(****
   Compiler Correctness proof for 
   Biorthogonality, Step-Indexing & Compiler Correctness
   Nick Benton & Chung-Kil Hur
   April 2009
 ****)

Require Import utility.
Require Import Predom.
Require Import PredomAux.
Require Import typedlambda.
Require Import typeddensem.
Require Import domoprs.
Require Import toptopsem_dom.
Set Implicit Arguments.
Unset Strict Implicit.

(***
   Compiler 
***)

Fixpoint compileE E T (exp:Exp E T) : Code :=
  match exp with
    | TOP op dt1 dt2 => (compileV dt1)++(compileV dt2)++[Op op]
    | TGT dt1 dt2 => (compileV dt1)++(compileV dt2)++[Swap, Op LT]
    | TAPP _ _ dt1 dt2 => (compileV dt1)++(compileV dt2)++[App]
    | TVAL _ dt => compileV dt
    | TLET _ _ dt1 dt2 => [PushC (compileE dt2++[Ret])]++(compileE dt1)++[App]
    | TIF t dtc dt1 dt2 => (compileV dtc)++[Sel (compileE dt1++[Join]) (compileE dt2++[Join])]
    | TFST _ _ dtp => (compileV dtp)++[ProjFst]
    | TSND _ _ dtp => (compileV dtp)++[ProjSnd]
  end

with compileV (E:Env) (T:Ty) (val:Value E T) : Code :=
  match val with
    | TINT n => [PushN n]
    | TBOOL b => [PushN (bool_to_nat b)]
    | TVAR _ i => [PushV (idxVar i)]
    | TLAM _ _ dtb => [PushC (compileE dtb++[Ret])]
    | TFIX _ _ dtb => [PushRC (compileE dtb++[Ret])]
    | TPAIR _ _ dt1 dt2 => (compileV dt1)++(compileV dt2)++[MkPair]
  end.


(****
   Compiler Correctness
 ****)

(* Compiler Correctness <c< *)
Theorem compiler_correctness_le : 
  (forall E T (val:Value E T), (compileV val) <P< sem_op _ eta (SemVal val)) /\
  (forall E T (exp:Exp E T), (compileE exp) <P< (SemExp exp)).
Proof.
apply ExpVal_ind.

(**
   TINT 
**)
do 4 red; simpl; intros.
eapply stepsfor_step.
apply step_pushn.
apply (H1 k0 (VN n)).
exists n.
split. 
rewrite sem_op_simpl, fcont_comp_simpl; rewrite eta_val; auto.
auto.
omega.
omega.

(**
   TBOOL
**)
do 4 red; simpl; intros.
eapply stepsfor_step.
apply step_pushn.
apply (H1 k0 (VN (bool_to_nat b))).
exists b.
split.
rewrite sem_op_simpl, fcont_comp_simpl; rewrite eta_val; auto.
destruct b.
simpl; auto.
simpl; auto.
omega.
omega.

(**
   TVAR 
**)
do 4 red; simpl; intros.
destruct (irelLeE_inv v H0) as [? [? ?]].
eapply stepsfor_step.
apply step_pushv.
apply nth_error_prefix.
apply H4.
apply (H1 k0 x).
exists (SemVar v de).
auto.
omega.
omega.


(**
   TLAM
**)
intros.
apply (compositionality_lamabs_le (c:= compileE e) (dc:= SemExp e)). auto.

(**
   TFIX
**)
do 3 red; intros.

(* Core of the proof appealing to the Lob rule *)
assert(forall m e', (m <= k0)%nat -> 
  @irelLeV _ m (VRC (e0++e') (compileE e ++ Ret:: nil)) 
  ((@FIXP _ _ @_ curry (curry (SemExp e))) de)
).
induction m.

(* Zero case *)
do 3 red; fold irelLeV; simpl; intros.
inversion H3.
rewrite H10 in H5.
inversion H5.
rewrite H11 in H8.
inversion H8.
apply stepsfor_zero.

(* Successor case *)
do 3 red; fold irelLeV; simpl; intros.
eapply stepsfor_step_strict.
apply step_rapp.

repeat rewrite app_comm_cons.
eapply H.

assert (Peano.pred k2 <= Peano.pred k1)%nat. omega. apply H10.

instantiate (1:= ((de,(FIXP @_ curry (curry (SemExp e))) de),dv)).
split.
apply irelLeV_decr with k1.
omega.
apply H4.

split.
apply irelLeV_decr with m. omega.
apply IHm. omega.

apply irelLeE_decr with k0. omega.
apply H1.

instantiate (1:= k3).
red; intros.
eapply stepsfor_step.
apply step_ret.

apply (H7 i).
do 2 destruct H10.
exists x.
split.

rewrite K_simpl.
rewrite fcont_comp_simpl in H10 |- *.
eapply Oeq_trans.
apply H10.

assert (TEMP := fcont_eq_elim (@FIXP_eq ((SemType _) -C-> DL (SemType _)) _ ((curry (curry (SemExp e)) de)))dv).
apply Oeq_sym. apply TEMP; clear TEMP.
apply H13.

omega.
omega.
omega.
omega.
(* End of the Core of the proof *)

red; simpl; intros.
eapply stepsfor_step.
apply step_pushrc.

apply (H3 k1).
exists ((@FIXP _ _ @_ curry (curry (SemExp e))) de).
split.
auto.
apply H2.

omega.
omega.
omega.

(**
   TP
**)
intros.
apply compositionality_pair_le.
auto. auto.

(**
   TFST
**)
do 4 red; simpl; intros. fold SemType in *.
repeat rewrite app_ass.
eapply H.
apply H0.
apply H1.
instantiate (1:= k1).

red; intros.
do 2 destruct H5.
destruct val.
inversion H8.
inversion H8.
inversion H8.
eapply stepsfor_step.
apply step_fst.

apply H2 with i.
simpl in x; destruct x.
exists t.
split.

rewrite sem_op_simpl, fcont_comp_simpl in H5.
rewrite eta_val in H5.
hreplace H5 with (DLleVal_eq2 H5).
repeat rewrite sem_op_simpl.
repeat rewrite fcont_comp_simpl.
rewrite eta_val.
simpl; apply Val_stable.
rewrite <- H5. auto.

simpl in H8; destruct H8.
assumption.

omega. omega. omega. omega.

(**
   TSND
**)
do 4 red; simpl; intros. fold SemType in *.
repeat rewrite app_ass.
eapply H.
apply H0.
apply H1.
instantiate (1:= k1).

red; intros.
do 2 destruct H5.
destruct val.
inversion H8.
inversion H8.
inversion H8.
eapply stepsfor_step.
apply step_snd.

apply H2 with i.
simpl in x; destruct x.
exists t0.
split.

rewrite sem_op_simpl, fcont_comp_simpl in H5.
rewrite eta_val in H5.
hreplace H5 with (DLleVal_eq2 H5).
repeat rewrite sem_op_simpl.
repeat rewrite fcont_comp_simpl.
rewrite eta_val.
simpl; apply Val_stable.
rewrite <- H5. auto.

simpl in H8; destruct H8.
assumption.

omega. omega. omega. omega.

(**
   TOP
**)
intros; rename o into op.
do 4 red; simpl; intros.
repeat rewrite app_ass.
eapply H.

apply H1.
apply H2.
instantiate (1:= k1).

red; intros.
eapply H0.
apply H1.
apply H2.
instantiate (1:= k2).

red; intros.
do 2 destruct H6.
simpl in x.
do 2 destruct H9.
simpl in x0.

destruct val.
destruct val0.

eapply stepsfor_step.
apply step_op.
apply H3 with k3.

simpl in H12.
inversion H12.
simpl in H13.
inversion H13.
exists (OpSem op x x0).
split.

rewrite sem_op_simpl, fcont_comp_simpl, sem_op2_eval, eta_val.
simpl; apply Val_stable.
rewrite sem_op_simpl, fcont_comp_simpl, eta_val in H6, H9.
hreplace H6 with (DLleVal_eq2 H6); simpl in H6.
hreplace H9 with (DLleVal_eq2 H9); simpl in H9.
rewrite <- H6.
rewrite <- H9.
auto.

auto.
omega.
omega.
inversion H13.
inversion H13.
inversion H13.
inversion H12.
inversion H12.
inversion H12.
omega.
omega.
omega.
omega.

(**
   TGT
**)
do 4 red; simpl; intros.
repeat rewrite app_ass.
eapply H.
apply H1.
apply H2.
instantiate (1:= k1).

red; intros.
eapply H0.
apply H1.
apply H2.
instantiate (1:= k2).

red; intros.
do 2 destruct H6.
simpl in x.
do 2 destruct H9.
simpl in x0.

destruct val.
destruct val0.

eapply stepsfor_step.
apply step_swap.
eapply stepsfor_step.
apply step_op.
apply H3 with k3.

simpl in H12.
inversion H12.
simpl in H13.
inversion H13.
exists (blt_nat x0 x).
split.

rewrite sem_op_simpl, fcont_comp_simpl, sem_op2_eval, eta_val.
simpl; apply Val_stable.
rewrite sem_op_simpl, fcont_comp_simpl, eta_val in H6, H9.
hreplace H6 with (DLleVal_eq2 H6); simpl in H6.
hreplace H9 with (DLleVal_eq2 H9); simpl in H9.
rewrite <- H6.
rewrite <- H9.
auto.

destruct (bool_dec (blt_nat x0 x) true).
repeat rewrite e1; simpl; auto.
rewrite e1.
auto.
hreplace n1 with (not_true_is_false (blt_nat x0 x) n1).
repeat rewrite n1; simpl; auto.
rewrite n1.
auto.

omega.
omega.
inversion H13.
inversion H13.
inversion H13.
inversion H12.
inversion H12.
inversion H12.
omega.
omega.
omega.
omega.

(**
   TVAL
**)
intros.
unfold compileE; fold compileV.
unfold SemExp; fold SemVal.
auto.

(**
   TLET
**)
intros.
unfold compileE; fold compileE.
eapply relLeC_le.
Focus 2.
apply (@compositionality_funapp_le _ _ _ (Lambda (compileE e0)) (compileE e) (sem_op _ eta (sem_curry _ _ _ (SemExp e0)))).
apply (@compositionality_lamabs_le _ _ _ (compileE e0) (SemExp e0) H0).
apply H.
fold SemType SemEnv.
rewrite (sem_app_curry_simpl (SemExp e0) (SemExp e)). auto.

(**
   TAPP 
**)
intros.
unfold compileE; fold compileV.
eapply relLeC_le.
Focus 2.
apply (@compositionality_funapp_le _ _ _ (compileV v) (compileV v0) (eta @_ SemVal v)).
apply H.
apply H0.
unfold SemExp; fold SemVal SemType.
rewrite (sem_app_val_simpl (SemVal v) (SemVal v0)). auto.

(**
   TIF
**)
do 4 red; simpl; intros.
repeat rewrite app_ass.
eapply H.
apply H2.
apply H3.
instantiate (1:= k1).

red; intros.
do 2 destruct H7.
rewrite sem_op_simpl, fcont_comp_simpl, eta_val in H7.
hreplace H7 with (DLleVal_eq2 H7).
simpl in H7.

destruct x.
(* true case *)
destruct val.
destruct n.
inversion H10.

eapply stepsfor_step.
apply step_seltrue.
red; intros; inversion H11.
eapply H0.
apply H2.
apply H3.
instantiate (1:= k2).

red; intros.
eapply stepsfor_step.
apply step_join.
apply H4 with i0.

do 2 destruct H11.
exists x.
split.

rewrite sem_op3_eval.
rewrite <- H7.
rewrite choose_t_simpl.
auto.

auto.
omega.
omega.
omega.
omega.
inversion H10.
inversion H10.
inversion H10.

(* false case *)
destruct val.
destruct n.

eapply stepsfor_step.
apply step_selfalse.
eapply H1.
apply H2.
apply H3.
instantiate (1:= k2).

red; intros.
eapply stepsfor_step.
apply step_join.
apply H4 with i0.

do 2 destruct H11.
exists x.
split.

rewrite sem_op3_eval.
rewrite <- H7.
rewrite choose_f_simpl.
auto.

auto.
omega.
omega.
omega.
omega.
inversion H10.
inversion H10.
inversion H10.
inversion H10.
omega.
omega.
Qed.

(* Compiler Correctness >c> *)

Fixpoint SemExpChain (env:Env) (t:Ty) (exp:Exp env t) {struct exp} : natO-m>(SemCTypeL env t) :=
match exp with
| TOP op dt1 dt2 => sem_op_cha eta (sem_op2_cha (SimpleOp2 (OpSem op)) (SemValChain dt1) (SemValChain dt2))
| TGT dt1 dt2 => sem_op_cha eta (sem_op2_cha (SimpleOp2 blt_nat) (SemValChain dt2) (SemValChain dt1))
| TAPP _ _ dt1 dt2 => sem_app_cha (sem_op_cha eta (SemValChain dt1)) (sem_op_cha eta (SemValChain dt2))
| TVAL _ dt => sem_op_cha eta (SemValChain dt)
| TLET _ _ dt1 dt2 => sem_app_cha (sem_op_cha eta (sem_curry_cha (SemExpChain dt2))) (SemExpChain dt1)
| TIF t dtc dt1 dt2 => sem_op3_cha (choose (DL (SemType t))) (SemExpChain dt1) (SemExpChain dt2) (SemValChain dtc)
| TFST _ _ dtp => sem_op_cha eta (sem_op_cha FST (SemValChain dtp))
| TSND _ _ dtp => sem_op_cha eta (sem_op_cha SND (SemValChain dtp))
end

with SemValChain (env:Env) (t:Ty) (val:Value env t) {struct val} : natO-m>(SemCTypeT env t) :=
match val with
| TINT n => fmon_cte _ (@K (SemEnv env) (SemType Int) n)
| TBOOL b => fmon_cte _ (@K (SemEnv env) (SemType Bool) b)
| TVAR t0 i => fmon_cte _ (SemVar i)
| TLAM _ _ dtb => sem_curry_cha (SemExpChain dtb)
| TFIX _ _ dtb => sem_fix_cha (SemExpChain dtb)
| TPAIR _ _ dt1 dt2 => sem_op2_cha PAIR (SemValChain dt1) (SemValChain dt2)
end.

Lemma SemChain_sound: 
  (forall E T (val:Value E T), @lub (SemCTypeT E T) (SemValChain val) == SemVal val) /\
  (forall E T (exp:Exp E T), @lub (SemCTypeL E T) (SemExpChain exp) == SemExp exp).

apply ExpVal_ind.

(* TINT *)
intros.
split.
apply lub_le; intros; unfold SemValChain, SemVal; simpl; auto.
apply (@le_lub (SemCTypeT env Int) (fmon_cte _ (@K _ (SemType Int) n)) 0).

(* TBOOL *)
intros.
split.
apply lub_le; intros; unfold SemValChain, SemVal; simpl; auto.
apply (@le_lub (SemCTypeT env Bool) (fmon_cte _ (@K _ (SemType Bool) b)) 0).

(* TVAR *)
intros.
split.
apply lub_le; intros; unfold SemValChain, SemVal; simpl; auto.
apply (@le_lub (SemEnv env -C-> (SemType _)) (fmon_cte _ (SemVar v)) 0).

(**
   TLAM
**)
intros.
unfold SemValChain, SemVal; fold SemExp SemExpChain. fold SemEnv.
rewrite (sem_curry_cha_lub (SemExpChain e)). fold SemEnv.
rewrite H. auto.

(**
   TFIX
**)
intros.
unfold SemValChain, SemVal; fold SemExp SemExpChain. fold SemEnv.
rewrite (sem_fix_cha_lub (SemExpChain e)).

rewrite H.
auto.

(**
   TP
**)
intros.

rewrite (sem_op2_cha_lub PAIR (SemValChain v) (SemValChain v0)).
setoid_replace (lub (SemValChain v0)) with (SemVal v0).
Focus 2.
auto.

rewrite H.
apply (@fcont_eq_intro (SemEnv env) (SemType (ty1**ty2))); intro.
auto.

(**
   TFST
**)
intros.
rewrite (sem_op_cha_lub eta (sem_op_cha FST (SemValChain v))); fold SemType.
rewrite (sem_op_cha_lub FST (SemValChain v)); fold SemType.
rewrite H.
apply (@fcont_eq_intro (SemEnv env) (DL (SemType ty1))); intro.
auto.

(**
   TSND
**)
intros.
rewrite (sem_op_cha_lub eta (sem_op_cha SND (SemValChain v))); fold SemType.
rewrite (sem_op_cha_lub SND (SemValChain v)); fold SemType.
rewrite H.
apply (@fcont_eq_intro (SemEnv env) (DL (SemType ty2))); intro.
auto.

(**
   TOP
**)
intros.
rename o into op.
rewrite (sem_op_cha_lub eta (sem_op2_cha (SimpleOp2 (OpSem op)) (SemValChain v) (SemValChain v0))).
rewrite (sem_op2_cha_lub (SimpleOp2 (OpSem op)) (SemValChain v) (SemValChain v0)).
apply (@fcont_eq_intro (SemEnv env) (DL (SemType Int))); intro.
rewrite H. rewrite H0. auto.

(**
   TGT
**)
intros.
rewrite (sem_op_cha_lub eta (sem_op2_cha (SimpleOp2 blt_nat) (SemValChain v0) (SemValChain v))).
rewrite (sem_op2_cha_lub (SimpleOp2 blt_nat) (SemValChain v0) (SemValChain v)).
apply (@fcont_eq_intro (SemEnv env) (DL (SemType Bool))); intro.
rewrite H. rewrite H0. auto.

(**
   TVAL
**)
intros.
rewrite (sem_op_cha_lub eta (SemValChain v)).
rewrite H.
auto.

(**
   TLET
**)
intros.
unfold SemExp; fold SemExp SemVal SemEnv.

rewrite (sem_app_cha_lub (sem_op_cha eta (sem_curry_cha (SemExpChain e0))) (SemExpChain e)); fold SemExpChain SemEnv.

assert (lub (sem_curry_cha (SemExpChain e0)) == sem_curry _ _ _ (SemExp e0)); fold SemEnv.
eapply Oeq_trans.
apply sem_curry_cha_lub.
destruct H0.
split. auto. auto.
fold SemEnv in H1.

rewrite sem_op_cha_lub.
rewrite H1. 
rewrite H.
auto.

(**
   TAPP
**)

intros.
rewrite (sem_app_cha_lub (sem_op_cha eta (SemValChain v)) (sem_op_cha eta (SemValChain v0))); fold SemType.

apply Oeq_trans with (sem_app _ _ _ (eta @_ SemVal v) (eta @_ SemVal v0)); fold SemType.

apply fcont_eq_compat2.
rewrite (sem_op_cha_lub eta (SemValChain v)). rewrite H. auto.
rewrite (sem_op_cha_lub eta (SemValChain v0)). rewrite H0. auto.
rewrite sem_app_val_simpl.
auto.

(* 
   TIF 
*)
intros.
rewrite (sem_op3_cha_lub (choose (DL (SemType ty))) (SemExpChain e) (SemExpChain e0) (SemValChain v)).
eapply Oeq_trans.
apply (fcont_eq_compat3 (sem_op3 (SemEnv env) (choose (DL (SemType ty)))) H0 H1 H).
apply (@fcont_eq_intro (SemEnv env) (DL (SemType ty))); intro.
auto.
Qed.

Lemma compiler_correctness_ge_chain:
  (forall E T (val:Value E T) i, CC (compileV val) >c> sem_op _ eta (SemValChain val i)) /\
  (forall E T (exp:Exp E T) i, CC (compileE exp) >c> (SemExpChain exp i)).

Proof.
apply ExpVal_ind.

(**
   TINT 
**)
do 4 red; simpl; intros.
eapply terminate_step. apply step_pushn.
apply H1. 
do 2 red; simpl.
hreplace H0 with (DLleVal_eq2 H0).
destruct H0. simpl in H0. 
rewrite H0. auto.

(**
   TBOOL
**)
do 4 red; simpl; intros.
eapply terminate_step. apply step_pushn.
apply H1.
hreplace H0 with (DLleVal_eq2 H0).
destruct H0. simpl in H0. rewrite H0.
destruct b.
do 2 red; simpl; auto.
do 2 red; simpl; auto.

(**
   TVAR
**)
do 4 red; simpl; intros.
destruct (erelGeE_inv v H).
destruct H2.

eapply terminate_step. eapply step_pushv. 
apply nth_error_prefix. apply H2.
apply H1.

rewrite sem_op_simpl, fcont_comp_simpl, eta_val in H0.
apply erelGeV_ge with (SemVar v de).
apply DLleVal_eq. apply H0.
assumption.

(**
   TLAM
**)
do 4 red. intros.
unfold compileV. fold compileE.
eapply terminate_step.
apply step_pushc.
apply H2. do 2 red.

red; fold erelGeV. do 3 red. intros.
repeat rewrite <- app_comm_cons.
eapply terminate_step.
apply step_app.
repeat rewrite app_comm_cons.
eapply H.
instantiate (1:= (de,dv)).

split. apply H3. apply H0.
rewrite sem_op_simpl, fcont_comp_simpl, eta_val in H1.
rewrite K_simpl in H5.
hreplace H1 with (DLleVal_eq2 H1).
instantiate (1 := i). instantiate (1 := d1).
eapply Oeq_trans. apply H5.
unfold SemValChain in H1. fold SemExpChain SemEnv in H1.

apply Oeq_trans with (sem_curry_cha (SemExpChain e) i de dv). fold SemEnv.
rewrite H1. reflexivity.
apply Oeq_trans with (sem_curry _ _ _ (SemExpChain e i) de dv). fold SemEnv.
apply (@fcont_eq_elim (SemType ty1) (DL (SemType ty2))).
apply (@fcont_eq_elim (SemEnv env) (SemType (ty1-->ty2))).
auto.
auto.

red. intros.
eapply terminate_step. apply step_ret.
apply H6. assumption.

(**
   TFIX
**)

do 2 red; intros.
(* main proof *)
assert (forall i0 e', @erelGeV (ty1-->ty2) (VRC (e0++e') (compileE e ++ Ret:: nil)) (sem_fix_cha (SemExpChain e) i0 de)).

induction i0.
(* base case *)
intros.
eapply erelGeV_ge.
apply (sem_fix_cha_ind_0 (SemExpChain e) de).
apply val_ge_bot_fun.

(* induction step *)
intros.
eapply erelGeV_ge.
apply (proj1 (sem_fix_cha_ind_S (SemExpChain e) i0 de)).

do 4 red; fold erelGeV; intros; simpl.
eapply terminate_step. apply step_rapp.
repeat rewrite app_comm_cons.
eapply H.
instantiate (1:= ((de,sem_fix_cha (SemExpChain e) i0 de),dv)).

split.
apply H2. 
split.
apply IHi0.
apply H0.
rewrite K_simpl in H4.
eapply Oeq_trans.
apply H4.
instantiate (1:= i0).
rewrite <- (sem_fix_cha_ind_S (SemExpChain e) i0 de).
rewrite sem_fix_cha_ind_S.
repeat rewrite curry_simpl.
auto.

red; intros.
eapply terminate_step. apply step_ret.
apply H5.
assumption.
(* end of main proof *)

red; simpl; intros.
eapply terminate_step.
apply step_pushrc.
apply H3; red.

assert (VRC (e0 ++ e1) (compileE e ++ [Ret]) >v> d).
apply erelGeV_ge with (sem_fix_cha (SemExpChain e) i de).
apply DLleVal_eq.
eapply DLle_trans.
apply H1.
apply Ole_to_DLle.
rewrite sem_op_simpl, fcont_comp_simpl, eta_val.
simpl. apply DLle_leVal.
auto.
apply H2.
apply H4.

(**
   TP
**)
do 4 red; simpl; intros.
repeat rewrite app_ass.

simpl in d; destruct d.
rewrite sem_op_simpl, fcont_comp_simpl, eta_val in H2.
hreplace H2 with (DLleVal_eq2 H2).
destruct H2.
simpl in H2, H4.

eapply H.
apply H1.
rewrite sem_op_simpl, fcont_comp_simpl, eta_val.
simpl. apply eqDLeq.
split. apply H2. apply H4.

red; intros.
eapply H0.
apply H1.
rewrite sem_op_simpl, fcont_comp_simpl, eta_val.
simpl. apply eqDLeq.
split. apply H2. apply H4.

red; intros.
eapply terminate_step. apply step_mkpair.

eapply H3.
do 2 red in H5, H6 |- *.
tauto.

(**
   TFST
**)
do 4 red; simpl; fold SemType; intros.
repeat rewrite app_ass.

repeat rewrite sem_op_simpl in H1.
repeat rewrite fcont_comp_simpl in H1.
rewrite eta_val in H1.
hreplace H1 with (DLleVal_eq2 H1).

eapply H.
apply H0.
rewrite sem_op_simpl, fcont_comp_simpl, eta_val.
apply eqDLeq.
instantiate (1:= i).
reflexivity.

red; intros.
destruct val; try (inversion H3).

eapply terminate_step. apply step_fst.

eapply H2.
do 2 red in H3 |- *.
assert (@erelGeV (ty1**ty2) (VP val1 val2) (fst (SemValChain v i de), snd (SemValChain v i de))).
eapply erelGeV_ge.
Focus 2. apply H3.
auto.
simpl in H4.
destruct H4.
eapply erelGeV_ge.
apply H1.
apply H4.

(**
   TSND
**)
do 4 red; simpl; fold SemType; intros.
repeat rewrite app_ass.

repeat rewrite sem_op_simpl in H1.
repeat rewrite fcont_comp_simpl in H1.
rewrite eta_val in H1.
hreplace H1 with (DLleVal_eq2 H1).

eapply H.
apply H0.
rewrite sem_op_simpl, fcont_comp_simpl, eta_val.
apply eqDLeq.
instantiate (1:= i).
reflexivity.

red; intros.
destruct val; try (inversion H3).

eapply terminate_step. apply step_snd.

eapply H2.
do 2 red in H3 |- *.
assert (@erelGeV (ty1**ty2) (VP val1 val2) (fst (SemValChain v i de), snd (SemValChain v i de))).
eapply erelGeV_ge.
Focus 2. apply H3.
auto.
simpl in H4.
destruct H4.
eapply erelGeV_ge.
apply H1.
apply H5.

(**
   TOP
**)
do 4 red.

intros. 
simpl.
rename o into op. 
repeat rewrite app_ass.

eapply H.
apply H1.
rewrite sem_op_simpl, fcont_comp_simpl, eta_val.
simpl. apply eqDLeq.
instantiate (1:= i).
reflexivity.

red; intros.
eapply H0.
apply H1.
rewrite sem_op_simpl, fcont_comp_simpl, eta_val.
simpl. apply eqDLeq.
instantiate (1:= i).
reflexivity.

red; intros.
destruct val0; try (inversion H5).
destruct val; try (inversion H4).
eapply terminate_step. apply step_op.

eapply H3.
do 2 red in H4, H5 |- *.
simpl in H4, H5.
inversion H4.
inversion H5.
simpl SemExpChain in H2.
rewrite sem_op_cha_eval, eta_val in H2.
hreplace H2 with (DLleVal_eq2 H2).
rewrite sem_op2_cha_eval in H2.
eapply erelGeV_ge.
apply H2.
simpl.
auto.

(**
   TGT
**)
do 4 red.

intros. 
simpl.
repeat rewrite app_ass.

eapply H.
apply H1.
rewrite sem_op_simpl, fcont_comp_simpl, eta_val.
simpl. apply eqDLeq.
instantiate (1:= i).
reflexivity.

red; intros.
eapply H0.
apply H1.
rewrite sem_op_simpl, fcont_comp_simpl, eta_val.
simpl. apply eqDLeq.
instantiate (1:= i).
reflexivity.

red; intros.
destruct val0; try (inversion H5).
destruct val; try (inversion H4).
eapply terminate_step. simpl. apply step_swap.
eapply terminate_step. apply step_op.
eapply H3.
red in H4, H5 |- *.
simpl in H4, H5.
inversion H4.
inversion H5.
simpl SemExpChain in H2.
rewrite sem_op_cha_eval, eta_val in H2.
hreplace H2 with (DLleVal_eq2 H2).
rewrite sem_op2_cha_eval, SimpleOp2_simpl in H2.
eapply erelGeV_ge.
apply H2.

simpl.
assert (forall b, match bool_to_nat b with | 0 => b = false | S _ => b = true end).
destruct b.
simpl; auto.
simpl; auto.
apply H6.

(**
   TVAL
**)
intros.
apply erelGeC_ge with (sem_op _ eta (SemValChain v i)).
auto.
auto.

(**
   TLET
**)
intros.
unfold compileE; fold compileE.
unfold SemExpChain; fold SemExpChain SemEnv.
eapply erelGeC_ge. rewrite (sem_app_cha_eval (E:=SemEnv env) (A:=SemType ty1) (B:= SemType ty2)). reflexivity.
assert (H1 := compositionality_funapp_ge_e). unfold ComposeApp in H1.
apply H1. clear H1.
apply (compositionality_lamabs_ge_e (H0 i)).
auto.

(**
   TAPP
**)
intros.
unfold compileE; fold compileV.
unfold SemExpChain; fold SemValChain SemType.
apply (compositionality_funapp_ge_e (H i) (H0 i)).

(**
   TIF
**)
do 4 red. intros.
unfold compileE; fold compileE compileV.
repeat rewrite app_ass.
eapply H.
apply H2.
rewrite sem_op_simpl, fcont_comp_simpl, eta_val.
apply eqDLeq.
instantiate (1:= i).
reflexivity.

assert (Val d == choose (DL (SemType ty)) (SemExpChain e i de) (SemExpChain e0 i de) (SemValChain v i de)).
eapply Oeq_trans.
apply H3.
auto.

red; intros.
destruct val; try (inversion H6).
destruct n.

(* false case *)
do 2 red in H6. simpl in H6.
assert (choose (DL (SemType ty)) (SemExpChain e i de) (SemExpChain e0 i de) (SemValChain v i de) == (SemExpChain e0 i de)).
apply Oeq_trans with (choose _ (SemExpChain e i de) (SemExpChain e0 i de) false).
apply fcont_eq_compat2.
auto. auto. auto.
rewrite H7 in H5.

eapply terminate_step. apply step_selfalse.

eapply H1.
apply H2.
apply H5.

red; intros.
eapply terminate_step. apply step_join.

eapply H4.
assumption.

(* true case *)
red in H6; simpl in H6.
assert (choose (DL (SemType ty)) (SemExpChain e i de) (SemExpChain e0 i de) (SemValChain v i de) == (SemExpChain e i de)).
apply Oeq_trans with (choose _ (SemExpChain e i de) (SemExpChain e0 i de) true).
apply fcont_eq_compat2.
auto. auto. auto.
rewrite H7 in H5.

eapply terminate_step. apply step_seltrue.
red; intros. inversion H8.

eapply H0.
apply H2.
apply H5.

red; intros.
eapply terminate_step. apply step_join.

eapply H4.
assumption.
Qed.

Lemma compiler_correctness_ge:
  (forall E T (val:Value E T), (compileV val) >P> sem_op _ eta (SemVal val)) /\
  (forall E T (exp:Exp E T), (compileE exp) >P> (SemExp exp)).
Proof.
split.
  intros. red.
  eapply ideal_closure_closed.
  rewrite <- (proj1 SemChain_sound).
  rewrite (lub_comp_eq_cont (sem_op (SemEnv E) (A:=SemType T) eta)).
  reflexivity.
  apply (proj1 (ideal_closure_closed (erelGeC (CC (compileV val))))).
  intros.
  apply ideal_closure_sub.
  eapply erelGeC_ge. rewrite fmon_comp_simpl, fcontit_simpl. reflexivity.
  apply compiler_correctness_ge_chain.

  intros. red.
  eapply ideal_closure_closed.
  rewrite <- (proj2 SemChain_sound). reflexivity.
  apply (proj1 (ideal_closure_closed (erelGeC (CC (compileE exp))))).
  intros.
  apply ideal_closure_sub.
  apply compiler_correctness_ge_chain.
Qed.

Corollary compiler_correctness :
  (forall E T (val:Value E T), (compileV val) =P= (@eta _ @_ (SemVal val))) /\
  (forall E T (exp:Exp E T), (compileE exp) =P= (SemExp exp)).

split.
split.
apply compiler_correctness_le.
apply compiler_correctness_ge.
split.
apply compiler_correctness_le.
apply compiler_correctness_ge.
Qed.


Corollary lam_correctness_le: forall k E A B (fbody: Exp (A::E) B) e de e',
  k:- e <e< de -> k:- VC (e++e') (compileE fbody++[Ret]) <v< SemVal (TLAM fbody) de.
intros.
do 3 red; fold irelLeV; simpl; intros.
eapply stepsfor_step.
apply step_app.
assert ((compileE fbody) <P< (SemExp fbody)) by apply compiler_correctness_le.
assert (k0:- (v::e) <e< ((de,dv):SemEnv (A::E))).
split. apply H1. apply irelLeE_decr with k. omega. assumption.

apply (H7 k k0 (v::e) (de,dv) H0 H8 k2 
  [Ret] e' nil ((c, e0 ++ e1, s) :: d)).
red. intros. 
eapply stepsfor_step.
apply step_ret.

do 2 destruct H9.
apply H4 with k3. rewrite K_simpl.
exists x.
split. apply H9.
apply irelLeV_decr with i. omega. assumption.
omega. omega. omega. omega.
Qed.

Corollary lam_correctness_ge: forall k E A B (fbody: Exp (A::E) B) e de e',
  e >e> de -> VC (e++e') (compileE fbody++[Ret]) >v> SemValChain (TLAM fbody) k de.
intros.
do 4 red; fold erelGeV; intros; simpl.
eapply terminate_step.
apply step_app.
destruct compiler_correctness_ge_chain. clear H4. 
hreplace H5 with (H5 _ _ fbody k).

eapply (H5 (v::e) (de,dv)).
split. apply H0. assumption.
instantiate (1:= d).
rewrite K_simpl in H2.
eapply Oeq_trans. apply H2.
unfold SemValChain; fold SemExpChain SemEnv.
rewrite (fcont_eq_elim (fcont_eq_elim (sem_curry_cha_eval (SemExpChain fbody) k) de) dv).
fold SemEnv. auto.

red; intros.
eapply terminate_step. apply step_ret.
apply H3. assumption.
Qed.

Corollary fix_correctness_le: forall k E A B (fbody: Exp (A::A-->B::E) B) e de e',
  k:- e <e< de -> k:- VRC (e++e') (compileE fbody++[Ret]) <v< SemVal (TFIX fbody) de.

assert(forall gk k E A B (fbody: Exp (A::A-->B::E) B) e de e', (k <= gk)%nat -> 
  k :- e <e< de -> k :- VRC (e++e') (compileE fbody ++ [Ret]) <v< SemVal (TFIX fbody) de).
induction gk.
intros. do 4 red. intros. assert (k2 = 0) by omega. rewrite H8. apply stepsfor_zero.

intros.
do 3 red; fold irelLeV; simpl; intros.
eapply stepsfor_step_strict.
apply step_rapp.
apply stepsfor_decr with (Peano.pred k2). omega.
assert ((compileE fbody) <P< (SemExp fbody)) by apply compiler_correctness_le.
assert (Peano.pred k2 :- (v::VRC (e++e') (compileE fbody ++ [Ret])::e) <e< ((de,SemVal(TFIX fbody) de ,dv):SemEnv (A::(A-->B)::E))).
split. apply irelLeV_decr with k0. omega. assumption.
split. rewrite FST_simpl, Fst_simpl, SND_simpl, Snd_simpl.
apply (IHgk (Peano.pred k2) E A B fbody e de). omega.
apply irelLeE_decr with k. omega. apply H0.
apply irelLeE_decr with k. omega. apply H0.

assert (Peano.pred k2 <= Peano.pred k2)%nat by omega.
rewrite app_comm_cons.
apply (H8 (Peano.pred k2) (Peano.pred k2) _ _ H10 H9 k2).
red; intros.
eapply stepsfor_step.
apply step_ret.
apply H5 with k3. rewrite K_simpl.
do 2 destruct H11. exists x.
split. unfold SemVal in H11. fold SemExp in H11. 
(* rewrite fcont_comp_simpl in H11 |- *. *)
apply Oeq_trans with (curry (curry (SemExp fbody)) de (FIXP (curry (curry (SemExp fbody)) de)) dv).
Focus 2. apply (@fcont_eq_elim (SemType A) (DL (SemType B))).
rewrite FIXP_eq at 2. auto.
repeat rewrite curry_simpl. assumption.
apply irelLeV_decr with i . omega. assumption.
omega. omega. omega. omega.

intros. apply (H k k). omega. assumption.
Qed.


Corollary fix_correctness_ge: forall k E A B (fbody: Exp (A::A-->B::E) B) e de e',
  e >e> de -> VRC (e++e') (compileE fbody++[Ret]) >v> SemValChain (TFIX fbody) k de.

induction k.

intros. eapply erelGeV_ge. Focus 2. apply val_ge_bot_fun. auto.

do 4 red; fold erelGeV; intros. simpl.
eapply terminate_step.
apply step_rapp.
destruct compiler_correctness_ge_chain. clear H4. 
hreplace H5 with (H5 _ _ fbody k).

repeat rewrite app_comm_cons.
rewrite K_simpl in H2.
unfold CC in H5.
eapply (H5 (v:: _ :: e) (de, SemValChain (TFIX fbody) k de, dv)).
split. apply H0.
split. eapply erelGeV_ge. Focus 2. apply IHk. apply H. auto.
apply H. eapply Oeq_trans. apply H2.

unfold SemValChain at 1; fold  SemExpChain.
rewrite (fcont_eq_elim (sem_fix_cha_ind_S (SemExpChain fbody) k de)). auto.

red; intros.
eapply terminate_step. apply step_ret.
apply H3. assumption.
Qed.

