(****
   Auxiliary facts about the denotational semantics for
   Biorthogonality, Step-Indexing & Compiler Correctness
   Nick Benton & Chung-Kil Hur
   April 2009
 ****)

Require Import utility.
Require Import Predom.
Require Import PredomAux.
Require Import typedlambda.
Require Import typeddensem.
Set Implicit Arguments.
Unset Strict Implicit.

Open Scope nat_scope.
Open Scope type_scope.

Definition prdc (A:Type) := A->Prop.
Definition sub (A : Type) (p q : prdc A) := forall a, p a -> q a.
Definition sub2 (A B: Type) (p q : A -> B -> Prop) := forall a b, p a b -> q a b.
Definition sub3 (A B C: Type) (p q : A -> B -> C -> Prop) := forall a b c, p a b c -> q a b c.
Definition union (A : Type) (p q : prdc A) : prdc A := fun a => p a \/ q a.
Definition intersection (A:Type) (p q : prdc A) : prdc A := fun a => p a /\ q a.
Fixpoint forall_elmt_of_list (A:Type) (P:A->Prop) (l:list A) : Prop := 
  match l with | nil => True | (a::tl) => P a /\ forall_elmt_of_list P tl end.


(****
   General Lemmas
 ****)
(* Section General_Lemmas. *)

Lemma le_simple_ind: forall (P:nat->Prop), (forall k, P (S k) -> P k) -> (forall k k', k <= k' -> P k' -> P k).
intros.
induction H0.
assumption.
apply IHle.
apply H.
assumption.
Qed.

Lemma min_le_split: forall k n m, k <= min n m -> k <= n /\ k <= m.
intros.
split.
eapply le_trans.
apply H.
auto with arith.
eapply le_trans.
apply H.
auto with arith.
Qed.

Lemma min_idemp: forall n, min n n = n.
intros.
auto with arith.
Qed.

Lemma min_assoc: forall n m l, (min (min n m) l) = min n (min m l).
induction n.
simpl.
auto.
destruct m.
auto.
destruct l.
auto.
simpl.
rewrite IHn.
auto.
Qed.

Lemma min_ordpreserve: forall n m n' m', n <= n' -> m <= m' -> min n m <= min n' m'.
induction n.
auto with arith.

intros.
destruct m.
auto with arith.
simpl.
destruct n'.
inversion H.
destruct m'.
inversion H0.
simpl.
apply le_n_S.
apply IHn.
auto with arith.
auto with arith.
Qed.

Lemma nth_error_prefix: forall (A:Type) l l' n (x:A),
  nth_error l n = value x -> nth_error (l++l') n = value x.
induction l.
intros.
destruct n.
inversion H.
inversion H.

intros.
destruct n.
auto.
simpl in H |- *.
apply IHl.
auto.
Qed.

Lemma nth_error_det: forall (A:Type) l n, (exists x:A, nth_error l n = value x) \/ nth_error l n = error.
induction l.

intros.
right.
destruct n.
auto.
auto.

intros.
destruct n.
left.
exists a.
auto.
simpl.
apply IHl.
Qed.

Lemma list_dec: forall (A:Type) (l:list A), l = nil \/ (exists a, exists tl, l = a::tl).
intros.
destruct l.
auto.
right.
exists a.
exists l.
auto.
Qed.

Lemma sub_trans: forall (A:Type) (P Q R:prdc A), sub P Q -> sub Q R -> sub P R.
intros.
red.
intros.
apply H0.
apply H.
assumption.
Qed.

Lemma minus_var: forall n n' m m', n <= n' -> m' <= m -> n - m <= n' - m'.
induction n.
intros.
auto with arith.
intros.
destruct n'.
inversion H.
destruct m.
inversion H0.
simpl.
auto.
destruct m'.
apply le_trans with (S n).
apply le_minus.
simpl.
auto.
simpl.
apply IHn.
apply le_S_n.
auto.
apply le_S_n.
auto.
Qed.

Lemma list_app_inv: forall (A:Type) (l:list A) s l' s', length l = length l' -> l++s = l'++s' -> l = l' /\ s = s'.

induction l.
intros.
destruct l'.
simpl in H, H0 |- *.
tauto.
inversion H.

intros.
destruct l'.
inversion H.
simpl in H, H0 |- *.
inversion H0.
inversion H.
destruct (IHl s l' s' H4 H3).
rewrite H1.
auto.
Qed.

Lemma cons_to_app: forall (A:Type) (a:A) (l:list A), a::l = (a::nil)++l.
simpl; auto.
Qed.

(* End General_Lemmas. *)




Open Scope O_scope.






(****
   Domain Operations
****)
(* Section Domain_Operations. *)


(* fmon <= *)
Definition fmon_less (A B: ord): relation (fmono A B).
intros. exact (@Ole (@fmon A B)).
Defined.

Definition fmon_less_preorder (A B: ord): PreOrder (@fmon_less A B).
intros. split. red; intros. apply Ole_refl.
red; intros. eapply Ole_trans. apply H. apply H0.
Defined.

Existing Instance fmon_less_preorder.

Add Parametric Morphism (A B : ord) :
  (@fmonot A B) with signature (@fmon_less A B ==> @Ole A ==> @Ole B)
 as fmon_less_compat.
intros. apply Ole_trans with (y x0). auto.
apply (fmonotonic y). auto.
Qed.

(* fmon == *)
Definition fmon_equal (A B: ord): relation (fmono A B).
intros. exact (@Oeq (@fmon A B)).
Defined.

Definition fmon_equal_equivalence (A B: ord): Equivalence (@fmon_equal A B).
intros. split. red; intros. apply Oeq_refl.
split. apply (Oeq_sym H). apply H.
red; intros. eapply Oeq_trans. apply H. apply H0.
Defined.

Existing Instance fmon_equal_equivalence.

Add Parametric Morphism (A B : ord) :
  (@fmonot A B) with signature (@fmon_equal A B ==> @Oeq A ==> @Oeq B)
 as fmon_equal_compat.
intros. apply Oeq_trans with (y x0). auto.
apply (monotonic_stable (fmonotonic  y)). auto.
Qed.

(* fcont <= *)
Definition fcont_less (A B: cpo): relation (fconti A B).
intros. exact (@Ole (@fcont_ord A B)).
Defined.

Definition fcont_less_preorder (A B: cpo): PreOrder (@fcont_less A B).
intros. split. red; intros. apply Ole_refl.
red; intros. eapply Ole_trans. apply H. apply H0.
Defined.

Existing Instance fcont_less_preorder.

Add Parametric Morphism (A B : cpo) :
  (@fconti_fun A B) with signature (@fcont_less A B ==> @Ole A ==> @Ole B)
 as fcont_less_compat.
intros. apply Ole_trans with (y x0). auto.
apply (fmonotonic (fcontit y)). auto.
Qed.

(* fcont == *)

Definition fcont_equal (A B: cpo): relation (fconti A B).
intros. exact (@Oeq (@fcont_ord A B)).
Defined.

Definition fcont_equal_equivalence (A B: cpo): Equivalence (@fcont_equal A B).
intros. split. red; intros. apply Oeq_refl.
split. apply (Oeq_sym H). apply H.
red; intros. eapply Oeq_trans. apply H. apply H0.
Defined.

Existing Instance fcont_equal_equivalence.

Add Parametric Morphism (A B : cpo) :
  (@fconti_fun A B) with signature (@fcont_equal A B ==> @Oeq A ==> @Oeq B)
 as fcont_equal_compat.
intros. apply Oeq_trans with (y x0). auto.
apply (monotonic_stable (fmonotonic (fcontit y))). auto.
Qed.

(*
Goal forall (A B C:cpo) (f g: A -c> B -C-> C) (a a' : A) (b b': B), f <= g -> a == a' -> b <= b' -> f a b <= g a' b'.
intros. rewrite H. rewrite H0. rewrite H1. auto.
Qed.

Goal forall (A B C:ord) (f g: A -m> B -m> C) (a a' : A) (b b': B), f <= g -> a == a' -> b <= b' -> f a b <= g a' b'.
intros. rewrite H. rewrite H0. rewrite H1. auto.
Qed.

Goal forall (A B C:cpo) (f g: A -m> B -m> C) (a a' : A) (b b': B), f <= g -> a == a' -> b <= b' -> f a b <= g a' b'.
intros. rewrite H. rewrite H0. rewrite H1. auto.
Qed.

Goal forall (A B C:cpo) (f g: A -c> B -M-> C) (a a' : A) (b b': B), f <= g -> a == a' -> b <= b' -> f a b <= g a' b'.
intros. rewrite H. rewrite H0. rewrite H1. auto.
Qed.
*)

Lemma Fcontit_simpl: forall (D1 D2:cpo) (f:D1-C->D2) (x:D1),
  Fcontit _ _ f x == f x.
auto.
Qed.

Lemma fcontit_simpl: forall (D1 D2:cpo) (f:D1-C->D2) (x:D1),
  fcontit f x == f x.
auto.
Qed.

Lemma kleisli_cc_eval: forall D E (f: D -C-> DL E) (a: D),
  kleisli_cc _ _ f (Val a) == f a.
intros. rewrite kleisli_cc_simpl. rewrite kleisli_inv. auto.
Qed.

Lemma KLEISLIR_simpl: forall A B C (f:Dprod A B -C-> DL C) a b,
  KLEISLIR f (a, Val b) == f (a, b).

intros. 
unfold KLEISLIR.
rewrite fcont_comp_simpl.
rewrite kleisli_cc_simpl, kleisli_inv.
auto.
Qed.

Definition KLEISLIR_m : forall A B C : cpo, (Dprod A B -C-> DL C) -m> Dprod A (DL B) -C-> DL C.
intros. exists (@KLEISLIR A B C). red; intros.
unfold KLEISLIR. rewrite H. auto.
Defined.

Lemma KLEISLIR_m_simpl : forall (A B C: cpo) x, KLEISLIR_m A B C x == KLEISLIR x.
auto.
Qed.

Definition KLEISLIR_cc : forall A B C : cpo, (Dprod A B -C-> DL C) -C-> Dprod A (DL B) -C-> DL C.
intros. exists (@KLEISLIR_m A B C).
red; intros. rewrite KLEISLIR_m_simpl. unfold KLEISLIR.
rewrite (@lub_comp_eq _ _ (fcontit (kleisli_cc (Dprod A B) C)) h). 
apply (@fcont_le_intro (Dprod A (DL B)) (DL C)). intros.
rewrite fcont_comp_simpl. 
repeat rewrite fcont_lub_simpl.
apply lub_le_compat.
apply fmon_le_intro. intros.
repeat rewrite fcont_app_simpl.
repeat rewrite fmon_comp_simpl.
rewrite KLEISLIR_m_simpl. unfold KLEISLIR.
rewrite fcont_comp_simpl.
auto. auto.
Defined.

Lemma KLEISLIR_cc_simpl : forall (A B C: cpo) x, KLEISLIR_cc A B C x == KLEISLIR_m A B C x.
auto.
Qed.


(* 
   General Lemmas 
*)
Lemma DLvalval2 : forall (D:cpo) (d:D) x, DLle (Val d) x -> 
 exists d', x == Val d' /\ (d<=d')%tord.
intros D d x H.
destruct (DLvalval H).
do 2 destruct H0.
exists x1.
split.
apply Oeq_sym.
apply Val_exists_pred_eq.
exists x0; auto.
auto.
Qed.

Lemma DL_bot_neq: forall (D:cpo) (d:D), ~ (Val d <= DL_bot D)%tord.
red; intros.
destruct (DLvalval H).
do 2 destruct H0.
induction x.
simpl in H0.
rewrite DL_bot_eq in H0.
inversion H0.
rewrite DL_bot_eq in H0.
simpl in H0.
apply IHx.
assumption.
Qed.

Lemma DLleVal_eq2 :
  forall (D:cpo) (d d':D), Val d == Val d' -> d == d'.

intros.
destruct H.
split.
apply DLleVal_eq; auto.
apply DLleVal_eq; auto.
Qed.


Lemma Val_stable: forall (D:cpo), stable (@Val D).
red; intros.
split. simpl. apply DLle_leVal. apply H.
simpl. apply DLle_leVal. apply H.
Qed.

Lemma DL_lub_val: forall (D:cpo) (f:natO-m> DL D) (d:D) , 
  (Val d <= DL_lub f)%tord -> exists k, exists d', f k == Val d'.

intros.
destruct (@DLnvalval _ 0 (Val d) d (DL_lub f)).
auto. assumption.
do 2 destruct H0.
unfold DL_lub in H0.
destruct (attempt2 H0).
do 2 destruct H2.
exists x1.
exists x2.
assumption.
Qed.

Lemma PROD_fun_mon: forall E A B (f f':E-C->A) (g g':E-C->B), 
  f <= f' -> g <= g' -> PROD_fun f g <= PROD_fun f' g'.

intros.
apply fcont_le_intro; intros.
repeat rewrite PROD_fun_simpl.
split.
simpl.
apply H.
simpl.
apply H0.
Qed.

Lemma Ole_to_DLle: forall (D:cpo) (x y: DL D), Ole x y -> DLle x y.
auto.
Qed.

Lemma DLle_to_Ole: forall (D:cpo) (x y: DL D), DLle x y -> Ole x y.
auto.
Qed.

Lemma lub_comp_eq_cont : forall (D1 D2 : cpo) (f : D1 -c> D2) (h : natO -m> D1),
  f (lub h) == lub (fcontit f @ h).
intros.
assert (continuous (fcontit f)) as cf by auto.
setoid_replace (f (lub h)) with (fcontit f (lub h)).
apply lub_comp_eq.
assumption.
auto.
Qed.

Lemma fcont_le_compat3 : forall (D1 D2 D3 D4:cpo) (f : D1-c>D2-C->D3-C->D4)
    (x x': D1) (y y':D2) (z z' : D3), x <= x' -> y <= y' -> z <= z' -> f x y z <= f x' y' z'.
intros.
apply Ole_trans with (f x y' z'). auto.
apply fcont_le_elim.
apply (@fcont_le_elim D2 (D3 -C-> D4)).
apply fmonotonic; auto.
Save.
Hint Resolve fcont_le_compat3.

Lemma fcont_eq_compat: forall (A B : cpo) (f: A-c> B),
  forall x0 y0 : A, x0 == y0 -> f x0 == f y0.
intros. apply fcont_equal_compat. red. auto. auto.
Qed.

Lemma fcont_eq_compat3 : forall (D1 D2 D3 D4:cpo) (f : D1-c>D2-C->D3-C->D4)
    (x x' : D1) (y y' : D2) (z z' : D3), x == x' -> y == y' -> z == z' -> f x y z == f x' y' z'.
intros; apply Ole_antisym; auto.
Save.
Hint Resolve fcont_eq_compat3.

Lemma kleisli_unit2: forall D d, kleisli d (@eta D @_ ID) == d.
intros D d.
split. simpl. apply DLless_cond.

intros dd kl. destruct (kleisliValVal kl) as [d1 [P1 P2]].
rewrite fcont_comp_simpl, eta_val in P2. rewrite P2 in P1.
rewrite (kleisli_Valeq (eta @_ ID) P1). auto.

simpl. apply DLless_cond.
intros dd kl.
rewrite (kleisli_Valeq (eta @_ ID) kl). auto.
Qed.

Lemma kleisli_cc_unit2: forall D, kleisli_cc D D (eta @_ ID) == @ID (DL D).
intros D. refine (fcont_eq_intro _).
intros d. rewrite kleisli_cc_simpl. rewrite ID_simpl. simpl.
assert (xx := kleisli_unit2). simpl. simpl in xx. apply xx.
Qed.

Lemma Smash_natural: forall (A A' B B': cpo) (f:A-C->A') (g:B-C->B') (a:DL A) (b:DL B),
  kleisli_cc _ _ (@eta _ @_ (PROD_map f g)) (Smash _ _ a b)
  == Smash _ _ (kleisli_cc _ _ (@eta _ @_ f) a) (kleisli_cc _ _ (@eta _ @_ g) b).
intros.
repeat rewrite kleisli_cc_simpl.
split. simpl. apply DLless_cond.

intros dd kl. destruct (kleisliValVal kl) as [d1 [P1 P2]].
rewrite kl.
unfold Smash in P1.
hreplace P1 with (Operator2Val P1).
destruct P1 as [d' [e' [Ha [Hb Hc]]]].
repeat rewrite <- kleisli_cc_simpl.

rewrite Ha, Hb.
repeat rewrite kleisli_cc_simpl, kleisli_inv.
rewrite <- P2.
repeat rewrite fcont_comp_simpl, eta_val.
unfold Smash. rewrite Operator2_simpl, eta_val.
rewrite eta_val in Hc.
hreplace Hc with (DLleVal_eq2 Hc).
simpl. apply DLle_leVal.

rewrite <- Hc. 
rewrite PROD_map_simpl.
auto.

simpl. apply DLless_cond.
intros dd kl.
rewrite kl.
unfold Smash in kl.
hreplace kl with (Operator2Val kl).
destruct kl as [d' [e' [Ha [Hb Hc]]]].
destruct (kleisliValVal Ha) as [d1 [P1 P2]].
destruct (kleisliValVal Hb) as [d1' [P1' P2']].
rewrite <- Hc.
rewrite <- kleisli_cc_simpl.

rewrite P1, P1'. unfold Smash. rewrite Operator2_simpl. auto.
rewrite kleisli_cc_simpl. rewrite kleisli_inv. simpl. rewrite fcont_comp_simpl.
repeat rewrite eta_val.
simpl. apply DLle_leVal.
rewrite PROD_map_simpl.
simpl.
rewrite fcont_comp_simpl, eta_val in P2, P2'.
hreplace P2 with (DLleVal_eq2 P2).
hreplace P2' with (DLleVal_eq2 P2').
rewrite P2.
rewrite P2'.
auto.
Qed.

Lemma fcont_app_Continuous: forall (O : ord) (D1 D2 : cpo) (f : O -m> D1 -c> D2) (h : natO -m> D1),
  (f <__> lub h) == lub (fcontit (fmon_fcont_shift f) @ h).
intros.
split.
apply fcont_app_continuous.
auto.
Qed.

Lemma lub_comp_simpl: forall (A B:cpo) (f:natO-m>A-C->B) (g: natO-m> A),
  lub f (lub g) == lub ((Fcontit A B @2 f) g).

intros.
rewrite fcont_lub_simpl.
rewrite fcont_app_Continuous.
rewrite double_lub_diag.
apply lub_eq_compat.
auto.
Qed.

Definition succO: natO-m> natO.
exists (fun n => S n).
red; intros. simpl in H |- *. omega.
Defined.

Lemma lub_succ_eq: forall (D:cpo) (f:natO-m>D),
  lub f == lub (f @ succO).
intros; split.
apply lub_le_compat; auto.
apply lub_le; intro; simpl; auto.
Qed.

Lemma SimpleOp2_simpl: forall A B C (op:A->B->C) a b, SimpleOp2 op a b == op a b.
auto.
Qed.

Lemma curry_cont: forall D1 D2 D3 (f:Dprod D1 D2 -c> D3), curry f == Curry f.
auto.
Qed.

(* 
   Domain operations
*)
Definition nat_to_bool (n:nat) : bool := match n with | 0 => false | _ => true end.
Definition bool_to_nat (b:bool) : nat := match b with | true => 1 | false => 0 end.

Definition lift_op C D (f:C-C->D) : (DL C)-C->(DL D) := 
  kleisli_cc _ _ (@eta _ @_ f).

Definition lift_op2 A B C (op:A-C->B-C->C) : DL A -C-> DL B -C-> DL C :=
  curry ((lift_op (uncurry _ _ _ op)) @_ (uncurry _ _ _ (@Smash A B))).

Definition lift_op3 A B C D (op:A-C->B-C->C-C->D) : DL A -C-> DL B -C-> DL C -C-> DL D :=
  curry (curry (
    lift_op (uncurry _ _ _ (uncurry _ _ _ op)) @_ 
    (uncurry _ _ _ (@Smash (Dprod A B) C)) @_ 
    (PROD_map (uncurry _ _ _ (@Smash A B)) ID))).


(*
   chain2
*)
Definition chain2 (A B : cpo) (ca : natO -m> A) (cb : natO -m> B) : natO -m> Dprod A B.
intros.
exists (fun n => (ca n, cb n)).
auto.
Defined.

Lemma chain2_lub_simpl : forall (A B C : cpo) (f : Dprod A B -c> C) (cp : natO -m> (Dprod A B)),
 f (lub cp) == f (lub (Fst A B @ cp), lub (Snd A B @ cp)).
intros.
auto.
Qed.

Lemma chain2_eval: forall A B ca cb n, @chain2 A B ca cb n == (ca n, cb n).
auto.
Qed.

Lemma chain2_fst : forall (A B : cpo) ca cb, Fst A B @ (chain2 ca cb) == ca.
intros.
auto.
Qed.

Lemma chain2_snd : forall (A B : cpo) ca cb, Snd A B @ (chain2 ca cb) == cb.
intros; auto.
Qed.

Definition chain2_fun (A B C : cpo) (f : A -C-> (B -C-> C)) 
                        (ca : natO -m> A) (cb : natO -m> B) : natO -m> C :=
 fcontit (uncurry A B C f) @ (chain2 ca cb).


Lemma chain2_fun_simpl : forall A B C (f : A -C-> (B -C-> C)) ca cb n,
 chain2_fun f ca cb n == f (ca n) (cb n).
intros.
unfold chain2_fun.
auto.
Qed.

Lemma chain2_fun_lub : forall A B C f ca cb, lub (@chain2_fun A B C f ca cb) == f (lub ca) (lub cb).
intros.
setoid_replace (f (lub ca) (lub cb)) with (uncurry A B C f (lub ca, lub cb)).
unfold chain2_fun.
rewrite <- lub_comp_eq_cont.
rewrite chain2_lub_simpl.
rewrite chain2_fst.
rewrite chain2_snd.
reflexivity.
auto.
Qed.

(*
   chain3
*)

Definition chain3 (A B C: cpo) (ca : natO -m> A) (cb : natO -m> B) (cc : natO -m> C): natO -m> Dprod (Dprod A B) C.
intros.
exists (fun n => ((ca n, cb n), cc n)).
red; intros.
auto.
Defined.

Lemma chain3_lub_simpl: forall (A B C D: cpo) (f : Dprod(Dprod A B) C -c> D) (cp : natO -m> Dprod (Dprod A B) C),
 f (lub cp) == f ((lub (Fst _ _ @ Fst _ _ @ cp), lub (Snd _ _ @ Fst _ _ @ cp)), lub (Snd _ _ @ cp)).
intros.
auto.
Qed.

Lemma chin3_fst : forall (A B C: cpo) ca cb cc, Fst _ _ @ Fst _ _ @ (@chain3 A B C ca cb cc) == ca.
intros.
auto.
Qed.

Lemma chain3_snd : forall (A B C: cpo) ca cb cc, Snd _ _ @ Fst _ _ @ (@chain3 A B C ca cb cc) == cb.
intros.
auto.
Qed.

Lemma chain3_third : forall (A B C: cpo) ca cb cc, Snd _ _ @ (@chain3 A B C ca cb cc) == cc.
intros.
auto.
Qed.

Definition chain3_fun (A B C D: cpo) (f : A -C-> B -C-> C -C-> D) 
                        (ca : natO -m> A) (cb : natO -m> B) (cc : natO -m> C): natO -m> D :=
 fcontit (uncurry _ _ _ (uncurry _ _ _ f)) @ (chain3 ca cb cc).

Lemma chain3_fun_simpl : forall A B C D (f : A -C-> B -C-> C -C-> D) ca cb cc n,
 chain3_fun f ca cb cc n == f (ca n) (cb n) (cc n).
intros.
unfold chain3_fun.
auto.
Qed.

Lemma chain3_fun_lub : forall A B C D f ca cb cc, lub (@chain3_fun A B C D f ca cb cc) == f (lub ca) (lub cb) (lub cc).
intros.
setoid_replace (f (lub ca) (lub cb) (lub cc)) with (uncurry _ _ _ (uncurry _ _ _ f) ((lub ca, lub cb), lub cc)).
unfold chain3_fun.
rewrite <- lub_comp_eq_cont.
rewrite chain3_lub_simpl.
apply fcont_eq_compat.
apply Dprod_eq_pair.
rewrite chin3_fst.
rewrite chain3_snd.
auto.
rewrite chain3_third.
auto.
auto.
Qed.


(*
   ideal closure
*)

Definition chain_closed (D:cpo) (P:prdc D) : Prop :=
  forall chs: natO-m>D, (forall i, P (chs i)) -> P (lub chs).

Definition down_closed (D:cpo) (P: prdc D) : Prop :=
  forall d d', (d <= d')%tord -> P d' -> P d.

Definition closed (D:cpo) (P:prdc D) := chain_closed P /\ down_closed P.

Definition ideal_closure (D:cpo) (P:prdc D) : prdc D := fun d =>
  forall (S: prdc D), sub P S -> chain_closed S -> down_closed S -> S d.

Lemma ideal_closure_covar: forall (D:cpo) (P P':prdc D),
  sub P P' -> sub (ideal_closure P) (ideal_closure P'). 
do 2 red; intros. apply H0.
eapply sub_trans. apply H. apply H1. auto. auto.
Qed.

Lemma ideal_closure_closed: forall (D:cpo) (P:prdc D),
  closed (ideal_closure P).
intros. split.
do 2 red; intros.
apply H1. intros. apply H. auto. auto. auto.

do 2 red; intros.
apply H3 with d'. auto.
apply H0. auto. auto. auto.
Qed.

Lemma ideal_closure_sub: forall (D:cpo) (P:prdc D),
  sub P (ideal_closure P).
do 2 red; intros. apply H0. auto.
Qed.

Lemma ideal_closure_flat: forall D (P: prdc (DL (Discrete D))) d,
  (forall p q, p == q -> P p -> P q) -> ideal_closure P (Val d) -> P (Val d).

intros. introvar P' (fun (x:DL (Discrete D)) =>  (exists x', x == Val x') -> P x) HP'.
assert (P' (Val d)).
apply H0. red; intros. auto.
  rewrite HP'. intros. auto.

  rewrite HP'. red; intros. destruct H2.
  destruct (lubval H2) as [i [x' [? ?]]]. simpl in H4. rewrite <- H4 in H2.
  eapply H. eapply Oeq_trans. apply H3. apply (Oeq_sym H2).
  apply H1. exists x'. auto.

  rewrite HP'. red; intros. destruct H3. rewrite H3 in H1.
  destruct (DLvalval2 H1) as [? [? ?]]. simpl in H5. rewrite <- H5 in *.
  eapply H. eapply Oeq_trans. apply H4. apply (Oeq_sym H3).
  apply H2. exists x. auto.

rewrite HP' in H1. apply H1. exists d. auto.
Qed.

Definition AppCpoSet (E D: cpo) (PC: prdc (E-C->D)) (PE: prdc E) : prdc D :=
  fun d => exists c, exists e, PC c /\ PE e /\ d == c e.

Lemma AppCpoSet_closed: forall (E D: cpo) (PC: prdc (E-C->D)) (PE: prdc E),
  sub (AppCpoSet (ideal_closure PC) (ideal_closure PE)) (ideal_closure (AppCpoSet PC PE)).

red; intros. destruct H as [c [e [? [? ?]]]].
eapply ideal_closure_closed. apply H1. clear H1 a.
generalize e H0. clear H0 e. apply H.
  red; intros. apply H1. 
    red; intros. apply ideal_closure_sub. exists a. exists a0. auto.
    red; intros. eapply ideal_closure_closed. apply (@lub_comp_eq _ _ _ _). auto.
    apply (ideal_closure_closed _). intros. simpl.
    eapply ideal_closure_closed. rewrite fcontit_simpl. apply Ole_refl. apply H2.

    red; intros. eapply ideal_closure_closed. apply fmonotonic. apply H2.
    eapply ideal_closure_closed. rewrite fcontit_simpl. apply Ole_refl. auto.

  red; intros. eapply ideal_closure_closed.
  rewrite <- (lub_cte e). apply (@lub_comp_simpl _ _ _ _).
  apply (ideal_closure_closed _). intros. 
  eapply ideal_closure_closed. simpl. 
  rewrite fcontit_simpl. apply Ole_refl. apply H0. auto.

  red; intros. eapply ideal_closure_closed.
  rewrite H0. apply Ole_refl. apply H1. auto.
Qed.

(*
   continuous closed
*)
Lemma cont_closed: forall D1 C (f:D1-C->C) (P: prdc D1) (S: prdc C),
  closed S -> (forall p, P p -> S (f p)) ->
  forall p, ideal_closure P p -> S (f p).

intros. apply H1. 
  red; intros. apply H0. auto.

  red; intros. eapply H. apply (lub_comp_eq_cont f chs).
  apply H. intros. eapply H. simpl. apply Ole_refl. apply H2.

  red; intros. eapply H. rewrite H2. apply Ole_refl. auto.
Qed.

Lemma cont2_closed: forall D1 D2 C (f:D1-C->D2-C->C) (P: prdc D1) (Q: prdc D2) (S: prdc C),
  closed S -> (forall p q, P p -> Q q -> S (f p q)) ->
  forall p q, ideal_closure P p -> ideal_closure Q q -> S (f p q).

intros. generalize q H2; clear H2 q. apply H1. 
  red; intros. apply H3.
    red; intros. eapply H0. auto. auto.

    red; intros. eapply H. rewrite <- (lub_cte a). 
    apply (chain2_fun_lub f _ _).
    apply H. intros. eapply H. simpl. apply Ole_refl. apply H4.

    red; intros. eapply H. rewrite H4. apply Ole_refl. auto.

  red; intros. eapply H. rewrite <- (lub_cte q).
    apply (chain2_fun_lub f _ _).
  apply H. intros. eapply H. simpl. apply Ole_refl. apply H2. auto.

  red; intros. eapply H. rewrite H2. apply Ole_refl. auto.
Qed.

Lemma cont3_closed: forall D1 D2 D3 C (f:D1-C->D2-C->D3-C->C) (P: prdc D1) (Q: prdc D2) (R: prdc D3) (S: prdc C),
  closed S -> (forall p q r, P p -> Q q -> R r -> S (f p q r)) ->
  forall p q r, ideal_closure P p -> ideal_closure Q q -> ideal_closure R r -> S (f p q r).

intros. generalize q H2 r H3; clear H3 r H2 q. apply H1.
  red; intros. generalize r H4; clear H4 r. apply H3. 
    red; intros. apply H5.
      red; intros. eapply H0. auto. auto. auto.

      red; intros. eapply H. rewrite <- (lub_cte a), <- (lub_cte a0). 
      apply (chain3_fun_lub f _ _ _).
      apply H. intros. eapply H. simpl. apply Ole_refl. apply H6.

      red; intros. eapply H. rewrite H6. apply Ole_refl. auto.

    red; intros. eapply H. rewrite <- (lub_cte a), <- (lub_cte r). 
    apply (chain3_fun_lub f _ _ _).
    apply H. intros. eapply H. simpl. apply Ole_refl. apply H4. auto.

    red; intros. eapply H. rewrite H4. apply Ole_refl. auto.

  red; intros. eapply H. rewrite <- (lub_cte q), <- (lub_cte r). 
  apply (chain3_fun_lub f _ _ _).
  apply H. intros. eapply H. simpl. apply Ole_refl. apply H2. auto. auto.

  red; intros. eapply H. rewrite H2. apply Ole_refl. auto.
Qed.

Lemma sem_precomp_simpl: forall E' E A (op:E'-C->E) (da:E-C->A),
  sem_precomp _ op da == da @_ op.
intros. apply (@fcont_eq_intro E' A). auto.
Qed.

Lemma sem_op_simpl: forall E A B (op:A-C->B) (da:E-C->A),
  sem_op E op da == op @_ da.
intros. auto.
Qed.

Lemma sem_op_eval: forall E A B (op:A-C->B) (da:E-C->A) e,
  sem_op E op da e == op (da e).
auto.
Qed.

Lemma sem_op2_eval: forall E A B C (op:A-C->B-C->C) (da:E-C->A) (db:E-C->B) (e:E),
  sem_op2 E op da db e == op (da e) (db e).
intros.
auto.
Qed.

Lemma sem_op3_eval: forall E A B C D (op:A-C->B-C->C-C->D) (da:E-C->A) (db:E-C->B) (dc:E-C->C) e,
  @sem_op3 E A B C D op da db dc e == op (da e) (db e) (dc e).
intros.
auto.
Qed.

Lemma sem_fix_eval: forall E A B f,
  sem_fix E A B f == FIXP @_ curry (curry f).
intros.
apply (@fcont_eq_intro (SemEnv E) (SemType (A-->B))); intros.
auto.
Qed.


Lemma sem_fix_unfold: forall E A B f e a,
  sem_fix E A B f e a == f (e, sem_fix E A B f e, a).

intros.
unfold sem_fix; fold SemEnv SemType.
rewrite curry_simpl.
repeat rewrite fcont_comp_simpl.
repeat rewrite uncurry_simpl.
repeat rewrite fcont_comp_simpl.
repeat rewrite CURRY_simpl.
rewrite FIXP_eq at 1.
repeat rewrite Curry_simpl.
auto.
Qed.

Fixpoint sem_fix_chain_core E A B (n:nat) : Dprod ((SemEnv E) -C-> (SemType (A-->B)) -C-> (SemType (A-->B))) (SemEnv E) -C-> (SemType (A-->B)) :=
  match n with 
    | 0 => @K _ (_ -C-> _) (@K _ (DL _) (DL_bot _))
    | S m => EV @_ (PROD_fun EV (sem_fix_chain_core E A B m))
  end.

Definition sem_fix_chain_cc E A B (n:nat) : (SemCTypeL (A::A-->B::E) B)-C->(SemCTypeT E (A-->B)) :=
  CURRY _ _ _ (sem_fix_chain_core E A B n) @_ CURRY _ _ _ @_ CURRY _ _ _.

Lemma sem_fix_chain_core_unfold: forall E A B n,
  sem_fix_chain_core E A B (S n) = EV @_ (PROD_fun EV (sem_fix_chain_core E A B n)).
auto.
Qed.

Definition sem_fix_chain E A B (F: SemCTypeL (A::A-->B::E) B) : natO -m> SemCTypeT E (A-->B).
intros. exists (fun n => sem_fix_chain_cc _ _ _ n F).
assert (forall n, sem_fix_chain_core E A B n <= sem_fix_chain_core _ _ _ (S n)).
induction n.
  intros. 
  apply (@fcont_le_intro (Dprod ((SemEnv E) -C-> (SemType (A-->B)) -C-> (SemType (A-->B))) (SemEnv E)) (SemType (A-->B))). intros.
  apply (@fcont_le_intro (SemType A) (DL (SemType B))). intros.
  unfold sem_fix_chain_core at 1.
  repeat rewrite K_simpl. apply (@DL_bot_least (SemType B)).

  intros.
  do 2 rewrite sem_fix_chain_core_unfold.
  apply (@fcont_le_intro (Dprod ((SemEnv E) -C-> (SemType (A-->B)) -C-> (SemType (A-->B))) (SemEnv E)) (SemType (A-->B))). intros.
  repeat rewrite fcont_comp_simpl.
  apply fmonotonic.
  apply fcont_le_elim.
  eapply Ole_trans. apply (PROD_FUN_simpl EV (sem_fix_chain_core E A B n)).
  eapply Ole_trans. Focus 2. apply (proj1 (PROD_FUN_simpl EV (sem_fix_chain_core E A B (S n)))).
  rewrite IHn. auto.

assert (forall x y, (x <= y)%nat -> sem_fix_chain_core E A B x <= sem_fix_chain_core E A B y).
intros. induction H0. auto. eapply Ole_trans. apply IHle. apply H. clear H.

red; intros.
unfold sem_fix_chain_cc. fold SemType.
rewrite (H0 x y H). auto.
Defined.

Lemma sem_fix_chain_simpl: forall E A B F n,
  sem_fix_chain F n == sem_fix_chain_cc E A B n F.
auto.
Qed.

Lemma sem_fix_chain_zero: forall E A B F,
  @sem_fix_chain E A B F 0 == @K _ (_-C->_) (@K _ (DL _) (DL_bot _)).
fold SemType. intros.
apply (@fcont_eq_intro (SemEnv E) (SemType (A-->B))). intros.
apply (@fcont_eq_intro (SemType A) (DL (SemType B))). intros.
auto.
Qed.

Lemma sem_fix_chain_unfold: forall E A B F n,
  @sem_fix_chain E A B F (S n) == @EV _ _ @_ (PROD_fun (curry (curry F)) (sem_fix_chain F n)).
intros.
apply (@fcont_eq_intro (SemEnv E) (SemType (A-->B))). intros.
apply (@fcont_eq_intro (SemType A) (DL (SemType B))). intros.
repeat rewrite sem_fix_chain_simpl.
unfold sem_fix_chain_cc. fold SemType.
rewrite sem_fix_chain_core_unfold.
auto.
Qed.

Lemma sem_fix_chain_lub: forall E A B(F: SemCTypeL (A::A-->B::E) B),
  lub (sem_fix_chain F) == sem_fix _ _ _ F.

intros. split.

apply lub_le.
intros. induction n.
  apply (@fcont_le_intro (SemEnv E) (SemType (A-->B))). intros.
  apply (@fcont_le_intro (SemType A) (DL (SemType B))). intros.
  rewrite sem_fix_chain_zero. repeat rewrite K_simpl. fold SemType.
  apply (@DL_bot_least (SemType B)).

  apply (@fcont_le_intro (SemEnv E) (SemType (A-->B))); intros.
  apply (@fcont_le_intro (SemType A) (DL(SemType B))); intros.
  rewrite sem_fix_chain_unfold. 
  rewrite sem_fix_unfold.
  
  rewrite fcont_comp_simpl, PROD_fun_simpl, PAIR_simpl, Pair_simpl. unfold fst, snd.
  unfold EV. rewrite fcont_comp2_simpl.
  rewrite FST_simpl, Fst_simpl, SND_simpl, Snd_simpl. unfold fst, snd.
  rewrite AP_simpl. repeat rewrite curry_simpl.
  apply fmonotonic. split. unfold fst. split. unfold fst. auto.
  unfold snd. auto. unfold snd. auto.

apply (@fcont_le_intro (SemEnv E) (SemType (A-->B))); intros.
apply (@fcont_le_intro (SemType A) (DL(SemType B))); intros.
unfold sem_fix. fold SemType SemEnv.

rewrite curry_simpl, fcont_comp_simpl, uncurry_simpl, fcont_comp_simpl. repeat rewrite CURRY_simpl. 
apply fcont_le_elim.
apply (@FIXP_inv _ _ (Curry (Curry F) x)).
apply (@fcont_le_intro (SemType A) (DL (SemType B))); intros.
repeat rewrite Curry_simpl.
pattern lub at 1. rewrite fcont_lub_simpl.
pattern lub at 1. rewrite fcont_lub_simpl.
rewrite (lub_succ_eq ((sem_fix_chain F <__> x) <__> x1)). fold SemType.

repeat rewrite <- Pair_simpl.
repeat rewrite <- PAIR_simpl.
rewrite fcont_lub_simpl.
repeat rewrite lub_comp_eq_cont. fold SemEnv.
rewrite fcont_lub_simpl, lub_comp_eq_cont.
apply lub_le_compat.
apply fmon_le_intro. intros.
repeat rewrite fmon_comp_simpl.
repeat rewrite fcontit_simpl.
apply fmonotonic. auto.
Qed.

Lemma sem_app_eval: forall E A B f a,
  sem_app E A B f a ==
  (kleisli_cc _ _ (@EV _ _)) @_ (uncurry _ _ _ (@Smash _ _)) @_ PROD_fun f a.
intros.
apply (@fcont_eq_intro E (DL B)); intro.
unfold sem_app.
repeat rewrite curry_simpl.
repeat rewrite fcont_comp_simpl.
repeat rewrite uncurry_simpl.
auto.
Qed.

Lemma sem_app_val_simpl: forall E A B (f:SemCTypeT E (A-->B)) (x:SemCTypeT E A),
  sem_app _ _ _ (eta @_ f) (eta @_ x) == EV @_ PROD_fun f x.

fold SemType; intros.
apply (@fcont_eq_intro (SemEnv E) (DL (SemType B))); intro.
unfold sem_app.
repeat rewrite curry_simpl.
repeat rewrite fcont_comp_simpl.
repeat rewrite uncurry_simpl.
rewrite PROD_FUN_simpl.
repeat rewrite PROD_fun_simpl, PAIR_simpl, Pair_simpl.
rewrite uncurry_simpl.
repeat rewrite fcont_comp_simpl, eta_val.
unfold Smash; rewrite Operator2_simpl.
rewrite kleisli_cc_simpl, kleisli_inv.
auto.
Qed.

Lemma sem_app_curry_simpl: forall E A B (f:SemCTypeL (A::E) B) (x:SemCTypeL E A),
  sem_app _ _ _ (sem_op _ eta (sem_curry _ _ _ f)) x == KLEISLIR f @_ PROD_fun ID x.
fold SemEnv; intros.
apply (@fcont_eq_intro (SemEnv E) (DL (SemType B))); intro.

unfold sem_app, sem_curry, KLEISLIR, LStrength.
repeat rewrite curry_simpl.
repeat rewrite fcont_comp_simpl.
repeat rewrite uncurry_simpl.
repeat rewrite PROD_FUN_simpl.
repeat rewrite PROD_fun_simpl.
repeat rewrite PAIR_simpl, Pair_simpl.
repeat rewrite uncurry_simpl.
unfold sem_op.
repeat rewrite fcont_comp_simpl.
repeat rewrite PROD_map_simpl. simpl.
repeat rewrite uncurry_simpl.
repeat rewrite ID_simpl, Id_simpl.
repeat rewrite eta_val.
rewrite (@curry_com _ _ _ f) at 2.
rewrite kleisli_cc_comp2. fold SemEnv.
rewrite fcont_comp_simpl.
rewrite Smash_natural.
rewrite kleisli_cc_unit2.
setoid_rewrite kleisli_cc_simpl at 3.
setoid_rewrite kleisli_inv.
repeat rewrite fcont_comp_simpl.
repeat rewrite eta_val.
rewrite ID_simpl, Id_simpl.
rewrite CURRY_simpl.
auto.
Qed.

Lemma sem_app_simpl: forall E A B (df:E-C->DL(A-C->DL B)) (dx:E-C->DL A) de f x,
  Val f == (df de) -> Val x == (dx de) -> f x == sem_app _ _ _ df dx de.

intros.
unfold sem_app.
repeat rewrite curry_simpl.
repeat rewrite fcont_comp_simpl.
repeat rewrite uncurry_simpl.
rewrite PROD_FUN_simpl, PROD_fun_simpl, PAIR_simpl, Pair_simpl, uncurry_simpl.
rewrite <- H. rewrite <- H0.
rewrite kleisli_cc_simpl, kleisli_inv.
auto.
Qed.

Lemma sem_app_simpl_le: forall E A B (df:E-C->DL(A-C->DL B)) (dx:E-C->DL A) de f x,
  Val f <= (df de) -> Val x <= (dx de) -> f x <= sem_app _ _ _ df dx de.

intros.

destruct (DLvalval2 H).
destruct (DLvalval2 H0).
destruct H1.
destruct H2.

apply Ole_trans with (f x1).
apply fcont_monotonic. auto.
apply Ole_trans with (x0 x1).
apply H3.
apply (sem_app_simpl (Oeq_sym H1) (Oeq_sym H2)).
Qed.

Lemma sem_app_inv: forall E A B fc xc e b,
  (Val b == sem_app E A B fc xc e) ->
  exists f, exists x, fc e == Val f /\ xc e == Val x /\ Val b == f x.

intros.
unfold sem_app in H.
repeat rewrite curry_simpl in H.
rewrite fcont_comp_simpl in H.
repeat rewrite uncurry_simpl in H.
rewrite PROD_FUN_simpl, fcont_comp_simpl, PROD_fun_simpl, PAIR_simpl, Pair_simpl, uncurry_simpl, kleisli_cc_simpl in H.
hreplace H with (@kleisliValVal _ _ (Smash (A -C-> DL B) A (fc e) (xc e)) EV b (Oeq_sym H)).
destruct H.
destruct x.
destruct H.
unfold Smash in H.
hreplace H with (Operator2Val H).
destruct H.
destruct H.
exists x.
exists x0.
destruct H.
destruct H1.
rewrite eta_val in H2.
split. assumption.
split. assumption.
hreplace H2 with (DLleVal_eq2 H2).
assert (fst (x,x0) == fst (t,t0)). auto. simpl in H3.
assert (snd (x,x0) == snd (t,t0)). auto. simpl in H4.
rewrite H4.
assert (x t0 == t t0). auto. rewrite H5.
auto.
Qed.



(******************
   Chain
 ******************)

(*
   sem_precomp
*)
Definition sem_precomp_cha: forall E E' C (gs:natO-m>(E-C->C)) (f:E'-C->E), natO-m>(E'-C->C).
intros.
exists (fun n => (gs n) @_ f).
red; intros.
apply fcont_le_intro.
intros.
repeat rewrite fcont_comp_simpl.
apply (fmonotonic gs).
assumption.
Defined.

Lemma sem_precomp_cha_eval: forall E E' C gs f i e,
  @sem_precomp_cha E E' C gs f i e == gs i (f e).
auto.
Qed.

Lemma sem_precomp_cha_lub: forall E E' C gs f,
  lub (@sem_precomp_cha E E' C gs f) == (lub gs) @_ f.

intros.
apply (@fcont_eq_intro _ _ (lub (sem_precomp_cha gs f)) (lub gs @_ f)); intros.
rewrite fcont_comp_simpl.
repeat rewrite fcont_lub_simpl.
apply lub_eq_compat.
apply fmon_eq_intro; intros.
simpl.
auto.
Qed.

(*
   sem_op
*)

Definition sem_op_cha: forall E A B (op:A-C->B) (vs:natO-m>(E-C->A)), natO-m>(E-C->B).
intros.
exists (fun n => sem_op _ op (vs n)).
red; intros.
rewrite H. auto.
Defined.

Lemma sem_op_cha_lub: forall E A B op vs,
  lub (@sem_op_cha E A B op vs) == sem_op _ op (lub vs).

intros.
apply (@fcont_eq_intro _ _ (lub (sem_op_cha op vs)) (sem_op _ op (lub vs))); intros.
rewrite sem_op_simpl. rewrite fcont_comp_simpl.
repeat rewrite fcont_lub_simpl.
apply Oeq_trans with (lub ((fcontit op) @ (vs <__> x))).
apply lub_eq_compat.
apply fmon_eq_intro; intros.
simpl.
auto.
apply Oeq_sym.
apply lub_comp_eq_cont.
Qed.

Lemma sem_op_cha_eval: forall E A B op vs i e,
  @sem_op_cha E A B op vs i e == op (vs i e).
auto.
Qed.

Lemma sem_op_cha_simpl: forall E A B (op:A-C->B) (chs:natO-m>(E-C->A)) i,
  sem_op_cha op chs i == sem_op _ op (chs i).
intros. apply (@fcont_eq_intro E B). intros. apply sem_op_cha_eval.
Qed.

Lemma sem_op_closed: forall E A B (op:A-C->B) (P: prdc (E-C->A)) (Q: prdc (E-C->B)),
  chain_closed Q -> down_closed Q -> (forall p, P p -> Q (sem_op _ op p)) ->
  forall p, ideal_closure P p -> Q (sem_op _ op p).
intros. apply H2. red; intros. apply H1. auto.
red; intros. eapply H0. apply (sem_op_cha_lub op chs).
apply H. intros. eapply H0. apply (sem_op_cha_simpl op chs i). simpl. apply H3.
red; intros. eapply H0. unfold sem_op. rewrite H3. apply Ole_refl. auto.
Qed.

(*
   sem_op2
*)

Definition sem_op2_cha E A B C (op:A-C->B-C->C) (vs1:natO-m>E-C->A) (vs2:natO-m>E-C->B)
 : natO-m> E-C->C
 := chain2_fun (sem_op2 _ op) vs1 vs2.

Lemma sem_op2_cha_lub: forall E A B C op vs1 vs2,
  lub (@sem_op2_cha E A B C op vs1 vs2) == sem_op2 _ op (lub vs1) (lub vs2).
intros.
apply chain2_fun_lub.
Qed.

Lemma sem_op2_cha_eval: forall E A B C op vs1 vs2 i e,
  @sem_op2_cha E A B C op vs1 vs2 i e == op (vs1 i e) (vs2 i e).
auto.
Qed.

Lemma sem_op2_cha_simpl: forall E A B C op vs1 vs2 i,
  @sem_op2_cha E A B C op vs1 vs2 i == sem_op2 _ op (vs1 i) (vs2 i).
intros. apply (@fcont_eq_intro E C). apply sem_op2_cha_eval.
Qed.

Lemma sem_op2_closed: forall E A B C (op:A-C->B-C->C) (P: prdc (E-C->A)) (Q: prdc (E-C->B)) (R: prdc (E-C->C)),
  chain_closed R -> down_closed R -> (forall p q, P p -> Q q -> R (sem_op2 _ op p q)) ->
  forall p q, ideal_closure P p -> ideal_closure Q q -> R (sem_op2 _ op p q).
intros. generalize q H3; clear H3 q. apply H2. 
  red; intros. apply H4.
    red; intros. apply H1. auto. auto.

    red; intros. eapply H0. rewrite <- (lub_cte a). apply (sem_op2_cha_lub op _ chs).
    apply H. intros. eapply H0. rewrite sem_op2_cha_simpl at 1. rewrite fmon_cte_simpl.
    apply Ole_refl. apply H5.

    red; intros. eapply H0. rewrite H5. apply Ole_refl. auto.
  
  red; intros. eapply H0. rewrite <- (lub_cte q). apply (sem_op2_cha_lub op chs _).
  apply H. intros. eapply H0. rewrite sem_op2_cha_simpl at 1. rewrite fmon_cte_simpl.
  apply Ole_refl. apply H3. auto.

  red; intros. eapply H0. rewrite H3. apply Ole_refl. auto.
Qed.

(*
   sem_op3
*)

Definition sem_op3_cha E A B C D (op:A-C->B-C->C-C->D) (vs1:natO-m>E-C->A) (vs2:natO-m>E-C->B) (vs3:natO-m>E-C->C)
  : natO-m>E-C->D
  := chain3_fun (sem_op3 _ op) vs1 vs2 vs3.

Lemma sem_op3_cha_lub: forall E A B C D op vs1 vs2 vs3,
  lub (@sem_op3_cha E A B C D op vs1 vs2 vs3) == sem_op3 _ op (lub vs1) (lub vs2) (lub vs3).

intros.
apply chain3_fun_lub.
Qed.

Lemma sem_op3_cha_eval: forall E A B C D op vs1 vs2 vs3 i e,
  @sem_op3_cha E A B C D op vs1 vs2 vs3 i e == op (vs1 i e) (vs2 i e) (vs3 i e).
auto.
Qed.

Lemma sem_op3_cha_simpl: forall E A B C D op vs1 vs2 vs3 i,
  @sem_op3_cha E A B C D op vs1 vs2 vs3 i == sem_op3 _ op (vs1 i) (vs2 i) (vs3 i).
intros. eapply (@fcont_eq_intro E D). apply sem_op3_cha_eval.
Qed.

Lemma sem_op3_closed: forall E A B C D (op:A-C->B-C->C-C->D) (P: prdc (E-C->A)) (Q: prdc (E-C->B)) (R: prdc (E-C->C)) (S: prdc (E-C->D)),
  chain_closed S -> down_closed S -> (forall p q r, P p -> Q q -> R r -> S (sem_op3 _ op p q r)) ->
  forall p q r, ideal_closure P p -> ideal_closure Q q -> ideal_closure R r -> S (sem_op3 _ op p q r).
intros. generalize q H3 r H4; clear H4 H3 q r. apply H2. 
  red; intros. generalize r H5; clear H5 r. apply H4.
    red; intros. apply H6. 
      red; intros. apply H1. auto. auto. auto.

      red; intros. eapply H0. rewrite <- (lub_cte a), <- (lub_cte a0). apply (sem_op3_cha_lub op _ _ chs).
      apply H. intros. eapply H0. rewrite sem_op3_cha_simpl at 1. repeat rewrite fmon_cte_simpl.
      apply Ole_refl. apply H7.

      red; intros. eapply H0. rewrite H7. apply Ole_refl. auto.

    red; intros. eapply H0. rewrite <- (lub_cte a), <- (lub_cte r). apply (sem_op3_cha_lub op _ chs _).
    apply H. intros. eapply H0. rewrite sem_op3_cha_simpl at 1. repeat rewrite fmon_cte_simpl.
    apply Ole_refl. apply H5. auto.

    red; intros. eapply H0. rewrite H5. apply Ole_refl. auto.

  red; intros. eapply H0. rewrite <- (lub_cte q), <- (lub_cte r). apply (sem_op3_cha_lub op chs _ _).
  apply H. intros. eapply H0. rewrite sem_op3_cha_simpl at 1. repeat rewrite fmon_cte_simpl.
  apply Ole_refl. apply H3. auto. auto.

  red; intros. eapply H0. rewrite H3. apply Ole_refl. auto.
Qed.

(*
   sem_curry
*)

Definition sem_curry_cha E A B (dcs:natO-m>(Dprod E A)-C->B) : natO -m> E -C-> A -C-> B :=
  fcontit (sem_curry _ _ _) @ dcs.

Lemma sem_curry_cha_eval: forall E A B dcs i,
  @sem_curry_cha E A B dcs i == sem_curry _ _ _ (dcs i).
intros.
auto.
Qed.

Lemma sem_curry_cha_lub: forall E A B dcs,
  lub (@sem_curry_cha E A B dcs) == sem_curry _ _ _ (lub dcs).
intros.
unfold sem_curry_cha.
rewrite <- lub_comp_eq_cont.
auto.
Qed.

Lemma sem_curry_closed: forall E A B (P:prdc (Dprod E A -C-> B)) (Q: prdc (E-C->(A-C->B))),
  chain_closed Q -> down_closed Q -> (forall p, P p -> Q (sem_curry _ _ _  p)) ->
  forall p, ideal_closure P p -> Q (sem_curry _ _ _ p).
intros. apply H2. red; intros. apply H1. auto.
red; intros. eapply H0. apply (sem_curry_cha_lub chs).
apply H. intros. eapply H0. apply (proj1 (sem_curry_cha_eval chs i)). apply H3.
red; intros. eapply H0. rewrite H3. apply Ole_refl. auto.
Qed.

(*
   sem_fix
*)

Fixpoint sem_fix_cha_u E A B (fs: natO-M->SemCTypeL (A::(A-->B)::E) B) (n:nat)
  {struct n} : SemEnv E -C-> SemType (A --> B)
 :=
    match n return (SemEnv E -C-> SemType (A --> B)) with
      | 0 => @K _ (_ -C-> _) (@K _ (DL _) (DL_bot _))
      | S m => @EV _ _ @_ (PROD_fun (curry (curry (fs m))) (sem_fix_cha_u fs m))
    end.

Lemma sem_fix_cha_u_monotonic: forall E A B fs,
  @monotonic natO _ (@sem_fix_cha_u E A B fs).

red; intros.

assert (forall z x y, (y <= z)%nat -> (x <= y)%nat ->
  sem_fix_cha_u (E:=E) (A:=A) (B:=B) fs x
  <= sem_fix_cha_u (E:=E) (A:=A) (B:=B) fs y).
induction z.
intros.
inversion H0.
rewrite H2 in H1.
inversion H1.
auto.

intros.
induction H1.
auto.

destruct x0.
apply (@fcont_le_intro (SemEnv E) (SemType (A-->B))).
intros.
unfold sem_fix_cha_u at 1; fold SemEnv SemType.
rewrite K_simpl.
apply (@fcont_le_intro _ _ (@K _ (DL _) (DL_bot _)) (sem_fix_cha_u fs (S m) x0)).
fold SemEnv SemType; intros.
rewrite K_simpl.
apply (DL_bot_least (sem_fix_cha_u fs (S m) x0 x1)).

unfold sem_fix_cha_u; fold sem_fix_cha_u.
apply (@fcont_comp_le_compat 
  (SemEnv E)
  (Dprod (SemType (A --> B) -C-> SemType A -C-> DL (SemType B)) (SemType (A --> B)))
  (SemType (A-->B))
).
auto.
apply PROD_fun_mon.

assert (Curry (Curry (fs x0)) <= Curry (Curry (fs m))).
apply fmonotonic.
apply fmonotonic.
apply fmonotonic.
simpl; omega.
apply H2.

eapply Ole_trans.
Focus 2.
apply IHle.
omega.
apply IHz.
omega.
omega.

eapply H0.
instantiate (1:= y).
omega.
simpl in H.
omega.
Qed.

Definition sem_fix_cha: forall E A B (fs: natO-m>SemCTypeL (A::A-->B::E) B),
  natO-m> SemCTypeT E (A-->B).
intros.
exists (sem_fix_cha_u fs).
apply (sem_fix_cha_u_monotonic fs).
Defined.

Lemma sem_fix_cha_ind_0: forall E A B fs e,
  @sem_fix_cha E A B fs 0 e == @K _ (DL _) (DL_bot _).
auto.
Qed.

Lemma sem_fix_cha_ind_S: forall E A B fs n e,
  @sem_fix_cha E A B fs (S n) e == (curry (curry (fs n)) e) (sem_fix_cha fs n e).
auto.
Qed.

Lemma sem_fix_cha_lub: forall E A B fs,
  lub (@sem_fix_cha E A B fs) == sem_fix _ _ _ (lub fs).
intros.
split.

apply lub_le.
intros.
induction n.
apply (@fcont_le_intro (SemEnv E) (SemType (A-->B))); intros.
rewrite sem_fix_cha_ind_0. fold SemType.
apply (@fcont_le_intro (SemType A) (DL(SemType B))); intros.
rewrite K_simpl.
apply (DL_bot_least (sem_fix _ _ _ (lub fs) x x0)).

apply (@fcont_le_intro (SemEnv E) (SemType (A-->B))); intros.
rewrite sem_fix_cha_ind_S.
apply (@fcont_le_intro (SemType A) (DL(SemType B))); intros.
repeat rewrite curry_simpl.
rewrite sem_fix_unfold.

apply Ole_trans with (lub fs (x, sem_fix_cha fs n x, x0)).
apply fcont_le_elim.
apply (le_lub fs n).

apply fmonotonic.
split. split. auto.
Focus 2. auto.
simpl snd.
apply IHn.

apply (@fcont_le_intro (SemEnv E) (SemType (A-->B))); intros.
apply (@fcont_le_intro (SemType A) (DL(SemType B))); intros.
unfold sem_fix. fold SemType SemEnv.

rewrite curry_simpl, fcont_comp_simpl, uncurry_simpl, fcont_comp_simpl.
repeat rewrite CURRY_simpl.
apply fcont_le_elim.
apply (@FIXP_inv _ _ (Curry (Curry (lub fs)) x)).
apply (@fcont_le_intro (SemType A) (DL (SemType B))); intros.
repeat rewrite Curry_simpl.
pattern lub at 1.
rewrite fcont_lub_simpl.

repeat rewrite <- Pair_simpl.
repeat rewrite <- PAIR_simpl.
repeat rewrite lub_comp_eq_cont. fold SemEnv.
pattern lub at 1.
rewrite fcont_lub_simpl.
rewrite lub_comp_simpl.
rewrite fcont_lub_simpl.
rewrite (lub_succ_eq ((sem_fix_cha fs <__> x) <__> x1)).
fold SemType.
apply lub_le_compat.

apply fmon_le_intro; intros.
rewrite fmon_comp_simpl.
rewrite fcont_app_simpl.
rewrite (@fcont_app_simpl natO (SemEnv E) (SemType (A-->B))).
rewrite fmon_comp2_simpl.
rewrite fcont_app_simpl.
repeat rewrite fmon_comp_simpl.
rewrite fcont_app_simpl.
rewrite Fcontit_simpl.
auto.
Qed.

(*
   sem_app
*)

Definition sem_app_cha E A B (fs: natO-m>E-C->DL(A-C->DL B)) (xs: natO-m>E-C->DL A):
  natO-m>E-C->DL B
  := chain2_fun (sem_app _ _ _) fs xs.

Lemma sem_app_cha_lub: forall E A B fs xs,
  lub (@sem_app_cha E A B fs xs) == sem_app _ _ _ (lub fs) (lub xs).
intros.
apply chain2_fun_lub.
Qed.

Lemma sem_app_cha_eval: forall E A B fs xs i,
  sem_app_cha fs xs i == sem_app E A B (fs i) (xs i).
auto.
Qed.

Lemma sem_app_cha_inv: forall E A B fs xs i e b,
  (Val b == @sem_app_cha E A B fs xs i e) ->
  exists f, exists a, fs i e == Val f /\ xs i e == Val a /\ Val b == f a.
intros.
rewrite sem_app_cha_eval in H.
apply sem_app_inv.
auto.
Qed.

Lemma sem_app_cha_inv_le: forall E A B fs xs i e b,
  (Val b <= @sem_app_cha E A B fs xs i e) ->
  exists f, exists a, fs i e == Val f /\ xs i e == Val a /\ Val b <= f a.

intros.

destruct (DLvalval2 H).
destruct H0.

destruct (sem_app_cha_inv (Oeq_sym H0)).
do 2 destruct H2.
destruct H3.

exists x0.
exists x1.
split. auto.
split. auto.

apply Ole_trans with (Val x).
auto.
auto.
Qed.

Lemma sem_app_closed: forall E A B (P:prdc (E-C->DL (A-C->DL B))) (Q:prdc (E-C->DL A)) (R: prdc (E-C->DL B)),
  chain_closed R -> down_closed R -> (forall p q, P p -> Q q -> R (sem_app _ _ _ p q)) ->
  forall p q, ideal_closure P p -> ideal_closure Q q -> R (sem_app _ _ _ p q).

intros. generalize q H3. clear H3 q. apply H2. 
  red; intros. apply H4. 
    red; intros. apply H1. auto. auto.

    red; intros. eapply H0. 
    rewrite <- (lub_cte a). apply (sem_app_cha_lub (fmon_cte _ a) chs).
    apply H. intros. eapply H0. 
    rewrite sem_app_cha_eval. rewrite fmon_cte_simpl. apply Ole_refl. apply H5.

    red; intros. eapply H0. rewrite H5. apply Ole_refl. auto.

  red; intros. eapply H0.
  rewrite <- (lub_cte q). apply (sem_app_cha_lub chs (fmon_cte _ q)).
  apply H. intros. eapply H0.
  rewrite sem_app_cha_eval. rewrite fmon_cte_simpl. apply Ole_refl. apply H3. auto.

  red; intros. eapply H0. rewrite H3. apply Ole_refl. auto.
Qed.


(* End Domain_Operations. *)



