(****
   Low level machine & logical relations for
   Biorthogonality, Step-Indexing & Compiler Correctness
   Nick Benton & Chung-Kil Hur
   April 2009
 ****)

Require Import FunctionalExtensionality.
Require Import utility.
Require Import Predom.
Require Import PredomAux.
Require Import typedlambda.
Require Import typeddensem.
Require Import domoprs.
Set Implicit Arguments.
Unset Strict Implicit.

Open Scope nat_scope.
Open Scope type_scope.

(****
   Definition of CESD machine
 ****)

Inductive Instruction : Set :=
 (* General stack manipulation *)
 | Swap  : 
   Instruction
 | Dup   : 
   Instruction
 | Pop   : 
   Instruction
 | PushV : 
   nat -> Instruction

 (* Integer operations *)
 | PushN : 
   nat -> Instruction
 | Op    : 
   Operation -> Instruction

 (* Function operations *)
 | PushC  : 
   list Instruction -> Instruction
 | PushRC : 
   list Instruction -> Instruction
 | App    : 
   Instruction
 | Ret    : 
   Instruction

 (* Branch operations *) 
 | Sel   : 
   list Instruction -> list Instruction -> Instruction
 | Join  : 
   Instruction

 (* Pair operations *)
 | MkPair  : 
   Instruction
 | ProjFst : 
   Instruction
 | ProjSnd : 
   Instruction

 (* Syntactic equality test operations *)
 | EqTest  : 
   Instruction 

 (* Virtual operations *)
 | EqHalt  : 
   Instruction
 | NeqHalt : 
   Instruction
 | Halt    : 
   Instruction
 | DoNothing :
   Instruction
.

Inductive ValueM : Set :=
 | VN : nat -> ValueM
 | VC : list ValueM -> (list Instruction) -> ValueM
 | VRC: list ValueM -> (list Instruction) -> ValueM
 | VP : ValueM -> ValueM -> ValueM
.

Definition Code := list Instruction.

Definition EnvM := list ValueM.

Definition Stack := list ValueM.

Definition Dump := (list (Code * EnvM * Stack)) % type. 

Definition CESD := Code * EnvM * Stack * Dump.

Definition CompM := Code * Stack.

Definition VToC (v:ValueM) : CompM  := (nil,v::nil).

Definition CC (c:Code) : CompM := (c,nil).

Definition GetEnv (cesd:CESD) : EnvM := match cesd with (_,e,_,_) => e end.

Definition RemoveEnv (cesd:CESD) : CESD := match cesd with (c,_,s,d) => (c,nil,s,d) end.

Definition mergeEnv (e:EnvM) (cesd:CESD) : CESD := match cesd with (c0,e0,s0,d0) => (c0,e++e0,s0,d0) end.

Definition mergeComp (comp:CompM) (cesd:CESD) : CESD := 
  match comp with (c,s) => 
    match cesd with (c0,e0,s0,d0) =>
      (c++c0,e0,s++s0,d0) 
    end
  end.

Lemma mergeEnv_assoc: forall e1 e2 cesd, mergeEnv (e1++e2) cesd = mergeEnv e1 (mergeEnv e2 cesd).
intros.
destruct cesd; destruct p; destruct p.
unfold mergeEnv; rewrite app_ass; simpl.
auto.
Qed.

Lemma mergeEnv_prj_rm: forall cesd,
  mergeEnv (GetEnv cesd) (RemoveEnv cesd) = cesd.

intros.
destruct cesd; destruct p; destruct p.
simpl.
rewrite <- app_nil_end.
auto.
Qed.




(****
   Operational Semantics for CESD
 ****)

Inductive step : CESD -> CESD -> Prop :=
 (* General stack manipulation *)
 | step_swap : forall c e s d v1 v2, step (Swap ::c, e, v1::v2::s, d) (c, e, v2::v1::s, d)
 | setp_dup : forall c e s d v, step (Dup ::c, e, v::s, d) (c, e, v::v::s, d)
 | step_pop : forall c e s d v, step (Pop ::c, e, v::s, d) (c, e, s, d)
 | step_pushv : forall c e s d i v, nth_error e i = value v -> step (PushV  i :: c, e, s, d) (c, e, v::s, d)

 (* Integer operations *)
 | step_pushn : forall  c e s d n, step (PushN  n :: c, e, s, d) (c, e, VN n :: s, d)
 | step_op : forall  c e s d op v1 v2, step (Op  op::c,e, (VN v2) ::(VN v1) ::s, d) (c,e,VN (OpSem op v1 v2)::s,d)

 (* Function operations *)
 | step_pushc : forall c e s d body, step (PushC  body :: c, e, s, d) (c, e, VC e body :: s, d)
 | step_pushrc : forall c e s d body, step (PushRC  body :: c, e, s, d) (c, e, VRC e body :: s, d)
 | step_app : forall c e s d e' body v, step (App :: c, e, v :: VC e' body :: s, d) (body, v::e', nil, (c,e,s)::d)
 | step_rapp : forall c e s d e' body v, step (App :: c, e, v :: VRC e' body :: s, d) (body, v:: (VRC e' body) :: e', nil, (c,e,s)::d)
 | step_ret : forall c e s d v c' e' s', step (Ret ::c, e, v::s, (c',e',s')::d) (c',e',v::s',d)

 (* Branch operations *) 
 | step_seltrue : forall c e s d v c1 c2, v <> VN 0 -> step (Sel  c1 c2::c, e, v::s,d) (c1,e,s,(c,nil,nil)::d)
 | step_selfalse : forall c e s d c1 c2, step (Sel  c1 c2::c, e, VN 0::s,d) (c2,e,s,(c,nil,nil)::d)
 | step_join : forall c e s d c' e' s', step (Join ::c, e, s, (c',e',s')::d) (c', e, s, d)

 (* Pair operations *)
 | step_mkpair : forall c e s d v1 v2, step (MkPair ::c, e, v2::v1::s, d) (c, e, VP v1 v2::s, d)
 | step_fst : forall c e s d v1 v2, step (ProjFst ::c, e, VP v1 v2::s, d) (c, e, v1::s, d)
 | step_snd : forall c e s d v1 v2, step (ProjSnd ::c, e, VP v1 v2::s, d) (c, e, v2::s, d)

 (* Syntactic equality test operation *)
 | step_eqtrue : forall c e s d v1 v2, v1=v2 -> step (EqTest ::c,e,v1::v2::s,d) (c,e,VN 1::s,d) 
 | step_eqfalse : forall c e s d v1 v2, v1<>v2 -> step (EqTest ::c,e,v1::v2::s,d) (c,e,VN 0::s,d) 

 (* Virtual operations *)
 | step_eqhalt : forall c e s d v1 v2, v1<>v2 -> step (EqHalt ::c,e,v1::v2::s,d) (c,e,s,d) 
 | step_neqhalt : forall c e s d v1 v2, v1=v2 -> step (NeqHalt ::c,e,v1::v2::s,d) (c,e,s,d) 
 | step_donothing : forall c e s d, step (DoNothing ::c,e,s,d) (c,e,s,d)
.

Inductive kstepreduce: nat -> CESD -> CESD -> Prop :=
  | kstep0 : forall cesd, kstepreduce 0 cesd cesd
  | kstepS : forall k cesd cesd1 cesd', step cesd cesd1 -> kstepreduce k cesd1 cesd' -> kstepreduce (S k) cesd cesd'
.

Definition stepsfor k cesd := exists k', exists cesd', k' >= k /\ kstepreduce k' cesd cesd'.

Definition diverge cesd := forall k, stepsfor k cesd.

Definition terminate cesd := ~ (diverge cesd).

Inductive reduceSet: CESD -> CESD -> Set :=
| reduce0 : forall cesd, reduceSet cesd cesd
| reduceS : forall cesd cesd1 cesd',  
  step cesd cesd1 -> reduceSet cesd1 cesd' -> reduceSet cesd cesd'.

Fixpoint count_steps cesdfrom cesdto (pf: reduceSet cesdfrom cesdto) {struct pf} : nat :=
  match pf with
    | reduce0 _ => 0
    | reduceS cesd cesd1 cesd' _ pf_ind => S (count_steps pf_ind)
  end.

(*
   Properties of kstepreduce
*)

Lemma step_uniq: forall cesd cesd' cesd'', step cesd cesd' -> step cesd cesd'' -> cesd' = cesd''.
intros.
destruct H.

inversion H0; auto.
inversion H0; auto.
inversion H0; auto.

inversion H0.
rewrite H in H7. 
inversion H7. auto.

inversion H0; auto.
inversion H0; auto.

inversion H0; auto.
inversion H0; auto.
inversion H0; auto.
inversion H0; auto.
inversion H0; auto.

inversion H0; auto; absurd False; auto.

inversion H0. absurd False; auto. auto.

inversion H0; auto.

inversion H0; auto.
inversion H0; auto.
inversion H0; auto.

inversion H0; auto; absurd False; auto.
inversion H0; auto; absurd False; auto.

inversion H0; auto.
inversion H0; auto.
inversion H0; auto.
Qed.

Lemma kstepreduce_uniq: forall k cesd cesd' cesd'', kstepreduce k cesd cesd' -> kstepreduce k cesd cesd'' -> (cesd' = cesd'').

assert (forall gk k cesd cesd' cesd'', k <= gk -> kstepreduce k cesd cesd' -> kstepreduce k cesd cesd'' -> (cesd' = cesd'')).
induction gk.

intros. inversion H. rewrite H2 in H0, H1.
inversion H0. inversion H1.
rewrite <- H5, H7. auto.

intros.
inversion H0.
rewrite <- H2 in H1. inversion H1.
rewrite <- H4, H7. auto.
inversion H1. rewrite <- H7 in H4. inversion H4.
assert (H12 := step_uniq H2 H7).
assert (k0 = k1). omega.
eapply IHgk.
instantiate (1:= k0). omega.
apply H3.
rewrite H13. rewrite H12. assumption.

intros. eapply (H k k).
omega. apply H0. apply H1.
Qed.

Lemma kstepreduce_next: forall k cesd cesd' cesd'', 
  kstepreduce k cesd cesd' -> step cesd' cesd'' -> kstepreduce (S k) cesd cesd''.

assert (forall gk k cesd cesd' cesd'', k <= gk ->
  kstepreduce k cesd cesd' -> step cesd' cesd'' -> kstepreduce (S k) cesd cesd'').

induction gk.
intros. inversion H. rewrite H2 in H0. inversion H0.
eapply kstepS. apply H1. apply kstep0. 

intros. inversion H0. rewrite <- H2 in H0. inversion H0.
eapply kstepS. apply H1. apply kstep0.

assert (S (S k0) = S (S k0))%nat. omega.
eapply kstepS. apply H2.
eapply IHgk. omega. apply H3. apply H1.

intros. eapply (H k k). omega. apply H0. assumption.
Qed.

Lemma kstepreduce_compose: forall k1 k2 cesd cesd1 cesd2,
  kstepreduce k1 cesd cesd1 -> kstepreduce k2 cesd1 cesd2 -> kstepreduce (k1+k2) cesd cesd2.

intros.
induction H.
apply H0.

simpl.
eapply kstepS. apply H.
apply IHkstepreduce.
assumption.
Qed.

Lemma kstepreduce_inv: forall k j cesd cesd' cesd'',
  kstepreduce k cesd cesd' -> kstepreduce (k+j) cesd cesd''-> kstepreduce j cesd' cesd''.

intros. induction H. apply H0.
apply IHkstepreduce.
inversion H0.
assert (H7 := step_uniq H H3).
rewrite H7.
assumption.
Qed.

Lemma stepsfor_zero: forall cesd, stepsfor 0 cesd.
intros.
red. exists 0. exists cesd.
split. auto. apply kstep0.
Qed.

Lemma stepsfor_decr: forall cesd k k', k <= k' -> stepsfor k' cesd -> stepsfor k cesd.
intros.
do 3 destruct H0.
exists x. exists x0. split. omega. assumption.
Qed.

Lemma stepsfor_step_strict: forall cesd cesd' k,
  step cesd cesd' -> stepsfor (k - 1) cesd' -> stepsfor k cesd.
intros.
do 3 destruct H0.
exists (S x)%nat. exists x0.
split. omega.
eapply kstepS. apply H. assumption.
Qed.

Corollary stepsfor_step: forall cesd cesd' k,
  step cesd cesd' -> stepsfor k cesd' -> stepsfor k cesd.
intros. eapply stepsfor_step_strict.
apply H. apply stepsfor_decr with k. omega. assumption.
Qed.

Lemma terminate_step: forall cesd cesd', step cesd cesd' -> terminate cesd' -> terminate cesd.
intros.
repeat red; intros.
apply H0.
red; intros.
hreplace H1 with (H1 (S k)%nat).
do 3 destruct H1.
destruct H2. assert(k = 0). omega. rewrite H2. apply stepsfor_zero.
assert (H4 := step_uniq H H2). destruct H4.
exists k0. exists cesd'0. split. omega.
assumption.
Qed.

Lemma kstepreduce_terminate: forall k cesd cesd',
  kstepreduce k cesd cesd' -> terminate cesd -> terminate cesd'.
intros.
do 2 red; intros. apply H0. red; intros.
destruct (H1 k0) as [k' [cesd'' [H3 H4]]].
exists (k + k')%nat.
exists cesd''.
split. omega.
eapply kstepreduce_compose.
apply H. assumption.
Qed.

Lemma kstepreduce_stepsfor: forall i j k cesd cesd',
  kstepreduce i cesd cesd' -> stepsfor j cesd -> j >= i+k ->stepsfor k cesd'.
intros.
do 3 destruct H0. 
exists (x-i). exists x0. split. omega.
eapply kstepreduce_inv. apply H.
replace (i+(x-i))%nat with x. assumption. omega.
Qed.


(*
   Properties of reduceSet
*)

Definition reduce_compose: forall cesd0 cesd1 cesd2,
  reduceSet cesd0 cesd1 -> reduceSet cesd1 cesd2 -> reduceSet cesd0 cesd2.
intros.
induction H.
exact H0.
eapply reduceS.
apply s.
apply IHreduceSet.
apply H0.
Defined.

Lemma reduce_to_kstepreduce: forall cesd cesd' (pf: reduceSet cesd cesd') k, 
  k = (count_steps pf) -> kstepreduce k cesd cesd'.
induction pf.
intros. simpl in H. rewrite H. apply kstep0.
intros. simpl in H. rewrite H. eapply kstepS.
apply s. apply IHpf. auto.
Qed.

Lemma reduce_terminate: forall cesd cesd',
  reduceSet cesd cesd' -> terminate cesd -> terminate cesd'.

intros. eapply kstepreduce_terminate.
apply reduce_to_kstepreduce with H.
reflexivity. assumption.
Qed.




(****
   Logical Relations
 ****)
Open Scope nat_scope.
Open Scope type_scope.

Definition indexed (A:Type) := nat * A.
Definition iValue := indexed ValueM.
Definition iComp := indexed CompM.
Definition iCESD  := indexed CESD.

Definition idecr A (P:prdc (indexed A)) := forall k k' a, k <= k' -> P (k',a) -> P (k,a).

Definition irel (A B:Type) := nat -> A -> B -> Prop.
Definition rel (A B:Type) := A -> B -> Prop.

Definition ireldecr A B (r: irel A B) := forall k k' a b, k <= k' -> r k' a b -> r k a b.


(**
   << Relation
 **)

(* 
   Top Top semantics 
*)


Definition pton_kstepv (pe:EnvM) (P: prdc iValue) : prdc iCESD  := fun jcesd =>
  match jcesd with (j,(c,e,s,d)) =>
    forall i val, P (i,val) ->
      forall k, k <= i -> k <= j -> stepsfor k (c, pe++e, val::s, d)
      (* stepsfor (min i j) (mergeComp (VToC val) (mergeEnv e cesd)) *)
  end.
  
Definition ntop_kstepc (pe:EnvM) (P: prdc iCESD) : prdc iComp := fun icomp =>
  match icomp with (i,(c1,s1)) =>
    forall j c e s d, P (j,(c,e,s,d)) ->
      forall k, k <= i -> k <= j -> stepsfor k (c1++c, pe++e, s1++s, d)
      (* stepsfor (min i j) (mergeComp comp (mergeEnv e (c,e,s,d))) *)
  end.


(* 
   Definition of << Divergence Monad:
   Given relations between Values and domain elements,
   defines relations between Computations and domain elements.
 *)

Definition irel_to_ipred (A B:Type) (r:irel A B) (b:B) : prdc (indexed A) := fun ia =>
  match ia with (i,a) => r i a b end.

Definition irel_lift (D:cpo) (RD: irel ValueM D) : irel ValueM (DL D) := fun i val dv =>
  exists d, (Val d == dv)%tord /\ RD i val d.

Definition ivalset_le (D:cpo) (RD: irel ValueM D) (d:DL D) : prdc (indexed ValueM) := 
  irel_to_ipred (irel_lift RD) d.

Definition DivMonadLe (S D: cpo) (RS: irel EnvM S) (RD: irel ValueM D) : irel CompM (S -C-> (DL D)) := fun i comp df =>
  forall k e de, k <= i -> RS k e de -> 
    ntop_kstepc e (pton_kstepv e (ivalset_le RD (df de))) (k,comp).

(*
   Approximation relation
*)
Definition irelLeENil : irel EnvM DOne := (fun _ e _ => match e with | nil => True | _ => False end).
Lemma irelLeENil_nil : forall i, irelLeENil i nil tt.
intros; red; auto.
Qed.

(* define in this way to prevent automatic unfolding of irelLeV. *)
Definition irelLeV_body 
  (irelLeV: forall (T:Ty), irel ValueM (SemType T)) 
  (T:Ty) :=
  match T return irel ValueM (SemType T) with
    | Int => fun _ v n => v = VN n
    | Bool => fun _ v b =>
      match v with 
        | VN 0 => b = false
        | VN (S _) => b = true
        | _ => False
      end
    | (T1-->T2) => fun i f df => 
      forall k v dv, k <= i -> irelLeV T1 k v dv -> 
        DivMonadLe irelLeENil (irelLeV T2)
          k (App::nil, v::f::nil) (K DOne (df dv))
    | (T1**T2) => fun k v dv =>
      match (v, dv) with
        | (VP v1 v2, (dv1, dv2)) => irelLeV T1 k v1 dv1 /\ irelLeV T2 k v2 dv2
        | _ => False
      end
  end.
  
Fixpoint irelLeV (T:Ty) : irel ValueM (SemType T) :=
  @irelLeV_body irelLeV T.

Fixpoint irelLeE (E:Env) : irel EnvM (SemEnv E) :=
  match E return irel EnvM (SemEnv E) with
    | nil => irelLeENil
    | T :: TS => fun i e de =>
      match e with 
        | nil => False
        | v :: vs => 
          irelLeV i v  (SND de) /\ 
          irelLeE i vs (FST de)
      end
  end.

Definition irelLeC (E:Env) (T:Ty) : irel CompM (SemCTypeL E T) := 
  DivMonadLe (@irelLeE E) (@irelLeV T).

(*
   Strict relation
*)
Definition relLeV (T:Ty) : rel ValueM (SemType T) := fun v dv =>
  forall k, irelLeV k v dv.

Definition relLeE (E:Env) : rel EnvM (SemEnv E) := fun e de =>
  forall k, irelLeE k e de.

Definition relLeC (E:Env) (T:Ty) : rel CompM (SemCTypeL E T) := fun comp df =>
  forall k, irelLeC k comp df.


(**
   >> relation
 **)

(* 
   Top Top Semantics 
*)


Definition pton_termv (pe:EnvM) (P: prdc ValueM) : prdc CESD := fun cesd =>
  match cesd with (c,e,s,d) =>
    forall val, P val ->
      terminate (c, pe++e, val::s, d)
  end.

Definition ntop_termc (pe:EnvM) (P: prdc CESD) : prdc CompM := fun comp =>
  match comp with (c1,s1) =>
    forall c e s d, P (c,e,s,d) ->
      terminate (c1++c, pe++e, s1++s, d)
  end.

(*
   Definition of >> Divergence Monad:
   Given relations between Values and domain elements,
   defines relations between Computations and domain elements.
*)

Definition rel_to_pred (A B:Type) (r:rel A B) (b:B) : prdc A := fun a => r a b.
Definition valset_ge := rel_to_pred.

Definition DivMonadGe (S D: cpo) (RS: rel EnvM S) (RD: rel ValueM D) : rel CompM (S -C-> DL D) := fun comp df =>
    forall e de, RS e de -> 
      forall d, (Val d == df de)%tord ->
        ntop_termc e (pton_termv e (valset_ge RD d)) comp.

(* 
   Approximation relation
*)

Definition erelGeENil : rel EnvM DOne := (fun e _ => match e with | nil => True | _ => False end).

Lemma erelGeENil_nil : erelGeENil nil tt.
intros; red; auto.
Qed.

(* define in this way to prevent automatic unfolding of erelGeV. *)
Definition erelGeV_body
  (erelGeV: forall (T:Ty), rel ValueM (SemType T))
  (T:Ty) :=
  match T return rel ValueM (SemType T) with
    | Int => fun v n => v = VN n
    | Bool => fun v b => 
      match v with 
        | VN 0 => b = false
        | VN (S _) => b = true
        | _ => False
      end
    | (T1 --> T2) => fun f df => 
      forall v dv, erelGeV T1 v dv ->
        DivMonadGe erelGeENil (erelGeV T2)
        (App::nil, v::f::nil) (K DOne (df dv))
    | (T1**T2) => fun v dv =>
      match (v, dv) with
        | (VP v1 v2, (dv1, dv2)) => erelGeV T1 v1 dv1 /\ erelGeV T2 v2 dv2
        | _ => False
      end
  end.

Fixpoint erelGeV (T:Ty) : rel ValueM (SemType T) :=
  @erelGeV_body erelGeV T.

Fixpoint erelGeE (E:Env) : rel EnvM (SemEnv E) :=
  match E return rel EnvM (SemEnv E) with
    | nil => erelGeENil
    | T :: TS => fun e de =>
      match e with 
        | nil => False
        | v :: vs => 
          erelGeV v  (SND de) /\ 
          erelGeE vs (FST de)
      end
  end.

Definition erelGeC (E:Env) (T:Ty) : rel CompM (SemCTypeL E T) := 
  DivMonadGe (@erelGeE E) (@erelGeV T).

(* 
   Strict relation
*)

Definition relGeV (T:Ty) : rel ValueM (SemType T) := fun v dv =>
  ideal_closure (erelGeV v) dv.

Definition relGeE (E:Env) : rel EnvM (SemEnv E) := fun e de =>
  ideal_closure (erelGeE e) de.

Definition relGeC (E:Env) (T:Ty) : rel CompM (SemCTypeL E T) := fun comp df =>
  ideal_closure (erelGeC comp) df.

(*
   == relations
*)

Definition relEqV T v d := @relLeV T v d /\ relGeV v d.
Definition relEqE E e d := @relLeE E e d /\ relGeE e d.
Definition relEqC E T c d := @relLeC E T c d /\ relGeC c d.

(*
   Observational equivalence
*)
Definition ObsEqV T v1 v2 := exists d: SemType T, relEqV v1 d /\ relEqV v2 d.
Definition ObsEqE E e1 e2 := exists d: SemEnv E, relEqE e1 d /\ relEqE e2 d.
Definition ObsEqC E T c1 c2 := exists d: SemCTypeL E T, relEqC c1 d /\ relEqC c2 d.

(*
   Notations
*)

Notation "i ':-' v '<v<' d" := (irelLeV i v d) (at level 55).
Notation "i ':-' e '<e<' d" := (irelLeE i e d) (at level 55).
Notation "i ':-' c '<c<' d" := (irelLeC i c d) (at level 55).

Notation "v '>v>' d" := (erelGeV v d) (at level 55).
Notation "e '>e>' d" := (erelGeE e d) (at level 55).
Notation "c '>c>' d" := (erelGeC c d) (at level 55).

Notation "v '<V<' d" := (relLeV v d) (at level 55).
Notation "e '<E<' d" := (relLeE e d) (at level 55).
Notation "c '<C<' d" := (relLeC c d) (at level 55).
Notation "c '<P<' d" := (relLeC (CC c) d) (at level 55).

Notation "v '>V>' d" := (relGeV v d) (at level 55).
Notation "e '>E>' d" := (relGeE e d) (at level 55).
Notation "c '>C>' d" := (relGeC c d) (at level 55).
Notation "c '>P>' d" := (relGeC (CC c) d) (at level 55).

Notation "v '=V=' d" := (relEqV v d) (at level 55).
Notation "e '=E=' d" := (relEqE e d) (at level 55).
Notation "c '=C=' d" := (relEqC c d) (at level 55).
Notation "c '=P=' d" := (relEqC (CC c) d) (at level 55).

Notation "       v1 '~V~' v2 ':::' T" := (ObsEqV T v1 v2) (at level 55).
Notation "       e1 '~E~' e2 ':::' E" := (ObsEqE E e1 e2) (at level 55).
Notation "E '|-' c1 '~C~' c2 ':::' T" := (ObsEqC E T c1 c2) (at level 55).
Notation "E '|-' c1 '~P~' c2 ':::' T" := (ObsEqC E T (CC c1) (CC c2)) (at level 55).

(*
   Lemmas for <<
*)

Lemma pton_kstepv_contra: forall pe p q, sub p q -> sub (pton_kstepv pe q) (pton_kstepv pe p).
do 2 red; destruct a; destruct c; do 2 destruct p0; intros.
eapply H0.
apply H.
apply H1.
assumption.
assumption.
Qed.

Lemma ntop_kstepc_contra: forall pe p q, sub p q -> sub (ntop_kstepc pe q) (ntop_kstepc pe p).
do 2 red; destruct a; destruct c; intros.
eapply H0.
apply H.
apply H1.
assumption.
assumption.
Qed.

Lemma irel_to_ipred_variance: forall A B (r r':irel A B) b b', 
  (forall k a, r k a b -> r' k a b') -> sub (irel_to_ipred r b) (irel_to_ipred r' b').
intros.
red.
intros.
destruct a.
simpl in H0 |- *.
apply H.
auto.
Qed.

Lemma irelLeV_fold: forall T k v (dv: SemType T),
  irelLeV_body irelLeV k v dv = irelLeV k v dv.
intros. destruct T. auto. auto. auto. auto.
Qed.

Lemma irelLeV_decr: forall T, ireldecr (@irelLeV T).
red; induction T.

(* Int *)
intros.
simpl in b, H0 |- *.
auto.

(* Bool *)
intros.
simpl in b, H0 |- *.
auto.

(* TArrow *)
intros.
simpl in H0 |- *.
intros.
apply H0.
omega.
auto.

(* TPair *)
intros.
simpl in H0 |- *.
destruct a.
auto.
auto.
auto.
simpl in b; destruct b.
destruct H0.
split.
eapply IHT1.
apply H.
assumption.
eapply IHT2.
apply H.
assumption.
Qed.

Lemma irelLeE_decr: forall E, ireldecr (@irelLeE E).
red; induction E.

intros.
auto.

intros; rename a0 into e.
destruct e.
inversion H0.
simpl in H0 |- *.
destruct H0.
split.
eapply irelLeV_decr.
apply H.
auto.
eapply IHE.
apply H.
auto.
Qed.

Lemma irelLeC_decr: forall E T, ireldecr (@irelLeC E T).
do 4 red; destruct a; intros.
eapply H0. instantiate (1:= k1); omega.
apply irelLeE_decr with k0. omega. apply H2. apply H3.
omega. omega.
Qed.

Lemma irelLeV_le: forall T k v (d d':SemType T), (d <= d')%tord -> k:- v <v< d -> k:- v <v< d'.
induction T.

(* Int *)
intros.
simpl in d, d'.
simpl in H.
rewrite H in H0.
auto.

(* Bool *)
intros.
simpl in d, d'.
simpl in H.
rewrite H in H0.
auto.

(* T1-->T2 *)
intros.
do 4 red; fold irelLeV; intros.
repeat red in H0.
eapply H0.
apply H1. fold irelLeV.
apply H2.
apply H3.
apply H4. fold irelLeV.

instantiate (1 := j).
instantiate (1 := de).
red; intros.
apply (H5 k3).
do 2 destruct H8. rewrite K_simpl in *.
assert (Val x <= d' dv)%tord.
eapply Ole_trans. apply H8. apply H.
destruct (DLvalval2 H12) as [? [? ?]].
exists x0.
split. auto.
eapply irelLeV_decr.
apply H9.
eapply IHT2. apply H14. auto.
omega. omega. omega. omega.

(* T1 ** T2 *)
intros.
destruct v.
simpl in H0; inversion H0.
simpl in H0; inversion H0.
simpl in H0; inversion H0.
simpl in d, d'; destruct d; destruct d'.
simpl in H; destruct H.
simpl in H0; destruct H0.
simpl; split.
eapply IHT1.
apply H.
assumption.
eapply IHT2.
apply H1.
assumption.
Qed.

Lemma irelLeC_le: forall E T k c (d d':SemCTypeL E T), (d <= d')%tord -> k:- c <c< d -> k:- c <c< d'.
do 3 red; destruct c; intros.
eapply H0.
apply H1.
apply H2.

instantiate (1 := j).
red; intros.
apply (H3 k2).

do 2 destruct H6.
assert (Val x <= d' de)%tord.
eapply Ole_trans. apply H6. apply H.
destruct (DLvalval2 H10) as [? [? ?]].
exists x0.
split. auto.
eapply irelLeV_decr. apply H7.
eapply irelLeV_le. apply H12. auto.
omega. omega. omega. omega.
Qed.

Lemma irelLeE_inv: forall E T k (var:Var E T) e (de:SemEnv E), k:- e <e< de ->
  exists v, nth_error e (idxVar var) = value v /\ k:- v <v< (SemVar var de).

induction var.

intros.
destruct e.
inversion H.
exists v.
split.
auto.
apply H.

intros.
destruct e.
inversion H.
simpl in de.
destruct de.
simpl in H.
destruct H.
destruct (IHvar _ _ H0); clear IHvar.
eexists.
split. apply H1.
simpl. rewrite fcont_comp_simpl.
apply H1.
Qed.

Lemma irelLeE_inv_initseg: forall E1 E2 k e (de:SemEnv (E1++E2)), k:- e <e< de ->
  exists e1, exists e2,
    e = e1++e2 /\
    k:- e1 <e< (projEnv E1 E2 de).

induction E1.

intros.
exists nil.
exists e.
simpl.
auto.

rename a into T.
intros.
simpl in de.
destruct de.
rename t0 into dv, t into de.
destruct e.
inversion H.
simpl in H.
fold SemEnv in H.
destruct H.
rewrite SND_simpl in H; simpl in H.
rewrite FST_simpl in H0; simpl in H0.
destruct (IHE1 E2 k e de H0).
destruct H1.
destruct H1.
exists (v::x).
exists x0.
split.
rewrite H1.
auto.
simpl.
fold SemEnv.
split.
rewrite SND_simpl; simpl; assumption.
rewrite FST_simpl; simpl; assumption.
Qed.

Lemma irelLeC_extends_irelLeV: forall T k v (d:SemType T),
  k:- v <v< d -> k:- (VToC v) <c< (DVtoDC d).

do 3 red; simpl; intros.
apply (H2 k1 v).
exists d.
split.
auto.
apply irelLeV_decr with k.
omega.
assumption.
omega.
omega.
Qed.

Lemma relLeC_le: forall E T c (d d':SemCTypeL E T), 
  (d <= d')%tord -> c <C< d -> c <C< d'.
red; intros.
apply irelLeC_le with d.
auto.
apply (H0 k).
Qed.

(*
   Lemmas for >>
*)

Lemma pton_termv_contra: forall e p q, sub p q -> sub (pton_termv e q) (pton_termv e p).
do 2 red; destruct a; do 2 destruct p0; intros.
eapply H0.
apply H.
apply H1.
Qed.

Lemma ntop_termc_contra: forall e p q, sub p q -> sub (ntop_termc e q) (ntop_termc e p).
do 2 red; destruct a; intros.
eapply H0.
apply H.
apply H1.
Qed.

Lemma rel_to_pred_variance: forall A B (r r':rel A B) b b', 
  (forall a, r a b -> r' a b') -> sub (rel_to_pred r b) (rel_to_pred r' b').
red.
intros.
red in H0 |- *.
apply H.
auto.
Qed.

Lemma erelGeV_fold: forall T v (dv: SemType T),
  erelGeV_body erelGeV v dv = erelGeV v dv.
intros. destruct T. auto. auto. auto. auto.
Qed.

Lemma erelGeV_ge: forall T v (d d': SemType T), (d' <= d)%tord -> v >v> d -> v >v> d'.
induction T.

(* Int *)
intros.
simpl in H.
rewrite H.
auto.

(* Bool *)
intros.
simpl in H.
rewrite H.
auto.

(* T1 --> T2 *)
do 4 red; fold erelGeV; intros.
rewrite K_simpl in *.
assert (Val d0 <= d dv)%tord. eapply Ole_trans. apply H3. apply H.
destruct (DLvalval2 H5) as [d0' [? ?]].
eapply H0. fold erelGeV. apply H1. apply H2. instantiate (1 := de). rewrite K_simpl in *. 
apply Oeq_sym. apply H6. fold erelGeV. 
generalize H4. apply pton_termv_contra. apply rel_to_pred_variance.
intros. eapply IHT2. apply H7. auto.

(* T1 ** T2 *)
intros.
destruct v.
simpl in H0; inversion H0.
simpl in H0; inversion H0.
simpl in H0; inversion H0.
simpl in d; destruct d.
simpl in d'; destruct d'.
simpl in H; destruct H.
simpl in H0; destruct H0.
simpl.
split.
eapply IHT1.
apply H.
assumption.
eapply IHT2.
apply H1.
assumption.
Qed.

Lemma erelGeC_ge: forall E T c (d d': SemCTypeL E T), (d' <= d)%tord -> c >c> d -> c >c> d'.
intros.
do 3 red; intros. destruct c. intros.
assert (Val d0 <= d de)%tord. eapply Ole_trans. apply H2. apply H.
destruct (DLvalval2 H4) as [d0' [? ?]].
eapply H0. apply H1. instantiate (1:= d0'). auto.
generalize H3. apply pton_termv_contra. apply rel_to_pred_variance.
intros. eapply erelGeV_ge. apply H6. auto.
Qed.

Lemma erelGeC_general: forall E T c (dc: SemCTypeL E T),
  c >c> dc -> 
  forall (e : EnvM) (de : SemEnv E),
    e >e> de ->
    forall d : SemType T,
      (Val d <= dc de)%tord  ->
      ntop_termc e (pton_termv e (valset_ge (erelGeV (T:=T)) d)) c.

intros. destruct c. red; intros.
destruct (DLvalval2 H1) as [d' [? ?]].
eapply H. apply H0. rewrite H3. apply Oeq_refl.
red; intros. apply H2.
do 2 red in H5 |- *.
eapply erelGeV_ge with d'. auto. auto.
Qed.

Lemma erelGeE_inv: forall E T (var:Var E T) e (de:SemEnv E), e >e> de ->
  exists v, nth_error e (idxVar var) = value v /\ v >v> (SemVar var de).

induction var.

intros.
destruct e.
inversion H.
exists v.
split.
auto.
apply H.

intros.
destruct e.
inversion H.
simpl in de.
destruct de.
simpl in H.
destruct H.
destruct (IHvar _ _ H0); clear IHvar.
eexists.
split. simpl. apply H1.
simpl. rewrite fcont_comp_simpl. apply H1.
Qed.

Lemma erelGeE_inv_initseg: forall E1 E2 e (de:SemEnv (E1++E2)), e >e> de ->
  exists e1, exists e2,
    e = e1++e2 /\
    e1 >e> (projEnv E1 E2 de).

induction E1.

intros.
exists nil.
exists e.
simpl.
auto.

rename a into T.
intros.
simpl in de.
destruct de.
rename t0 into dv, t into de.
destruct e.
inversion H.
simpl in H; fold SemEnv in H.
destruct H.
rewrite SND_simpl in H; simpl in H.
rewrite FST_simpl in H0; simpl in H0.
destruct (IHE1 E2 e de H0).
destruct H1.
destruct H1.
exists (v::x).
exists x0.
split.
rewrite H1.
auto.
simpl; fold SemEnv.
split.
rewrite SND_simpl; simpl; assumption.
rewrite FST_simpl; simpl; assumption.
Qed.

Lemma erelGeC_extends_erelGeV: forall T v (d:SemType T),
  v >v> d -> (VToC v) >c> (DVtoDC d).

repeat red.
intros.
apply (H2 v).
unfold rel_to_pred.
unfold DVtoDC in H1.
rewrite K_simpl in H1.
apply erelGeV_ge with d.
apply DLleVal_eq.
apply H1.
assumption.
assumption.
Qed.

Lemma relGeC_ge: forall E T c (d d': SemCTypeL E T), (d' <= d)%tord -> c >C> d -> c >C> d'.
intros. eapply ideal_closure_closed. apply H. apply H0.
Qed.

Lemma relEqC_eq: forall E T c (d d': SemCTypeL E T), (d' == d)%tord -> c =C= d -> c =C= d'.
intros. destruct H0. split.
eapply relLeC_le. apply H. apply H0.
eapply relGeC_ge. apply H. apply H1.
Qed.

Lemma val_ge_bot_fun: forall f T1 T2,
  @erelGeV (T1-->T2) f (@K _ (DL (SemType T2)) (DL_bot _)).

intros.
do 3 red. fold erelGeV SemType.
intros.
repeat rewrite K_simpl in H1.
absurd False; auto.
apply (@DL_bot_neq (SemType T2) d).
apply H1.
Qed.


(*
   Soundness (?) of Logical Relation
*)
Open Scope O_scope.

(* Functional operations on machine code *)
Definition Lambda (c:Code) := [PushC  (c++[Ret])]. 

Definition ComposeApp (cf cx:Code) := cf++cx++[App].

Hint Resolve Lambda ComposeApp.


(* 
   Soundness of Environment Extension

            E |- c1 ~P~ c2 ::: T 
     -----------------------------
      (E++E') |- c1 ~P~ c2 ::: T.
*)

Lemma compositionality_envext_le: forall E E' T c (dc: SemCTypeL E T),
  c <C< dc -> c <C< sem_precomp _ (projEnv E E') dc.
do 4 red; destruct c; intros.
destruct (irelLeE_inv_initseg H1).
do 2 destruct H5.
rewrite H5 in H1 |- *.
simpl. rewrite app_ass.

eapply H.
apply H0.
apply H6.

instantiate (1:= j).
red; intros; simpl.
rewrite <- app_ass.
rewrite <- H5.
apply (H2 k2 val).

do 2 destruct H7.
exists x1.
split.
apply H7.
eapply irelLeV_decr.
apply H8.
assumption.
omega.
omega.
omega.
omega.
Qed.


Lemma compositionality_envext_ge_e: forall E E' T c (dc: SemCTypeL E T), c >c> dc -> c >c> sem_precomp _ (projEnv E E') dc.
do 3 red; destruct c; intros.
destruct (erelGeE_inv_initseg H0).
do 2 destruct H3.
rewrite H3 in H0 |- *.
rewrite app_ass.

eapply H.
apply H4.
apply H1.

red; intros; simpl.
rewrite <- app_ass.
rewrite <- H3.
apply (H2 val).
assumption.
Qed.

Lemma compositionality_envext_ge: forall E E' T c (dc: SemCTypeL E T),
  c >C> dc -> c >C> sem_precomp _ (projEnv E E') dc.

red; intros. eapply cont_closed. apply ideal_closure_closed.
intros. apply ideal_closure_sub. apply compositionality_envext_ge_e. apply H0. auto.
Qed.

Corollary compositionality_envext: forall E E' T c (dc: SemCTypeL E T),
  c =P= dc -> c =P= sem_precomp _ (projEnv E E') dc.
intros.
destruct H.
split.

apply compositionality_envext_le; auto.
apply compositionality_envext_ge; auto.
Qed.

Corollary compositionality_envext_obs: forall E E' T c1 c2,
  E |- c1 ~P~ c2 ::: T -> (E++E') |- c1 ~P~ c2 ::: T.

intros.
do 2 destruct H.
exists (sem_precomp _ (projEnv E E') x).
split.
apply compositionality_envext; auto.
apply compositionality_envext; auto.
Qed.

(* 
   Soundness of Lambda Abstraction

            A :: E |- c1 ~P~ c2 ::: B 
   ---------------------------------------------
    E |- (Lambda c1) ~P~ (Lambda c2) ::: (A-->B)
*)

Lemma compositionality_lamabs_le: forall E A B c (dc:SemCTypeL (A::E) B),
  c <P< dc -> @relLeC E (A-->B) (CC(Lambda c)) (sem_op _ eta (sem_curry _ _ _ dc)).

do 4 red; simpl; fold SemEnv; intros.
eapply stepsfor_step.
apply step_pushc.
apply (H2 k1 (VC (e++e0) (c++[Ret]))).
exists (curry dc de).
split.
auto.

do 2 red; fold irelLeV; intros; simpl.
eapply stepsfor_step.
apply step_app.
rewrite app_comm_cons.

eapply H.
apply H5.
instantiate (1:= (de,dv)).
split.
rewrite SND_simpl; simpl; auto.
rewrite FST_simpl; simpl; auto.
eapply irelLeE_decr with k0. omega. assumption.

red; intros; simpl.
eapply stepsfor_step.
apply step_ret.
apply (H9 k5 val).

do 2 destruct H12.
exists x.
split.
rewrite K_simpl.
rewrite curry_simpl.
auto.
apply irelLeV_decr with i.
omega.
auto.

omega.
apply H14.
omega.
omega.
omega.
omega.
Qed.

Lemma compositionality_lamabs_ge_e: forall E A B c (d:SemCTypeL (A::E) B),
  CC c >c> d -> CC (Lambda c) >c> (sem_op _ eta (sem_curry _ _ _ d) : SemCTypeL E (A-->B)).

fold SemEnv. do 4 red; fold erelGeV; intros. simpl.
eapply terminate_step.
apply step_pushc.
apply H2.

do 5 red; fold erelGeV; simpl; intros.
eapply terminate_step.
apply step_app.
rewrite app_comm_cons.
eapply H.

instantiate (1 := (de,dv)).
split.
apply H3.
apply H0.

rewrite K_simpl in *.
eapply Oeq_trans.
apply H5. 
unfold sem_op in H1. rewrite fcont_COMP_simpl, fcont_comp_simpl, eta_val in H1.
rewrite (DLleVal_eq2 H1). auto.

red; intros.
eapply terminate_step.
apply step_ret.
apply (H6 val).
auto.
Qed.

Lemma compositionality_lamabs_ge: forall E A B c (dc:SemCTypeL (A::E) B),
  c >P> dc -> @relGeC E (A-->B) (CC(Lambda c)) (sem_op _ eta (sem_curry _ _ _ dc)).

fold SemEnv; intros. red. 
eapply cont_closed. apply ideal_closure_closed.
intros. apply ideal_closure_sub. pattern p. apply H0.
eapply cont_closed. apply ideal_closure_closed.
intros. apply ideal_closure_sub. apply compositionality_lamabs_ge_e. apply H0. auto.
Qed.

Corollary compositionality_lamabs: forall E A B c (dc: SemCTypeL (A::E) B),
  c =P= dc -> (Lambda c) =P= (sem_op _ eta (sem_curry _ _ _ dc) : SemCTypeL E (A-->B)).
fold SemEnv; intros.
split.
apply compositionality_lamabs_le.
apply (proj1 H).
apply compositionality_lamabs_ge.
apply (proj2 H).
Qed.

Corollary compositionality_lamabs_obs: forall E A B c1 c2,
  (A::E) |- c1 ~P~ c2  ::: B -> E |- (Lambda c1) ~P~ (Lambda c2)::: (A-->B).

intros.
do 2 destruct H.
exists (sem_op _ eta (sem_curry _ _ _ x)).
split.
apply compositionality_lamabs; auto.
apply compositionality_lamabs; auto.
Qed.


(****
   Soundness of Function application

     E |- f ~P~ g ::: (A-->B) 
     E |- x ~P~ y :::  A 
   ------------------------------------------------------
     E |- (ComposeApp f x) ~P~ (ComposeApp g y) ::: B 
****)

Lemma compositionality_funapp_le: forall E A B f x (df:SemCTypeL E (A-->B)) (dx:SemCTypeL E A),
  (f <P< df) -> (x <P< dx) -> (ComposeApp f x) <P< (sem_app _ _ _ df dx).

do 4 red; fold SemEnv SemType; simpl; intros.
unfold ComposeApp; simpl; repeat rewrite app_ass.
eapply H.
apply H1.
apply H2.
instantiate (1 := k1).

red; intros; simpl.
eapply H0.
apply H1.
apply H2.
instantiate (1 := k2).

red; intros; simpl.
do 2 destruct H6.
do 2 destruct H9.
replace (App::c) with ([App]++c).
replace (val0::val::s) with ([val0,val]++s).
replace (e++e0) with (nil++(e++e0)).

eapply H12.
instantiate (1 := k3). omega.
fold irelLeV; apply irelLeV_decr with i0. auto with arith. apply H13.
instantiate (1 := k3); auto.
apply irelLeENil_nil. fold irelLeV.
instantiate (1 := k3).

red; intros; simpl.
eapply H3.
do 2 destruct H14. rewrite K_simpl in *.
exists x2.
split.
eapply Oeq_trans.
apply H14.
rewrite sem_app_simpl. apply Oeq_refl.
assumption. assumption. apply H17.
omega. omega. omega. omega.
auto. auto. auto.
omega. omega. omega. omega.
Qed.

Lemma compositionality_funapp_ge_e: forall E A B f x (df:SemCTypeL E (A-->B)) (dx:SemCTypeL E A),
  CC f >c> df -> (CC x) >c> dx -> CC (ComposeApp f x) >c> sem_app _ _ _ df dx.

do 4 red; fold SemType; intros.
destruct (sem_app_inv H2) as [? [? [? [? ?]]]].

unfold ComposeApp. repeat rewrite app_ass.
eapply H. apply H1. apply Oeq_sym. apply H4.

red; intros.
eapply H0. apply H1. apply Oeq_sym. apply H5.

red; intros.
replace (val0::val::s) with ([val0,val]++s).
replace (e++e0) with (nil++(e++e0)).
eapply H7. fold erelGeV.
apply H8.
instantiate (1:= tt).
apply erelGeENil_nil.

rewrite K_simpl.
apply H6. fold erelGeV.

red; simpl; intros.
apply (H3 val1).
assumption.
auto.
auto.
Qed.

Lemma compositionality_funapp_ge: forall E A B f x (df:SemCTypeL E (A-->B)) (dx:SemCTypeL E A),
  f >P> df -> x >P> dx -> (ComposeApp f x) >P> (sem_app _ _ _ df dx).

intros. red. fold SemType. eapply cont2_closed. apply ideal_closure_closed.
intros. apply ideal_closure_sub. apply compositionality_funapp_ge_e. apply H1. apply H2.
auto. auto.
Qed.

Corollary compositionality_funapp: forall E A B f x (df:SemCTypeL E (A-->B)) (dx:SemCTypeL E A),
  f =P= df -> x =P= dx ->
  (ComposeApp f x) =P= (sem_app _ _ _ df dx).

intros.
split.
apply compositionality_funapp_le.
apply (proj1 H).
apply (proj1 H0).
apply compositionality_funapp_ge.
apply H.
apply H0.
Qed.

Corollary compositionality_funapp_obs: forall E A B f g x y,
  E |- f ~P~ g ::: (A-->B) -> E |- x ~P~ y ::: A -> 
    E |- (ComposeApp f x) ~P~ (ComposeApp g y) ::: B. 

intros.
do 2 destruct H; do 2 destruct H0.
exists (sem_app _ _ _ x0 x1).
split.
apply compositionality_funapp; auto.
apply compositionality_funapp; auto.
Qed.

(* compositionality pair *)

Lemma compositionality_pair_le: forall E T1 T2 c1 c2 (dv1:SemCTypeT E T1) (dv2:SemCTypeT E T2),
  c1 <P< sem_op _ eta dv1 -> c2 <P< sem_op _ eta dv2 -> (c1++c2++[MkPair]) <P< (sem_op _ eta (sem_op2 _ PAIR dv1 dv2) : SemCTypeL E (T1**T2)).

do 4 red; simpl; intros.
repeat rewrite app_ass.
eapply H.

apply H1.
apply H2.
instantiate (1:= k1).

red; intros.
eapply H0.

apply H1.
apply H2.
instantiate (1:= k2).

red; intros.
eapply stepsfor_step.
apply step_mkpair. simpl.
apply (H3 k3).

do 2 destruct H6.
do 2 destruct H9.
exists (x,x0).
split.

rewrite sem_op_simpl, fcont_comp_simpl, eta_val in H6, H9 |- *.
apply Val_stable.
rewrite (DLleVal_eq2 H6).
rewrite (DLleVal_eq2 H9). auto.

split.
eapply irelLeV_decr with i. omega. auto.
apply irelLeV_decr with i0. omega. auto.

omega. omega. omega. omega. omega. omega.
Qed.

Lemma compositionality_pair_ge: forall E T1 T2 c1 c2 (dv1:SemCTypeT E T1) (dv2:SemCTypeT E T2),
  c1 >P> sem_op _ eta dv1 -> c2 >P> sem_op _ eta dv2 -> (c1++c2++[MkPair]) >P> (sem_op _ eta (sem_op2 _ PAIR dv1 dv2) : SemCTypeL E (T1**T2)).
intros.
eapply ideal_closure_closed.
instantiate (1:= (uncurry _ _ _ (Smash _ _)) @_ (PROD_FUN _ _ _ (sem_op _ eta  dv1) (sem_op _ eta dv2))).
repeat rewrite sem_op_simpl.
apply (@fcont_le_intro (SemEnv E) (DL (Dprod (SemType T1) (SemType T2)))). intros.
repeat rewrite fcont_comp_simpl.
rewrite sem_op2_eval, PAIR_simpl, Pair_simpl.
rewrite PROD_FUN_simpl, PROD_fun_simpl, PAIR_simpl, Pair_simpl.
repeat rewrite fcont_comp_simpl.
repeat rewrite eta_val.
rewrite uncurry_simpl.
unfold Smash. rewrite Operator2_simpl, eta_val.
auto.

eapply cont2_closed.
Focus 3. apply H.
Focus 3. apply H0.

split.
red; intros. eapply ideal_closure_closed.
rewrite <- fcont_COMP_simpl.
rewrite (lub_comp_eq_cont _ chs). apply Ole_refl.
apply (proj1 (ideal_closure_closed _)). intros.
apply (H1 i).

red; intros. eapply ideal_closure_closed.
rewrite H1. apply Ole_refl. auto.

intros.
apply ideal_closure_sub.
unfold CC. do 3 red; intros. repeat rewrite app_ass. simpl.
rewrite fcont_comp_simpl, PROD_FUN_simpl, PROD_fun_simpl, PAIR_simpl, Pair_simpl in H4.
rewrite uncurry_simpl in H4. unfold Smash in H4.
destruct (Operator2Val (Oeq_sym H4)) as [x1 [x2 [? [? ?]]]].
eapply H1. apply H3. apply (Oeq_sym H6).

red; intros.
eapply H2. apply H3. apply (Oeq_sym H7).
red; intros.
eapply terminate_step. apply step_mkpair.
rewrite eta_val in H8. hreplace H8 with (DLleVal_eq2 H8).
apply H5.
do 2 red. eapply erelGeV_ge. apply H8. split.
auto. auto.
Qed.

Corollary compositionality_pair: forall E T1 T2 c1 c2 (dv1:SemCTypeT E T1) (dv2:SemCTypeT E T2),
  c1 =P= sem_op _ eta dv1 -> c2 =P= sem_op _ eta dv2 -> (c1++c2++[MkPair]) =P= (sem_op _ eta (sem_op2 _ PAIR dv1 dv2) : SemCTypeL E (T1**T2)).

intros.
split.
apply compositionality_pair_le.
apply (proj1 H).
apply (proj1 H0).
apply compositionality_pair_ge.
apply H.
apply H0.
Qed.

(****
   Adequacy of Logical Relation
****)

Definition comp_converge_to_int comp e n :=
  forall cesd,
    (terminate (mergeComp (VToC (VN n)) cesd) -> terminate (mergeComp comp (mergeEnv e cesd)))
    /\
    (diverge (mergeComp (VToC (VN n)) cesd) -> diverge (mergeComp comp (mergeEnv e cesd))).

Definition comp_converge_to_bool comp e b :=
  forall cesd,
    ((forall n, nat_to_bool n = b -> terminate (mergeComp (VToC (VN n)) (mergeEnv e cesd))) 
      -> terminate (mergeComp comp cesd))
    /\
    ((forall n, nat_to_bool n = b -> diverge (mergeComp (VToC (VN n)) (mergeEnv e cesd))) 
      -> diverge (mergeComp comp cesd)).

Definition comp_diverge comp e :=
  forall cesd, diverge (mergeComp comp (mergeEnv e cesd)).

Notation "c '-~>!'"   := (comp_diverge (CC c) nil) (at level 55).
Notation "c '-~>N' n" := (comp_converge_to_int (CC c) nil n) (at level 55).
Notation "c '-~>B' b" := (comp_converge_to_bool (CC c) nil b) (at level 55).


Lemma adequacy_le_bot: forall E T e (de:SemEnv E) c (dc:SemCTypeL E T),
  e <E< de -> c <C< dc -> (dc de == DL_bot _) -> comp_diverge c e.

do 2 red; intros.
destruct c; destruct cesd; do 2 destruct p; simpl.
eapply H0.

assert (k <= k)%nat. auto. apply H2.
apply H.

red; intros.

do 2 destruct H2.
rewrite H1 in H2.
absurd False; auto.
eapply DL_bot_neq.
apply H2.
omega.
instantiate (1:= k); auto.
Qed.

Lemma adequacy_le_int: forall E e (de:SemEnv E) c (dc:SemCTypeL E Int) n,
  e <E< de -> c <C< dc -> (dc de == Val n) ->
  forall cesd, 
    diverge (mergeComp (VToC (VN n)) (mergeEnv e cesd)) ->
    diverge (mergeComp c (mergeEnv e cesd)).

red; intros.
destruct c; destruct cesd; do 2 destruct p; simpl.
eapply H0.
assert (k <= k)%nat. auto. apply H3.
apply H.

red; intros.
do 2 destruct H3.
rewrite H1 in H3.
hreplace H3 with (DLleVal_eq2 H3).
simpl in H3, H6.
rewrite H6. destruct H3. simpl in *.
rewrite H3.
apply H2.
omega. apply le_n.
Qed.

Lemma adequacy_ge_int: forall E e (de:SemEnv E) c (dc:SemCTypeL E Int) n,
  e >E> de -> c >C> dc -> (dc de == Val n) ->
  forall cesd, 
    terminate (mergeComp (VToC (VN n)) (mergeEnv e cesd)) ->
    terminate (mergeComp c (mergeEnv e cesd)).

intros.
assert (AppCpoSet (erelGeC c) (erelGeE (E:=E) e) (Val n)).
  apply (@ideal_closure_flat _ _ n).
  intros. destruct H4 as [? [? ?]]. exists x. exists x0. rewrite <- H3. auto.

  apply AppCpoSet_closed.
  exists dc. exists de. auto.

destruct H3 as [dc' [de' [? [? ?]]]].

destruct c as [? ?]. destruct cesd as [[[? ?] ?] ?]. simpl in H2 |- *.
eapply H3. apply H4. apply H5.
red; intros. repeat red in H6. rewrite H6. auto.
Qed.

Lemma adequacy_le_bool: forall E e (de:SemEnv E) c (dc:SemCTypeL E Bool) b,
  e <E< de -> c <C< dc -> (dc de == Val b) ->
  forall cesd, 
    (forall n, nat_to_bool n = b -> diverge (mergeComp (VToC (VN n)) (mergeEnv e cesd))) ->
    diverge (mergeComp c (mergeEnv e cesd)).

red; destruct c; destruct cesd; do 2 destruct p; intros; simpl.
eapply H0.
assert (k <= k)%nat. auto. apply H3.
apply H.

red; intros.
do 2 destruct H3.
rewrite H1 in H3.
hreplace H3 with (DLleVal_eq2 H3).
red in H3. simpl in H3.

destruct val.
apply (H2 n).
destruct n.
simpl in H6 |- *. rewrite <- H6. intuition.
simpl in H6 |- *. rewrite <- H6. intuition.

inversion H6.
inversion H6.
inversion H6.
omega. apply le_n.
Qed.

Lemma adequacy_ge_bool: forall E e (de:SemEnv E) c (dc:SemCTypeL E Bool) b,
  e >E> de -> c >C> dc -> (dc de == Val b) ->
  forall cesd, 
    (forall n, nat_to_bool n = b -> terminate (mergeComp (VToC (VN n)) (mergeEnv e cesd))) ->
    terminate (mergeComp c (mergeEnv e cesd)).

intros.
assert (AppCpoSet (erelGeC c) (erelGeE (E:=E) e) (Val b)).
  apply (@ideal_closure_flat _ _ b).
  intros. destruct H4 as [? [? ?]]. exists x. exists x0. rewrite <- H3. auto.

  apply AppCpoSet_closed.
  exists dc. exists de. auto.

destruct H3 as [dc' [de' [? [? ?]]]].

destruct c as [? ?]. destruct cesd as [[[? ?] ?] ?]. simpl in H2 |- *.
eapply H3. apply H4. apply H5.
red; intros. repeat red in H6.
destruct val. destruct n.
  apply H2. simpl. auto.
  apply H2. simpl. auto.
contradiction. contradiction. contradiction.
Qed.

Corollary adequacy_bot: forall T c (dc: SemCTypeL nil T),
  c =P= dc -> (dc tt == DL_bot _) -> (c -~>!) .

intros.
destruct H.
red; destruct cesd; destruct p; destruct p.
unfold mergeComp; simpl.
apply (@adequacy_le_bot nil T nil tt (CC c) dc irelLeENil_nil H H0 (c0,e,s,d)).
Qed.

Corollary adequacy_int: forall c (dc: SemCTypeL nil Int) n,
  c =P= dc -> (dc tt == Val n) -> (c -~>N n).

red; intros; destruct H.
destruct cesd; destruct p; destruct p.
split.
unfold mergeComp; simpl.
assert (@relGeE nil nil tt). apply ideal_closure_sub. simpl. auto.
apply (@adequacy_ge_int nil nil tt (CC c) dc n H2 H1 H0 (c0,e,s,d)).

unfold mergeComp; simpl.
assert (@relLeE nil nil tt). red. apply irelLeENil_nil.
apply (@adequacy_le_int nil nil tt (CC c) dc n H2 H H0 (c0,e,s,d)).
Qed.

Corollary adequacy_bool: forall c (dc: SemCTypeL nil Bool) b,
  c =P= dc -> (dc tt == Val b) -> (c -~>B b).

red; intros; destruct H.
destruct cesd; destruct p; destruct p.
split.
unfold mergeComp; simpl.
assert (@relGeE nil nil tt). apply ideal_closure_sub. simpl. auto.
apply (@adequacy_ge_bool nil nil tt (CC c) dc b H2 H1 H0 (c0,e,s,d)).

unfold mergeComp; simpl.
assert (@relLeE nil nil tt). red. apply irelLeENil_nil.
apply (@adequacy_le_bool nil nil tt (CC c) dc b H2 H H0 (c0,e,s,d)).
Qed.

