(** * Predom.v: Specification and properties of a cpo 
   This version is part of the formalization for
   Biorthogonality, Step Indexing and Compiler Correctness
   N Benton & C-K Hur
   April 2009 
*)

(* Modified from Christine's original *)

Require Import utility.
Require Import ConstructiveEpsilon.

Set Implicit Arguments.
Unset Strict Implicit.
Open Scope nat_scope.


(** ** Ordered type *)
Record ord : Type := mk_ord
  {tord:>Type; Ole:tord->tord->Prop; Ole_refl : forall x :tord, Ole x x; 
   Ole_trans : forall x y z:tord, Ole x y -> Ole y z -> Ole x z}.

Hint Resolve Ole_refl Ole_trans.

Hint Extern 2  (Ole (o:=?X1) ?X2 ?X3 ) => simpl Ole.

Bind Scope O_scope with tord.
Delimit Scope O_scope with tord.

Infix "<=" := Ole : O_scope.
Open Scope O_scope.

(** *** Associated equality *)
Definition Oeq (O:ord) (x y : O) := x <= y /\ y <= x.
Infix "==" := Oeq (at level 70) : O_scope.

Lemma Ole_refl_eq : forall  (O:ord) (x y:O), x=y -> x <= y.
intros O x y H; rewrite H; auto.
Save.

Hint Resolve Ole_refl_eq.

Lemma Ole_antisym : forall (O:ord) (x y:O), x<=y -> y <=x -> x==y.
red; auto.
Save.
Hint Immediate Ole_antisym.

Lemma Oeq_refl : forall (O:ord) (x:O), x == x.
red; auto.
Save.
Hint Resolve Oeq_refl.

Lemma Oeq_refl_eq : forall (O:ord) (x y:O), x=y -> x == y.
intros O x y H; rewrite H; auto.
Save.
Hint Resolve Oeq_refl_eq.

Lemma Oeq_sym : forall (O:ord) (x y:O), x == y -> y == x.
unfold Oeq; intuition.
Save.

Lemma Oeq_le : forall (O:ord) (x y:O), x == y -> x <= y.
unfold Oeq; intuition.
Save.

Lemma Oeq_le_sym : forall (O:ord) (x y:O), x == y -> y <= x.
unfold Oeq; intuition.
Save.

Hint Resolve Oeq_le.
Hint Immediate Oeq_sym Oeq_le_sym.

Lemma Oeq_trans : forall (O:ord) (x y z:O), x == y -> y == z -> x == z.
unfold Oeq; split; apply Ole_trans with y; auto.
Save.
Hint Resolve Oeq_trans.

(** *** Setoid relations *)

Add Parametric Relation (O:ord) : O (@Oeq O)
       reflexivity proved by (@Oeq_refl O) symmetry proved by (@Oeq_sym O)
       transitivity proved by (@Oeq_trans O) as Oeq_Relation.

Add Parametric Relation (O:ord) : O (@Ole O)
       reflexivity proved by (@Ole_refl O) 
       transitivity proved by (@Ole_trans O) as Ole_Relation.

Add Parametric Morphism (O:ord) : (@Ole O) with signature (@Oeq O) ==> (@Oeq O) ==> iff as Ole_eq_compat_iff.
split; intuition.
apply Ole_trans with x; firstorder using Ole_trans.
apply Ole_trans with y.
firstorder.
apply Ole_trans with y0.
assumption.
intuition.
Save.

Lemma Ole_eq_compat : 
     forall (O : ord) (x1 x2 : O),
       x1 == x2 -> forall x3 x4 : O, x3 == x4 -> x1 <= x3 -> x2 <= x4.
intros; apply Ole_trans with x1; firstorder using Ole_trans.
Save.

Lemma Ole_eq_right : forall (O : ord) (x y z: O),
             x <= y -> y==z -> x<=z.
intros; apply Ole_eq_compat with x y; auto.
Save.

Lemma Ole_eq_left : forall (O : ord) (x y z: O),
             x == y -> y<=z -> x<=z.
intros; apply Ole_eq_compat with y z; auto.
Save.

(** *** Dual order *)

Definition Iord (O:ord):ord.
intros O; exists O (fun x y : O => y <= x); intros; auto.
apply Ole_trans with y; auto.
Defined.

(** *** Order on functions *)

Definition ford (A:Type) (O:ord) : ord.
intros A O; exists (A->O) (fun f g:A->O => forall x, f x <= g x); intros; auto.
apply Ole_trans with (y x0); auto.
Defined.

Infix "-o>" := ford (right associativity, at level 30) : O_scope .

Lemma ford_le_elim : forall A (O:ord) (f g:A -o> O), f <= g ->forall n, f n <= g n.
auto.
Save.
Hint Immediate ford_le_elim.

Lemma ford_le_intro : forall A (O:ord) (f g:A-o>O), (forall n, f n <= g n) -> f <= g.
auto.
Save.
Hint Resolve ford_le_intro.

Lemma ford_eq_elim : forall A (O:ord) (f g:A -o> O), f == g ->forall n, f n == g n.
firstorder.
Save.
Hint Immediate ford_eq_elim.

Lemma ford_eq_intro : forall A (O:ord) (f g:A -o> O), (forall n, f n == g n) -> f == g.
red; auto.
Save.
Hint Resolve ford_eq_intro.

Hint Extern 2 (Ole (o:=ford ?X1 ?X2) ?X3 ?X4) => intro.

(** ** Monotonicity *)

(** *** Definition and properties *)

Definition monotonic (O1 O2:ord) (f : O1 -> O2) := forall x y, x <= y -> f x <= f y.
Hint Unfold monotonic.

Definition stable (O1 O2:ord) (f : O1 -> O2) := forall x y, x == y -> f x == f y.
Hint Unfold stable.

Lemma monotonic_stable : forall (O1 O2 : ord) (f:O1 -> O2), 
             monotonic f -> stable f.
unfold monotonic, stable; firstorder.
Save.
Hint Resolve monotonic_stable.

(** *** Type of monotonic functions *)

Record fmono (O1 O2:ord) : Type := mk_fmono
            {fmonot :> O1->O2; fmonotonic: monotonic fmonot}.
Hint Resolve fmonotonic.

Definition fmon (O1 O2:ord) : ord.
intros O1 O2; exists (fmono O1 O2) (fun f g:fmono O1 O2 => forall x, f x <= g x); intros; auto.
apply Ole_trans with (y x0); auto.
Defined.

Infix "-m>" := fmon (at level 30, right associativity) : O_scope.

Lemma fmon_stable : forall (O1 O2:ord) (f:O1-m>O2), stable f.
intros; apply monotonic_stable; auto.
Save.
Hint Resolve fmon_stable.

Lemma fmon_le_elim : forall (O1 O2:ord) (f g:O1 -m> O2), f <= g ->forall n, f n <= g n.
auto.
Save.
Hint Immediate fmon_le_elim.

Lemma fmon_le_intro : forall (O1 O2:ord) (f g:O1 -m> O2), (forall n, f n <= g n) -> f <= g.
auto.
Save.
Hint Resolve fmon_le_intro.

Lemma fmon_eq_elim : forall (O1 O2:ord) (f g:O1 -m> O2), f == g ->forall n, f n == g n.
firstorder.
Save.
Hint Immediate fmon_eq_elim.

Lemma fmon_eq_intro : forall (O1 O2:ord) (f g:O1 -m> O2), (forall n, f n == g n) -> f == g.
red; auto.
Save.
Hint Resolve fmon_eq_intro.

Hint Extern 2 (Ole (o:=fmon ?X1 ?X2) ?X3 ?X4) => intro.

(** *** Monotonicity and dual order *)

Definition Imon : forall O1 O2, (O1-m>O2) -> Iord O1 -m> Iord O2.
intros O1 O2 f; exists (f:Iord O1 -> Iord O2); red; simpl; intros.
apply (fmonotonic f); auto.
Defined.

Definition Imon2 : forall O1 O2 O3, (O1-m>O2-m>O3) -> Iord O1 -m> Iord O2 -m> Iord O3.
intros O1 O2 O3 f; exists (fun (x:Iord O1) => Imon (f x)); red; simpl; intros.
apply (fmonotonic f); auto.
Defined.

(** *** Monotonic functions with 2 arguments *)
Definition le_compat2_mon : forall (O1 O2 O3:ord)(f:O1 -> O2 -> O3),
     (forall (x y:O1) (z t:O2), x<=y -> z <= t -> f x z <= f y t) -> (O1-m>O2-m>O3).
intros O1 O2 O3 f Hle; exists (fun (x:O1) => mk_fmono (fun z t => Hle x x z t (Ole_refl x))).
red; intros; intro a; simpl; auto. 
Defined.

(** ** Sequences *)
(** *** Order on natural numbers *)
Definition natO : ord.
    exists nat (fun n m : nat => (n <= m)%nat); intros; auto with arith.
    apply le_trans with y; auto.
Defined.

Definition fnatO_intro : forall (O:ord) (f:nat -> O), (forall n, f n <= f (S n)) -> natO-m>O.
intros; exists f; red; simpl; intros.
elim H0; intros; auto.
apply Ole_trans with (f m); trivial.
Defined.

Lemma fnatO_elim : forall (O:ord) (f:natO -m> O) (n:nat), f n <= f (S n).
intros; apply (fmonotonic f); auto.
Save.
Hint Resolve fnatO_elim.


(** - (mseq_lift_left f n) k = f (n+k) *)
Definition mseq_lift_left : forall (O:ord) (f:natO -m> O) (n:nat), natO -m> O.
intros; exists (fun k => f (n+k)%nat); red; intros.
apply (fmonotonic f); auto with arith.
Defined.

Lemma mseq_lift_left_le_compat : forall (O:ord) (f g:natO -m> O) (n:nat), 
             f <= g -> mseq_lift_left f n <= mseq_lift_left g n.
intros; intro; simpl; auto.
Save.
Hint Resolve mseq_lift_left_le_compat.

Add Parametric Morphism (O:ord) : (@mseq_lift_left O)
 with signature (@Oeq (natO -m> O))  ==> (@eq nat) ==> (@Oeq (natO -m> O))
as mseq_lift_left_eq_compat.
intros; apply Ole_antisym; auto.
Save.
Hint Resolve mseq_lift_left_eq_compat.

(** - (mseq_lift_left f n) k = f (k+n) *)
Definition mseq_lift_right : forall (O:ord) (f:natO -m> O) (n:nat), natO -m> O.
intros; exists (fun k => f (k+n)%nat); red; intros.
apply (fmonotonic f); auto with arith.
Defined.

Lemma mseq_lift_right_le_compat : forall (O:ord) (f g:natO -m> O) (n:nat), 
             f <= g -> mseq_lift_right f n <= mseq_lift_right g n.
intros; intro; simpl; auto.
Save.
Hint Resolve mseq_lift_right_le_compat.

Add Parametric Morphism (O:ord) : (@mseq_lift_right O) 
with signature (@Oeq (natO -m> O)) ==> (@eq nat) ==> (@Oeq (natO -m> O))
as mseq_lift_right_eq_compat.
intros; apply Ole_antisym; auto.
Save.

Lemma mseq_lift_right_left : forall (O:ord) (f:natO -m> O) n, 
       mseq_lift_left f n  == mseq_lift_right f n.
intros; apply fmon_eq_intro; unfold mseq_lift_left,mseq_lift_right; simpl; intros.
replace (n0+n)%nat with (n+n0)%nat; auto with arith.
Save.

(** *** Monotonicity and functions *)
(** -  (ford_app f x) n = f n x *) 
Definition ford_app : forall (A:Type)(O1 O2:ord)(f:O1 -m> (A -o> O2))(x:A), O1 -m> O2.
intros; exists (fun n => f n x); intros.
intro n; intros.
assert (f n <= f y); auto.
apply (fmonotonic f); trivial.
Defined.

Infix "<o>" := ford_app (at level 30, no associativity) : O_scope.

Lemma ford_app_simpl : forall (A:Type)(O1 O2:ord)  (f : O1 -m> A -o> O2) (x:A)(y:O1),
            (f <o> x) y = f y x.
trivial.
Save.

Lemma ford_app_le_compat : forall (A:Type)(O1 O2:ord) (f g:O1 -m> A -o> O2) (x:A), 
             f <= g -> f <o> x  <= g <o> x.
intros; intro; simpl; auto.
Save.
Hint Resolve ford_app_le_compat.

Add Parametric Morphism (A:Type) (O1 O2 : ord) : (@ford_app A O1 O2)
with signature (@Oeq (O1 -m> (A -o> O2))) ==> (@eq A) ==> (@Oeq (O1 -m> O2))
as ford_app_eq_compat.
intros; apply Ole_antisym; auto.
Save.

(** - ford_shift f x y == f y x *)
Definition ford_shift : forall (A:Type)(O1 O2:ord)(f:A -o> (O1 -m> O2)), O1 -m> (A-o>O2).
intros; exists (fun x y => f y x); intros.
intros n x H y.
apply (fmonotonic (f y)); trivial.
Defined.

Lemma ford_shift_le_compat : forall (A:Type)(O1 O2:ord) (f g: A -o> (O1 -m> O2)), 
             f <= g -> ford_shift f  <= ford_shift g.
intros; intro; simpl; auto.
Save.
Hint Resolve ford_shift_le_compat.

Add Parametric Morphism (A:Type) (O1 O2 : ord) : (@ford_shift A O1 O2)
with signature (@Oeq (A -o> (O1 -m> O2))) ==> (@Oeq (O1 -m> (A -o> O2)))
as ford_shift_eq_compat.
intros; apply Ole_antisym; auto.
Save.


(**  - (fmon_app f x) n = f n x *) 

Definition fmon_app : forall (O1 O2 O3:ord)(f:O1 -m> O2 -m> O3)(x:O2), O1 -m> O3.
intros; exists (fun n => f n x); intros.
intro n; intros.
assert (f n <= f y); auto.
apply (fmonotonic f); trivial.
Defined.

Infix "<_>" := fmon_app (at level 35, no associativity) : O_scope.

Lemma fmon_app_simpl : forall  (O1 O2 O3:ord)(f:O1 -m> O2 -m> O3)(x:O2)(y:O1),
      (f <_> x)  y = f y x.
trivial.
Save.

Lemma fmon_app_le_compat : forall (O1 O2 O3:ord) (f g:O1 -m> (O2 -m> O3)) (x y:O2), 
             f <= g -> x <= y -> f <_> x  <= g <_> y.
red; intros; simpl; intros; auto.
apply Ole_trans with (f x0 y); auto.
apply (fmonotonic (f x0)); auto.
Save.
Hint Resolve fmon_app_le_compat.

Add Parametric Morphism (O1 O2 O3 : ord) : (@fmon_app O1 O2 O3)
with signature (@Oeq (O1 -m> O2 -m> O3)) ==> (@Oeq O2) ==> (@Oeq (O1 -m> O3))
as fmon_app_eq_compat.
intros; apply Ole_antisym; intros; auto.
Save.

(** - fmon_id c = c *)
Definition fmon_id :  forall (O:ord), O -m> O.
intros; exists (fun (x:O)=>x).
intro n; auto.
Defined.

Lemma fmon_id_simpl : forall (O:ord) (x:O), fmon_id O x = x.
trivial.
Save.

(**  - (fmon_cte c) n = c *) 
Definition fmon_cte :  forall (O1 O2:ord)(c:O2), O1 -m> O2.
intros; exists (fun (x:O1)=>c).
intro n; auto.
Defined.

Lemma fmon_cte_simpl : forall  (O1 O2:ord) (c:O2) (x:O1), fmon_cte O1 c x = c.
trivial.
Save.


Definition mseq_cte : forall O:ord, O -> natO-m>O := fmon_cte natO.

Lemma fmon_cte_le_compat : forall (O1 O2:ord) (c1 c2:O2), 
             c1 <= c2 -> fmon_cte O1 c1 <= fmon_cte O1 c2.
intros; intro; auto.
Save.

Add Parametric Morphism (O1 O2 : ord) : (@fmon_cte O1 O2) 
with signature (@Oeq O2) ==>  (@Oeq (O1 -m> O2))
as fmon_cte_eq_compat.
intros; apply Ole_antisym; auto.
Save.

(** - (fmon_diag h) n = h n n *) 
Definition fmon_diag : forall (O1 O2:ord)(h:O1 -m> (O1 -m> O2)), O1 -m> O2.
intros; exists (fun n => h n n).
red; intros.
apply Ole_trans with (h x y); auto.
apply (fmonotonic (h x)); auto.
assert (h x <= h y); auto.
apply (fmonotonic h); trivial.
Defined.

Lemma fmon_diag_le_compat :  forall (O1 O2:ord) (f g:O1 -m> (O1 -m> O2)), 
             f <= g -> fmon_diag f <= fmon_diag g.
intros; intro; simpl; auto.
Save.
Hint Resolve fmon_diag_le_compat.

Lemma fmon_diag_simpl : forall (O1 O2:ord) (f:O1 -m> (O1 -m> O2)) (x:O1), 
             fmon_diag f x = f x x.
trivial.
Save.

Add Parametric Morphism (O1 O2 : ord) : (@fmon_diag O1 O2)
with signature (@Oeq (O1 -m> (O1 -m> O2))) ==> (@Oeq (O1 -m> O2))
as fmon_diag_eq_compat.
intros; apply Ole_antisym; auto.
Save.

(** - (fmon_shift h) n m = h m n *) 
Definition fmon_shift : forall (O1 O2 O3:ord)(h:O1 -m> O2 -m>  O3), O2 -m> O1 -m>  O3.
intros; exists (fun m => h <_> m).
intro n; simpl; intros.
apply (fmonotonic (h x)); trivial.
Defined.

Lemma fmon_shift_simpl : forall (O1 O2 O3:ord)(h:O1 -m> O2 -m>  O3) (x : O2) (y:O1),
      fmon_shift h x y = h y x.
trivial.
Save.

Lemma fmon_shift_le_compat :  forall (O1 O2 O3:ord) (f g:O1 -m>  O2 -m>  O3), 
             f <= g -> fmon_shift f <= fmon_shift g.
intros; intro; simpl; intros.
assert (f x0 <= g x0); auto.
Save.
Hint Resolve fmon_shift_le_compat.

Add Parametric Morphism (O1 O2 O3 : ord) : (@fmon_shift O1 O2 O3)
with signature (@Oeq (O1 -m> O2 -m> O3)) ==> (@Oeq (O2 -m> O1 -m> O3))
as fmon_shift_eq_compat.
intros; apply Ole_antisym; auto.
Save.

Lemma fmon_shift_shift_eq :  forall (O1 O2 O3:ord) (h : O1 -m> O2 -m>  O3),
             fmon_shift (fmon_shift h) == h.
intros; apply fmon_eq_intro; unfold fmon_shift; simpl; auto.
Save.

(** - (f@g) x = f (g x) *)
Definition fmon_comp : forall O1 O2 O3:ord, (O2 -m> O3) -> (O1 -m> O2) -> O1 -m> O3.
intros O1 O2 O3 f g; exists (fun n => f (g n)); red; intros.
apply (fmonotonic f).
apply (fmonotonic g); auto.
Defined.

Infix "@" := fmon_comp (at level 35) : O_scope.

Lemma fmon_comp_simpl : forall (O1 O2 O3:ord) (f :O2 -m> O3) (g:O1 -m> O2) (x:O1),
            (f @ g) x = f (g x).
trivial.
Save.

(** - (f@2 g) h x = f (g x) (h x) *)
Definition fmon_comp2 : 
    forall O1 O2 O3 O4:ord, (O2 -m> O3 -m> O4) -> (O1 -m> O2) -> (O1 -m> O3) -> O1-m>O4.
intros O1 O2 O3 O4 f g h; exists (fun n => f (g n) (h n)); red; intros.
apply Ole_trans with (f (g x) (h y)); auto.
apply (fmonotonic (f (g x))).
apply (fmonotonic h); auto.
apply (fmonotonic f); auto. 
apply (fmonotonic g); auto. 
Defined.

Infix "@2" := fmon_comp2 (at level 70) : O_scope.

Lemma fmon_comp2_simpl : 
    forall (O1 O2 O3 O4:ord) (f:O2 -m> O3 -m> O4) (g:O1 -m> O2) (h:O1 -m> O3) (x:O1),
            (f @2 g) h x = f (g x) (h x).
trivial.
Save.

Add Parametric Morphism (O1 O2 O3 : ord) : (@fmon_comp O1 O2 O3)
with signature (@Ole (O2 -m> O3)) ++> (@Ole (O1 -m> O2)) ++> (@Ole (O1 -m> O3)) 
as fmon_comp_le_compat_morph.
intros f1 f2 H g1 g2 H1; red; intros.
simpl.
intros.
apply Ole_trans with (f2 (g1 x)); auto.
apply (fmonotonic f2); auto.
Save.

Lemma fmon_comp_le_compat : 
      forall (O1 O2 O3:ord) (f1 f2: O2 -m> O3) (g1 g2:O1 -m> O2),
                 f1 <= f2 -> g1<= g2 -> f1 @ g1 <= f2 @ g2.
intros; exact (fmon_comp_le_compat_morph H H0).
Save.
Hint Immediate fmon_comp_le_compat.

Add Parametric Morphism (O1 O2 O3 : ord) : (@fmon_comp O1 O2 O3) 
with signature (@Oeq (O2 -m> O3)) ==> (@Oeq (O1 -m> O2)) ==> (@Oeq (O1 -m> O3)) 
as fmon_comp_eq_compat.
intros; apply Ole_antisym; apply fmon_comp_le_compat; auto.
Save.
Hint Immediate fmon_comp_eq_compat.


Lemma fmon_comp_monotonic2 : 
      forall (O1 O2 O3:ord) (f: O2 -m> O3) (g1 g2:O1 -m> O2),
               g1<= g2 -> f @ g1 <= f @ g2.
auto.
Save.
Hint Resolve fmon_comp_monotonic2.

Lemma fmon_comp_monotonic1 : 
      forall (O1 O2 O3:ord) (f1 f2: O2 -m> O3) (g:O1 -m> O2),
               f1<= f2 -> f1 @ g <= f2 @ g.
auto.
Save.
Hint Resolve fmon_comp_monotonic1.

Definition fcomp : forall O1 O2 O3:ord,  (O2 -m> O3) -m> (O1 -m> O2) -m> (O1 -m> O3).
   intros; exists (fun f : O2 -m> O3 => 
                  mk_fmono (fmonot:=fun g : O1 -m> O2 => fmon_comp f g)
                                    (fmon_comp_monotonic2 f)).
red; intros; simpl; intros.
apply H.
Defined.

Lemma fmon_le_compat : forall (O1 O2:ord) (f: O1 -m> O2) (x y:O1), x<=y -> f x <= f y.
intros; apply (fmonotonic f); auto.
Save.
Hint Resolve fmon_le_compat.

Lemma fmon_le_compat2 : forall (O1 O2 O3:ord) (f: O1 -m> O2 -m> O3) (x y:O1) (z t:O2),
             x<=y -> z <=t -> f x z <= f y t.
intros; apply Ole_trans with (f x t).
apply (fmonotonic (f x)); auto.
apply (fmonotonic f); auto.
Save.
Hint Resolve fmon_le_compat2.

Lemma fmon_cte_comp : forall (O1 O2 O3:ord)(c:O3)(f:O1-m>O2),
              fmon_cte O2 c @ f == fmon_cte O1 c.
intros; apply fmon_eq_intro; intro x; auto.
Save.

(** ** Basic operators of omega-cpos *)
(** - Constant : $0$ REMOVED
     - lub : limit of monotonic sequences
*)


(** *** Definition of cpos *)
Record cpo : Type := mk_cpo 
  {tcpo:>ord; (* D0 : tcpo;*) lub: (natO -m> tcpo) -> tcpo;
(*   Dbot : forall x:tcpo, D0 <= x;  *)
   le_lub : forall (f : natO -m> tcpo) (n:nat), f n <= lub f;
   lub_le : forall (f : natO -m> tcpo) (x:tcpo), (forall n, f n <= x) -> lub f <= x}.

(*
Implicit Arguments D0 [c].
Notation "0" := D0 : O_scope.
*)

Hint Resolve (* Dbot *) le_lub lub_le.

(** *** Least upper bounds *)

Add Parametric Morphism (c:cpo) : (@lub c) 
with signature (@Ole (natO -m> c)) ++> (@Ole c) 
as lub_le_compat_morph.
intros f g H; apply lub_le; intros.
apply Ole_trans with (g n); auto.
Save.
Hint Resolve lub_le_compat_morph.

Lemma lub_le_compat : forall (D:cpo) (f g:natO -m> D), f <= g -> lub f <= lub g.
intros; apply lub_le; intros.
apply Ole_trans with (g n); auto.
Save.
Hint Resolve lub_le_compat.

Definition Lub : forall (D:cpo), (natO -m> D) -m> D.
intro D; exists (fun (f :natO-m>D) => lub f); red; auto.
Defined.

Add Parametric Morphism (c:cpo) : (@lub c) 
with signature (@Oeq (natO -m> c)) ==> (@Oeq c) as lub_eq_compat.
intros; apply Ole_antisym; auto.
Save.
Hint Resolve lub_eq_compat.

Lemma lub_cte : forall (D:cpo) (c:D), lub (fmon_cte natO c) == c.
intros; apply Ole_antisym; auto.
apply le_lub with (f:=fmon_cte natO c) (n:=O); auto.
Save.

Hint Resolve lub_cte.

Lemma lub_lift_right : forall (D:cpo) (f:natO -m> D) n, lub f == lub (mseq_lift_right f n).
intros; apply Ole_antisym; auto.
apply lub_le_compat; intro.
unfold mseq_lift_right; simpl.
apply (fmonotonic f); auto with arith.
Save.
Hint Resolve lub_lift_right.

Lemma lub_lift_left : forall (D:cpo) (f:natO -m> D) n, lub f == lub (mseq_lift_left f n).
intros; apply Ole_antisym; auto.
apply lub_le_compat; intro.
unfold mseq_lift_left; simpl.
apply (fmonotonic f); auto with arith.
Save.
Hint Resolve lub_lift_left.

Lemma lub_le_lift : forall (D:cpo) (f g:natO -m> D) (n:natO), 
        (forall k, n <= k -> f k <= g k) -> lub f <= lub g.
intros; apply lub_le; intros.
apply Ole_trans with (f (n+n0)).
apply (fmonotonic f); simpl; auto with arith.
apply Ole_trans with (g (n+n0)); auto.
apply H; simpl; auto with arith.
Save.

Lemma lub_eq_lift : forall (D:cpo) (f g:natO -m> D) (n:natO), 
        (forall k, n <= k -> f k == g k) -> lub f == lub g.
intros; apply Ole_antisym; apply lub_le_lift with n; intros; auto.
apply Oeq_le_sym; auto.
Save.

(** - (lub_fun h) x = lub_n (h n x) *)
Definition lub_fun : forall (O:ord) (D:cpo) (h : natO -m> O -m> D), O -m> D.
intros; exists (fun m => lub (h <_> m)).
red; intros.
apply lub_le_compat; simpl; intros.
apply (fmonotonic (h x0)); auto.
Defined.

Lemma lub_fun_eq : forall (O:ord) (D:cpo) (h : natO -m> O -m> D) (x:O), 
           lub_fun h x == lub (h <_> x).
auto.
Save.

Lemma lub_fun_shift :  forall (D:cpo) (h : natO -m> (natO -m>  D)),
             lub_fun h == Lub D @ (fmon_shift h).
intros; apply fmon_eq_intro; unfold lub_fun; simpl; auto.
Save.

Lemma double_lub_simpl : forall (D:cpo) (h : natO -m> natO -m>  D),
        lub (Lub D @ h) == lub (fmon_diag h). 
intros; apply Ole_antisym.
apply lub_le; intros; simpl; apply lub_le; intros.
apply Ole_trans with (h n (n+n0)).
apply (fmonotonic (h n)); auto with arith.
apply Ole_trans with (h (n+n0) (n+n0)).
apply (fmonotonic h); auto with arith.
apply le_lub with (f:=fmon_diag h) (n:=(n + n0)%nat).
apply lub_le_compat.
unfold fmon_diag; simpl; auto.
Save.

Lemma lub_exch_le : forall (D:cpo) (h : natO -m> (natO -m>  D)),
                    lub (Lub D @ h) <= lub (lub_fun h).
intros; apply lub_le; intros; simpl.
apply lub_le; intros.
apply Ole_trans with (lub (h n)); auto.
apply lub_le_compat; intro.
unfold lub_fun; simpl.
apply le_lub with (f:=h <_> x) (n:=n).
Save.
Hint Resolve lub_exch_le.

Lemma lub_exch_eq : forall (D:cpo) (h : natO -m> (natO -m>  D)),
 lub (Lub D @ h) == lub (lub_fun h).
intros; apply Ole_antisym; auto.
Save.

Hint Resolve lub_exch_eq.

(** *** Functional cpos *)

Definition fcpo : Type -> cpo -> cpo.
intros A D; exists (ford A D)  
                   (* (fun x:A => (0:D)) *) 
                   (fun f (x:A) => lub(c:=D) (f <o> x)); intros; auto.
intro x; apply Ole_trans with ((f <o> x) n); auto.
Defined.

Infix "-O->" := fcpo (right associativity, at level 30) : O_scope.

Lemma fcpo_lub_simpl : forall A (D:cpo) (h:natO-m> A-O->D)(x:A),
      (lub h) x = lub(c:=D) (h<o> x).
trivial.
Save.

(** ** Continuity *)

Lemma lub_comp_le : 
    forall (D1 D2 : cpo) (f:D1 -m> D2) (h : natO -m> D1),  lub (f @ h) <= f (lub h).
intros; apply lub_le; simpl; intros.
apply (fmonotonic f); auto.
Save.
Hint Resolve lub_comp_le.

Lemma lub_comp2_le : forall (D1 D2 D3: cpo) (F:D1 -m> D2-m>D3) (f : natO -m> D1) (g: natO -m> D2), 
        lub ((F @2 f) g) <= F (lub f)  (lub g).
intros; apply lub_le; simpl; auto.
Save.
Hint Resolve lub_comp2_le.

Definition continuous (D1 D2 : cpo) (f:D1 -m> D2)
                := forall h : natO -m> D1,  f (lub h) <= lub (f @ h).

Lemma continuous_eq_compat : forall (D1 D2 : cpo) (f g:D1 -m> D2), 
                 f==g -> continuous f -> continuous g.
red; intros.
apply Ole_trans with (f (lub h)).
assert (g <= f); auto.
rewrite <- H; auto.
Save. 

Add Parametric Morphism (D1 D2 : cpo) : (@continuous D1 D2)
with signature (@Oeq (D1 -m> D2)) ==> iff as continuous_eq_compat_iff.
split; intros.
apply continuous_eq_compat with x; trivial.
apply continuous_eq_compat with y; auto.
Save.

Lemma lub_comp_eq : 
    forall (D1 D2 : cpo) (f:D1 -m> D2) (h : natO -m> D1),  continuous f -> f (lub h) == lub (f @ h).
intros; apply Ole_antisym; auto.
Save.
Hint Resolve lub_comp_eq.


(*
(** - mon0 x == 0 *)
Definition mon0 (O1:ord) (D2 : cpo) : O1 -m> D2 := fmon_cte O1 (0:D2).

Lemma cont0 : forall (D1 D2 : cpo), continuous (mon0 D1 D2).
red; simpl; auto.
Save.
Implicit Arguments cont0 [].
*)

(** - double_app f g n m = f m (g n) *)
Definition double_app (O1 O2 O3 O4: ord) (f:O1 -m> O3 -m> O4) (g:O2 -m> O3) 
        : O2 -m> (O1 -m> O4) := (fmon_shift f) @ g.
 
(** ** Cpo of monotonic functions *)

Definition fmon_cpo : forall (O:ord) (D:cpo), cpo.
intros; exists (fmon O D) 
               (* (mon0 O D) *)
               (lub_fun (O:=O) (D:=D)); auto.
simpl; intros.
apply le_lub with (f:=fmon_app f x) (n:=n); auto.
Defined.

Infix "-M->" := fmon_cpo (at level 30, right associativity) : O_scope.

Lemma fmon_lub_simpl : forall (O:ord) (D:cpo) (h:natO-m>O-M->D) (x:O),
             (lub h) x = lub (h <_> x).
trivial.
Save.

Lemma double_lub_diag : forall (D:cpo) (h:natO-m>natO-M->D),
             lub (lub h) == lub (fmon_diag h).
intros.
intros; apply Ole_antisym.
apply lub_le; intros; simpl; apply lub_le; intros; simpl.
apply Ole_trans with (h (n+n0) (n+n0)); auto with arith.
apply le_lub with (f:=fmon_diag h) (n:=(n + n0)%nat).
apply lub_le_compat.
unfold fmon_diag; simpl; intros.
apply le_lub with (f:=h <_> x) (n:=x).
Save.



(** *** Continuity *)

Definition continuous2 (D1 D2 D3: cpo) (F:D1 -m> D2 -m> D3)
     := forall (f : natO-m>D1) (g :natO-m>D2), F (lub f) (lub g) <= lub ((F @2 f) g).

Lemma continuous2_app : forall (D1 D2 D3:cpo) (F : D1-m>D2-m>D3),
            continuous2 F -> forall k, continuous (F k).
red; intros.
apply Ole_trans with (F (lub (mseq_cte k))  (lub h)); auto.
apply Ole_trans with (lub ((F @2 (mseq_cte k)) h)); auto.
Save.

Lemma continuous2_continuous : forall (D1 D2 D3:cpo) (F : D1-m>D2-M->D3),
            continuous2 F -> continuous F.
red; intros; intro k.
apply Ole_trans with (F (lub h) (lub (mseq_cte k)) ); auto.
apply Ole_trans with (lub ((F @2 h) (mseq_cte k))); auto.
Save.

Lemma continuous2_left : forall (D1 D2 D3:cpo) (F : D1-m>D2-M->D3) (h:natO-m>D1) (x:D2),
            continuous2 F -> F (lub h) x <= lub ((F <_> x) @h).
intros; apply (continuous2_continuous H h x).
Save.

Lemma continuous2_right : forall (D1 D2 D3:cpo) (F : D1-m>D2-M->D3) (x:D1)(h:natO-m>D2),
            continuous2 F -> F x (lub h) <= lub (F x @h).
intros; apply (continuous2_app H x).
Save.

Lemma continuous_continuous2 : forall (D1 D2 D3:cpo) (F : D1-m>D2-M->D3),
      (forall k:D1, continuous (F k)) -> continuous F -> continuous2 F.
red; intros.
apply Ole_trans with (lub (F (lub f) @ g)); auto.
apply lub_le; simpl; intros.
apply Ole_trans with (lub ((F <_> (g n))@f)).
apply Ole_trans with (lub (c:=D2 -M-> D3) (F@f) (g n)); auto.
rewrite (lub_lift_right ((F @2 f) g) n).
apply lub_le_compat; simpl; intros.
apply Ole_trans with (F (f x) (g (x+n))); auto with arith.
Save.

Hint Resolve continuous2_app continuous2_continuous continuous_continuous2.

Lemma lub_comp2_eq : forall (D1 D2 D3:cpo) (F : D1 -m> D2 -M-> D3),
      (forall k:D1, continuous (F k)) -> continuous F -> 
      forall (f : natO-m>D1) (g :natO-m>D2),
      F (lub f) (lub g) == lub ((F@2 f) g).
intros; apply Ole_antisym; auto.
apply (continuous_continuous2 (F:=F)); trivial.
Save.

Lemma continuous_sym : forall (D1 D2:cpo) (F : D1-m> D1 -M-> D2),
      (forall x y, F x y == F y x) -> (forall k:D1, continuous (F k)) -> continuous F.
red; intros; intro k.
apply Ole_trans with (F k (lub h)); auto.
apply Ole_trans with (lub ((F k) @ h)); auto.
Save.

Lemma continuous2_sym : forall (D1 D2:cpo) (F : D1-m>D1-m>D2),
      (forall x y, F x y == F y x) -> (forall k, continuous (F k)) -> continuous2 F.
intros; apply continuous_continuous2; auto.
apply continuous_sym; auto.
Save.
Hint Resolve continuous2_sym.

(** - continuity is preserved by composition *)

Lemma continuous_comp : forall (D1 D2 D3:cpo) (f:D2-m>D3)(g:D1-m>D2),
             continuous f -> continuous g -> continuous (f@g).
red; intros.
rewrite fmon_comp_simpl.
apply Ole_trans with (f (lub (g@h))); auto.
apply Ole_trans with (lub (f@(g@h))); auto.
Save.
Hint Resolve continuous_comp.

(** ** Cpo of continuous functions *)

Lemma cont_lub : forall (D1 D2 : cpo) (f:natO -m> (D1 -m> D2)),
                                        (forall n, continuous (f n)) ->
                                        continuous  (lub (c:=D1-M->D2) f).
red; intros; simpl.
apply Ole_trans with   
                        (lub (c:=D2) (Lub D2 @ (fmon_shift (double_app f h)))).
apply lub_le_compat; intro n; simpl.
apply Ole_trans with (lub ((f n) @ h)); auto.
rewrite lub_exch_eq.
apply lub_le_compat; intro n; simpl.
apply lub_le_compat; intro m; simpl; auto.
Save.

Record fconti (D1 D2:cpo): Type 
     := mk_fconti {fcontit : D1 -m> D2; fcontinuous : continuous fcontit}.
Hint Resolve fcontinuous.

Definition fconti_fun (D1 D2 :cpo) (f:fconti D1 D2) : D1-> D2 :=fun x => fcontit f x.
Coercion fconti_fun : fconti >-> Funclass.

Definition fcont_ord : cpo -> cpo -> ord.
intros D1 D2; exists (fconti D1 D2) (fun f g => fcontit f <= fcontit g); intros; auto.
apply Ole_trans with (fcontit y); auto.
Defined.

Infix "-c>" := fcont_ord (at level 30, right associativity) : O_scope.

Lemma fcont_le_intro : forall (D1 D2:cpo) (f g : D1 -c> D2), (forall x, f x <= g x) -> f <= g.
trivial.
Save.

Lemma fcont_le_elim : forall (D1 D2:cpo) (f g : D1 -c> D2), f <= g -> forall x, f x <= g x.
trivial.
Save.

Lemma fcont_eq_intro : forall (D1 D2:cpo) (f g : D1 -c> D2), (forall x, f x == g x) -> f == g.
intros; apply Ole_antisym; apply fcont_le_intro; auto.
Save.

Lemma fcont_eq_elim : forall (D1 D2:cpo) (f g : D1 -c> D2), f == g -> forall x, f x == g x.
intros; apply Ole_antisym; apply fcont_le_elim; auto.
Save.

Lemma fcont_monotonic : forall (D1 D2:cpo) (f : D1 -c> D2) (x y : D1),
            x <= y -> f x <= f y.
intros; apply (fmonotonic (fcontit f) H).
Save.
Hint Resolve fcont_monotonic.

Lemma fcont_stable : forall (D1 D2:cpo) (f : D1 -c> D2) (x y : D1),
            x == y -> f x == f y.
intros; apply (fmon_stable (fcontit f) H).
Save.
Hint Resolve fcont_monotonic.

(*
Definition fcont0 (D1 D2:cpo) : D1 -c> D2 := mk_fconti (cont0 D1 D2).
*)

Definition Fcontit (D1 D2:cpo) : (D1 -c> D2) -m> D1-m> D2.
intros D1 D2; exists (fcontit (D1:=D1) (D2:=D2)); auto.
Defined.


Definition fcont_lub (D1 D2:cpo) : (natO -m> D1 -c> D2) -> D1 -c> D2.
intros D1 D2 f; exists (lub (c:=D1-M->D2) (Fcontit D1 D2 @f)).
apply cont_lub; intros; simpl; auto.
Defined.

Definition fcont_cpo : cpo -> cpo -> cpo.
intros D1 D2; exists (fcont_ord D1 D2) 
                     (* (fcont0 D1 D2) *)
                     (fcont_lub (D1:=D1) (D2:=D2)); 
simpl; intros; auto.
apply le_lub with (f:=(Fcontit D1 D2 @ f) <_> x) (n:=n).
Defined.

Infix "-C->" := fcont_cpo (at level 30, right associativity) : O_scope.

Add Parametric Morphism (D E:cpo) : (fun f d => f d)
with signature (@Oeq (D -C-> E)) ==> (@Oeq D) ==> (@Oeq E)
as app_eq_compat.
intros x y xeq a b aeq.
refine (Oeq_trans  _ _).
assert (stable x) by auto. unfold stable in H. apply H. apply aeq. auto.
Qed.

Add Parametric  Morphism (D E:cpo) (f:D -C-> E) : (@fcontit D E)
with signature (@Oeq (D -C-> E)) ==> (@Oeq (D -M-> E))
as fcontit_eq_compat.
intros ; auto.
Qed.

Definition fcont_app (O:ord) (D1 D2:cpo) (f: O -m> D1-c> D2) (x:D1) : O -m> D2
         := Fcontit D1 D2 @ f <_> x.

Infix "<__>" := fcont_app (at level 70) : O_scope.

Lemma fcont_app_simpl : forall (O:ord) (D1 D2:cpo) (f: O -m> D1-c> D2) (x:D1)(y:O),
            (f <__> x) y = f y x.
trivial.
Save.

Definition ford_fcont_shift (A:Type) (D1 D2:cpo) (f: A -o> (D1-c> D2)) : D1 -c> A -O-> D2.
intros; exists (ford_shift (fun x => Fcontit D1 D2 (f x))).
red; intros; intro x.
simpl.
apply Ole_trans with (lub (fcontit (f x) @ h)); auto.
Defined.

Definition fmon_fcont_shift (O:ord) (D1 D2:cpo) (f: O -m> D1-c> D2) : D1 -c> O -M-> D2.
intros; exists (fmon_shift (Fcontit D1 D2@f)).
red; intros; intro x.
simpl.
rewrite (fcontinuous (f x) h); auto.
Defined.

Lemma fcont_app_continuous : 
       forall (O:ord) (D1 D2:cpo) (f: O -m> D1-c> D2) (h:natO-m>D1),
            f <__> (lub h) <= lub (c:=O-M->D2) (fcontit (fmon_fcont_shift f) @ h).
intros; intro x.
rewrite fcont_app_simpl.
rewrite (fcontinuous (f x) h); auto.
Save.

Lemma fcont_lub_simpl : forall (D1 D2:cpo) (h:natO-m>D1-C->D2)(x:D1),
            lub h x = lub (h <__> x).
trivial.
Save.

Definition continuous2_cont_app : forall (D1 D2 D3 :cpo) (f:D1-m> D2 -M-> D3), 
            (forall k, continuous (f k)) -> D1 -m> (D2 -C-> D3).
intros; exists (fun k => mk_fconti (H k)); red; intros; auto.
Defined.

Lemma continuous2_cont_app_simpl : 
forall (D1 D2 D3 :cpo) (f:D1-m> D2 -M-> D3)(H:forall k, continuous (f k))
        (k:D1),  continuous2_cont_app H k = mk_fconti (H k).
trivial.
Save.

Lemma continuous2_cont : forall (D1 D2 D3 :cpo) (f:D1-m> D2 -M-> D3), 
            continuous2 f -> D1 -c> (D2 -C-> D3).
intros; exists (continuous2_cont_app (f:=f) (continuous2_app H)).
red; intros; rewrite continuous2_cont_app_simpl; intro k; simpl.
apply Ole_trans with (lub (c:=D2-M->D3) (f@h) k); auto.
assert (continuous f); auto.
Defined.

Lemma Fcontit_cont : forall D1 D2, continuous (D1:=D1-C->D2) (D2:=D1-M->D2) (Fcontit D1 D2).
red; intros; auto.
Save.
Hint Resolve Fcontit_cont.


Definition fcont_comp : forall (D1 D2 D3:cpo), (D2 -c> D3) -> (D1-c> D2) -> D1 -c> D3.
intros D1 D2 D3 f g.
exists (fcontit f @ fcontit g); auto.
Defined.

Infix "@_" := fcont_comp (at level 35) : O_scope.

Lemma fcont_comp_assoc: forall (D E F G:cpo) (f:F -C-> G) (g:E -C-> F) (h:D -C-> E),
          (f @_ g) @_ h == f @_ (g @_ h).
intros D E F G f g h. apply fcont_eq_intro. auto.
Qed.

Lemma fcont_comp_simpl : forall (D1 D2 D3:cpo)(f:D2 -c> D3)(g:D1-c> D2) (x:D1),
       (f @_ g) x = f (g x).
trivial.
Save.

Lemma fcontit_comp_simpl : forall (D1 D2 D3:cpo)(f:D2 -c> D3)(g:D1-c> D2) (x:D1),
       fcontit (f @_ g) = fcontit f @ fcontit g.
trivial.
Save.

Lemma fcont_comp_le_compat : forall (D1 D2 D3:cpo) (f g : D2 -c> D3) (k l :D1-c> D2),
      f <= g -> k <= l -> f @_ k <= g @_ l.
intros; apply fcont_le_intro; intro x.
repeat (rewrite fcont_comp_simpl).
apply Ole_trans with (g (k x)); auto.
Save.
Hint Resolve fcont_comp_le_compat.

Add Parametric Morphism (D1 D2 D3 : cpo) : (@fcont_comp D1 D2 D3)
with signature (@Ole (D2 -c> D3)) ++> (@Ole (D1 -c> D2)) ++> (@Ole (D1 -c> D3)) 
as fcont_comp_le_morph.
intros.
exact (fcont_comp_le_compat H H0).
Save.

Add Parametric Morphism (D1 D2 D3 : cpo) : (@fcont_comp D1 D2 D3)
with signature (@Oeq (D2 -c> D3)) ==> (@Oeq (D1 -c> D2)) ==> (@Oeq (D1 -c> D3))
as fcont_comp_eq_compat.
intros.
apply Ole_antisym; auto.
Save.

Definition fcont_Comp (D1 D2 D3:cpo) : (D2 -C-> D3) -m> (D1-C-> D2) -m> D1 -C-> D3 
      := le_compat2_mon (fcont_comp_le_compat (D1:=D1) (D2:=D2) (D3:=D3)).

Lemma fcont_Comp_simpl : forall (D1 D2 D3:cpo) (f:D2 -c> D3) (g:D1-c> D2),
               fcont_Comp D1 D2 D3 f g = f @_ g.
trivial.
Save.

Lemma fcont_Comp_continuous2 : forall (D1 D2 D3:cpo), continuous2 (fcont_Comp D1 D2 D3).
red; intros.
change ((lub  f) @_ (lub g) <= lub (c:=D1 -C-> D3) ((fcont_Comp D1 D2 D3 @2 f) g)).
apply fcont_le_intro; intro x; rewrite fcont_comp_simpl.
repeat (rewrite fcont_lub_simpl).
rewrite fcont_app_continuous.
rewrite double_lub_diag.
apply lub_le_compat; simpl; auto.
Save.

Definition fcont_COMP  (D1 D2 D3:cpo) : (D2 -C-> D3) -c> (D1-C-> D2) -C-> D1 -C-> D3 
      := continuous2_cont (fcont_Comp_continuous2 (D1:=D1) (D2:=D2) (D3:=D3)).

Lemma fcont_COMP_simpl : forall (D1 D2 D3:cpo) (f: D2 -C-> D3) (g:D1-C-> D2),
	fcont_COMP D1 D2 D3 f g = f @_ g.
trivial.
Save.

Definition fcont2_COMP  (D1 D2 D3 D4:cpo) : (D3 -C-> D4) -c> (D1-C-> D2-C->D3) -C-> D1 -C-> D2 -C-> D4 := 
   (fcont_COMP D1 (D2-C->D3) (D2 -C-> D4)) @_ (fcont_COMP D2 D3 D4).

Definition fcont2_comp  (D1 D2 D3 D4:cpo) (f:D3 -C-> D4)(F:D1-C-> D2-C->D3) := fcont2_COMP D1 D2 D3 D4 f F.

Infix "@@_" := fcont2_comp (at level 35) : O_scope.

Lemma fcont2_comp_simpl : forall (D1 D2 D3 D4:cpo) (f:D3 -C-> D4)(F:D1-C-> D2-C->D3)(x:D1)(y:D2),
       (f @@_ F) x y = f (F x y).
trivial.
Save.
            
Lemma fcont_le_compat2 : forall (D1 D2 D3:cpo) (f : D1-c>D2-C->D3)
    (x y : D1) (z t : D2), x <= y -> z <= t -> f x z <= f y t.
intros; apply Ole_trans with (f y z); unfold fconti_fun; auto.
apply fmon_le_elim.
apply  (fmonotonic (Fcontit D2 D3) (x:=fcontit f x) (y:=fcontit f y)); auto.
Save.
Hint Resolve fcont_le_compat2.

Lemma fcont_eq_compat2 : forall (D1 D2 D3:cpo) (f : D1-c>D2-C->D3)
    (x y : D1) (z t : D2), x == y -> z == t -> f x z == f y t.
intros; apply Ole_antisym; auto.
Save.
Hint Resolve fcont_eq_compat2.

Lemma fcont_continuous : forall (D1 D2 : cpo) (f:D1 -c> D2)(h:natO-m>D1),
            f (lub h) <= lub (fcontit f @ h).
intros; apply (fcontinuous f h).
Save.
Hint Resolve fcont_continuous.

Lemma fcont_continuous2 : forall (D1 D2 D3:cpo) (f:D1-c>(D2-C->D3)), 
                                             continuous2 (Fcontit D2 D3 @ fcontit f).
intros; apply continuous_continuous2; intros.
change (continuous (fcontit (f k))); auto.
apply (continuous_comp (D1:=D1) (D2:=D2-C->D3) (D3:=D2-M->D3) (f:=Fcontit D2 D3)); auto.
Save.
Hint Resolve fcont_continuous2.

Definition fcont_shift (D1 D2 D3 : cpo) (f:D1-c>D2-C->D3) : D2-c>D1-C->D3.
intros.
apply continuous2_cont with (f:=fmon_shift (Fcontit D2 D3 @ fcontit f)).
red; intros; repeat rewrite fmon_shift_simpl.
repeat rewrite fmon_comp_simpl; auto.
change (f (lub g) (lub f0) <= lub ((fmon_shift (Fcontit D2 D3 @ fcontit f) @2 f0) g)).
apply Ole_trans with (1:=fcont_continuous2 f g f0).
apply lub_le_compat; intro n.
repeat rewrite fmon_comp_simpl; auto.
Defined.

Lemma fcont_shift_simpl : forall (D1 D2 D3 : cpo) (f:D1-c>D2-C->D3) (x:D2) (y:D1),
            fcont_shift f x y = f y x.
trivial.
Save.

Definition fcont_SEQ  (D1 D2 D3:cpo) : (D1-C-> D2) -C-> (D2 -C-> D3) -C-> D1 -C-> D3 
      := fcont_shift (fcont_COMP D1 D2 D3).

Lemma fcont_SEQ_simpl : forall (D1 D2 D3:cpo) (f: D1 -C-> D2) (g:D2-C-> D3),
	fcont_SEQ D1 D2 D3 f g = g @_ f.
trivial.
Save.

Definition fcont_comp2 : forall (D1 D2 D3 D4:cpo), 
                (D2 -c> D3 -C->D4) -> (D1-c> D2) -> (D1 -c> D3) -> D1-c>D4.
intros D1 D2 D3 D4 f g h.
exists (((Fcontit D3 D4 @fcontit f) @2 fcontit g) (fcontit h)).
red; intros; simpl.
change (f (g (lub h0)) (h (lub h0)) <= lub (c:=D4) ((Fcontit D3 D4 @ fcontit f @2 fcontit g) (fcontit h) @ h0)).
apply Ole_trans with (f (lub (c:=D2) (fcontit g @ h0)) (lub (c:=D3) (fcontit h @ h0))); auto.
apply (fcont_continuous2 f (fcontit g @ h0) (fcontit h @ h0)).
Defined.

Infix "@2_" := fcont_comp2 (at level 35, right associativity) : O_scope.

Lemma fcont_comp2_simpl : forall (D1 D2 D3 D4:cpo)
                (F:D2 -c> D3 -C->D4) (f:D1-c> D2) (g:D1 -c> D3) (x:D1), (F@2_ f) g x = F (f x) (g x).
trivial.
Save.

Add Parametric Morphism (D1 D2 D3 D4:cpo) : (@fcont_comp2 D1 D2 D3 D4)
with signature (@Ole (D2 -c> D3 -C->D4)) ++> (@Ole (D1-c> D2)) ++> (@Ole (D1 -c> D3)) ++> (@Ole (D1-c>D4))
as fcont_comp2_le_morph.
intros F G HF f1 f2 Hf g1 g2 Hg x.
apply Ole_trans with (fcontit (fcontit G (fcontit f1 x)) (fcontit g1 x)); auto.

(* auto.*)
Focus 2.
change (Fcontit D3 D4 (fcontit G (fcontit f1 x)) (fcontit g1 x) <=
              Fcontit D3 D4 (fcontit G (fcontit f2 x)) (fcontit g2 x)).
apply (fmon_le_compat2 (Fcontit D3 D4)); auto.
simpl.
auto.
(* not entirely sure I've got the right signature here
   auto was supposed to deal with first sibgoal but doesn't
   I want to just setoid rewrite with HF
   but that doesn't work either
*)
unfold Ole in HF.
simpl in HF.
apply HF.
Save.

Add Parametric Morphism (D1 D2 D3 D4:cpo) : (@fcont_comp2 D1 D2 D3 D4)
with signature (@Oeq (D2 -c> D3 -C->D4)) ==> (@Oeq (D1-c> D2)) ==> 
   (@Oeq (D1 -c> D3)) ==> (@Oeq (D1-c>D4))
as fcont_comp2_eq_compat.
intros F G (HF1,HF2) f1 f2 (Hf1,Hf2) g1 g2 (Hg1,Hg2).
apply Ole_antisym.
exact (fcont_comp2_le_morph HF1 Hf1 Hg1).
exact (fcont_comp2_le_morph HF2 Hf2 Hg2).
Save.

Lemma fcont_comp2_comp: forall (D E F G H:cpo) (f:D -C-> E -C-> F) (g:H -C-> D) (h: H -C-> E) (a:G -C-> H),
         (f @2_ g) h @_ a == (f @2_ (g @_ a)) (h @_ a).
intros. apply fcont_eq_intro.
intros gg. auto.
Qed.

(** - Identity function is continuous *)

Definition Id : forall O:ord, O-m>O.
intros; exists (fun x => x); red; auto.
Defined.

Definition ID : forall D:cpo, D-c>D.
intros; exists (Id D); red; auto.
Defined.

Lemma fcont_comp_unitR: forall D E (f:D -C-> E), f @_ (ID _) == f.
intros D E f. apply fcont_eq_intro. auto.
Qed.

Lemma fcont_comp_unitL: forall D E (f:D -C-> E), (ID _) @_ f == f.
intros. apply fcont_eq_intro. auto.
Qed.

Lemma Id_simpl : forall O x, Id O x = x.
trivial.
Save.

Lemma ID_simpl : forall D x, ID D x = Id D x.
trivial.
Save.

Definition AP (D1 D2:cpo) : (D1-C->D2)-c>D1-C->D2:=ID (D1-C->D2).

Lemma AP_simpl : forall (D1 D2:cpo) (f : D1-C->D2) (x:D1), AP D1 D2 f x = f x.
trivial.
Save.

Definition fcont_comp3 (D1 D2 D3 D4 D5:cpo)
                (F:D2 -c> D3 -C->D4-C->D5)(f:D1-c> D2)(g:D1 -c> D3)(h:D1-c>D4): D1-c>D5
  := (AP D4 D5 @2_ ((F @2_ f) g)) h.

Infix "@3_" := fcont_comp3 (at level 35, right associativity) : O_scope.

Lemma fcont_comp3_simpl : forall (D1 D2 D3 D4 D5:cpo)
                (F:D2 -c> D3 -C->D4-C->D5) (f:D1-c> D2) (g:D1 -c> D3) (h:D1-c>D4) (x:D1), 
                (F@3_ f) g h x = F (f x) (g x) (h x).
trivial.
Save.

Add Parametric Morphism (D E F G H:cpo) : (@fcont_comp3 D E F G H)
with signature (@Oeq (E -C-> F -C-> G -C-> H)) ==> (@Oeq (D -C-> E)) ==> (@Oeq (D -C-> F)) ==> (@Oeq (D -C-> G)) ==> (@Oeq (D -C-> H))
as fcont_comp3_eq_compat.
intros h0 h1 heq e0 e1 eeq f0 f1 feq g0 g1 geq.
simpl. apply fcont_eq_intro. intros d.
repeat (rewrite fcont_comp3_simpl).
refine (app_eq_compat _ _).
refine (app_eq_compat _ _).
refine (app_eq_compat _ _).
apply heq.
apply (fcont_eq_elim eeq d).
apply (fcont_eq_elim feq d).
apply (fcont_eq_elim geq d).
Qed.

Lemma fcont_comp3_comp: forall (D E F G H A:cpo) (f:D -C-> E -C-> F -C-> G) (g:H -C-> D)
           (h:H -C-> E) (k:H -C-> F) (a:A -C-> H),
        (f @3_ g) h k @_ a == (f @3_ (g @_ a)) (h @_ a) (k @_ a).
intros. refine (fcont_eq_intro _). auto.
Qed.

(** ** Product of two cpos *)

Definition Oprod : ord -> ord -> ord.
intros O1 O2; exists (O1 * O2)%type (fun x y => fst x <= fst y /\ snd x <= snd y); intuition.
apply Ole_trans with a0; trivial.
apply Ole_trans with b0; trivial.
Defined.

Definition Fst (O1 O2 : ord) : Oprod O1 O2 -m> O1.
intros O1 O2; exists (fst (A:=O1) (B:=O2)); red; simpl; intuition.
Defined.

Definition Snd (O1 O2 : ord) : Oprod O1 O2 -m> O2.
intros O1 O2; exists (snd (A:=O1) (B:=O2)); red; simpl; intuition.
Defined.

Definition Pairr (O1 O2 : ord) : O1 -> O2 -m> Oprod O1 O2.
intros O1 O2 x.
exists (fun (y:O2) => (x,y)).
red; auto.
Defined.

Definition Pair (O1 O2 : ord) : O1 -m> O2 -m> Oprod O1 O2.
intros O1 O2; exists (Pairr (O1:=O1) O2); red; auto.
Defined.

Lemma Fst_simpl : forall (O1 O2 : ord) (p:Oprod O1 O2), Fst O1 O2 p = fst p.
trivial.
Save.

Lemma Snd_simpl : forall (O1 O2 : ord) (p:Oprod O1 O2), Snd O1 O2 p = snd p.
trivial.
Save.

Lemma Pair_simpl : forall (O1 O2 : ord) (x:O1)(y:O2), Pair O1 O2 x y = (x,y).
trivial.
Save.

Lemma Oprod_unique: forall O1 O2 O3 (f g:O1 -m> Oprod O2 O3),
      Fst _ _ @ f == Fst _ _ @ g -> Snd _ _ @ f == Snd _ _ @ g -> f == g.
intros O1 O2 O3 f g ff ss.
apply fmon_eq_intro. intros x.
split ; auto using (fmon_eq_elim ff x), (fmon_eq_elim ss x).
Qed.

Definition Osum: ord -> ord -> ord.
intros O1 O2. exists ((O1 + O2)%type) 
     (fun x y => match x, y with | inl x, inl y => x <= y | inr x, inr y => x <= y | _,_ =>  False end).
intros x. case x ; intros t ; apply Ole_refl.
intros x y z. case x ; clear x ; intros x.
  case y ; clear y ; intros y ; try (intros ; intuition ; fail).
  case z ; clear z ; intros z ; try (intros ; intuition ; fail). apply Ole_trans.
case y ; clear y ; intros y ; try (intros ; intuition ; fail).
case z ; clear z ; intros z ; try (intros ; intuition ; fail). apply Ole_trans.
Defined.

Definition Inl (O1 O2:ord) : O1 -m> (Osum O1 O2).
intros. exists (fun x => @inl O1 O2 x). unfold monotonic.
intros. auto.
Defined.

Definition Inr (O1 O2 :ord) : O2 -m> (Osum O1 O2).
intros. exists (fun x => @inr O1 O2 x). unfold monotonic. auto.
Defined.

(*
Definition prod0 (D1 D2:cpo) : Oprod D1 D2 := (0: D1,0: D2).
*)

Definition prod_lub (D1 D2:cpo) (f : natO -m> Oprod D1 D2) := (lub (Fst D1 D2@f), lub (Snd D1 D2@f)).

Definition Dprod : cpo -> cpo -> cpo.
intros D1 D2; exists (Oprod D1 D2) 
                     (* (prod0 D1 D2) *)
                     (prod_lub (D1:=D1) (D2:=D2)); unfold prod_lub; intuition.
apply Ole_trans with (fst (fmonot f n), snd (fmonot f n)); simpl; intuition.
apply le_lub with (f:=Fst D1 D2 @ f) (n:=n).
apply le_lub with (f:=Snd D1 D2 @ f) (n:=n).
apply Ole_trans with (fst x, snd x); simpl; intuition.
apply lub_le; simpl; intros.
case (H n); auto.
apply lub_le; simpl; intros.
case (H n); auto.
Defined.

Lemma Dprod_eq_intro : forall (D1 D2:cpo) (p1 p2: Dprod D1 D2),
             fst p1 == fst p2 -> snd p1 == snd p2 -> p1 == p2.
split; simpl; auto.  
Save.
Hint Resolve Dprod_eq_intro.

Lemma Dprod_eq_pair : forall (D1 D2:cpo) (x1 y1:D1) (x2 y2:D2),
             x1==y1 -> x2==y2 -> ((x1,x2):Dprod D1 D2) == (y1,y2).
auto.  
Save.
Hint Resolve Dprod_eq_pair.

Lemma Dprod_eq_elim_fst : forall (D1 D2:cpo) (p1 p2: Dprod D1 D2),
             p1==p2 -> fst p1 == fst p2.
split; case H; simpl; intuition.  
Save.
Hint Immediate Dprod_eq_elim_fst.

Lemma Dprod_eq_elim_snd : forall (D1 D2:cpo) (p1 p2: Dprod D1 D2),
             p1==p2 -> snd p1 == snd p2.
split; case H; simpl; intuition.  
Save.
Hint Immediate Dprod_eq_elim_snd.

Lemma Dprod_le_elim_snd: forall (D1 D2:cpo) (p1 p2 : Dprod D1 D2),
        p1 <= p2 -> snd p1 <= snd p2.
simpl. intros D1 D2 p1 p2. intros x. destruct x. auto.
Qed.

Hint Immediate Dprod_le_elim_snd.

Lemma Dprod_le_elim_fst: forall (D1 D2:cpo) (p1 p2:Dprod D1 D2),
       p1 <= p2 -> fst p1 <= fst p2.
simpl. intros D1 D2 p1 p2 x. destruct x. auto.
Qed.

Hint Immediate Dprod_le_elim_fst.

Definition FST (D1 D2:cpo) : Dprod D1 D2 -c> D1.
intros; exists (Fst D1 D2); red; intros; auto.
Defined.

Definition SND (D1 D2:cpo) : Dprod D1 D2 -c> D2.
intros; exists (Snd D1 D2); red; intros; auto.
Defined.

Lemma Pair_continuous2 : forall (D1 D2:cpo), continuous2 (D3:=Dprod D1 D2) (Pair D1 D2).
red; intros; auto.
Save.

Definition PAIR (D1 D2:cpo) : D1 -c> D2 -C-> Dprod D1 D2 
                := continuous2_cont (Pair_continuous2 (D1:=D1) (D2:=D2)).

Add Parametric Morphism (D E:cpo) : (@pair D E)
with signature (@Oeq D) ==> (@Oeq E) ==> (@Oeq (Dprod D E))
as PAIR_eq_compat.
intros d0 d1 deq e0 e1 eeq.
destruct deq ; destruct eeq ; split ; auto.
Qed.

Lemma Dprod_unique: forall D E F (f g:F -C-> Dprod D E), FST _ _ @_ f == FST _ _ @_ g ->
               SND _ _ @_ f == SND _ _ @_ g -> f == g.
Proof.
intros D E F f g ff ss.
refine (Oprod_unique ff ss).
Qed.

Lemma FST_simpl : forall (D1 D2 :cpo) (p:Dprod D1 D2), FST D1 D2 p = Fst D1 D2 p.
trivial.
Save.

Lemma SND_simpl : forall (D1 D2 :cpo) (p:Dprod D1 D2), SND D1 D2 p = Snd D1 D2 p.
trivial.
Save.

Lemma PAIR_simpl : forall (D1 D2 :cpo) (p1:D1) (p2:D2), PAIR D1 D2 p1 p2 = Pair D1 D2 p1 p2.
trivial.
Save.

Lemma FST_PAIR_simpl : forall (D1 D2 :cpo) (p1:D1) (p2:D2), 
            FST D1 D2 (PAIR D1 D2 p1 p2) = p1.
trivial.
Save.

Lemma SND_PAIR_simpl : forall (D1 D2 :cpo) (p1:D1) (p2:D2), 
            SND D1 D2 (PAIR D1 D2 p1 p2) = p2.
trivial.
Save.

Definition Prod_map :  forall (D1 D2 D3 D4:ord)(f:D1-m>D3)(g:D2-m>D4) , 
      Oprod D1 D2 -m> Oprod D3 D4.
intros; exists (fun p => pair (f (fst p)) (g (snd p))); red; intros.
split; simpl fst; simpl snd; apply fmonotonic.
apply (fmonotonic (Fst D1 D2) H).
apply (fmonotonic (Snd D1 D2) H).
Defined.

Lemma Prod_map_simpl : forall (D1 D2 D3 D4:ord)(f:D1-m>D3)(g:D2-m>D4) (p:Oprod D1 D2),
      Prod_map f g p =  pair (f (fst p)) (g (snd p)).
trivial.
Save.

Definition PROD_map :  forall (D1 D2 D3 D4:cpo)(f:D1-c>D3)(g:D2-c>D4) , 
      Dprod D1 D2 -c> Dprod D3 D4.
intros; exists (Prod_map (fcontit f) (fcontit g)); red; intros ; rewrite (Prod_map_simpl (fcontit f) (fcontit g)).
split; simpl.
apply Ole_trans with (f (lub (Fst D1 D2 @ h))); trivial.
rewrite (fcont_continuous f).
apply lub_le_compat; intros; intro; simpl; auto.
apply Ole_trans with (g (lub (Snd D1 D2 @ h))); trivial.
rewrite (fcont_continuous g).
apply lub_le_compat; intros; intro; simpl; auto.
Defined.

Add Parametric Morphism (D E F G:cpo) : (@PROD_map D E F G)
with signature (@Oeq (D -C-> F)) ==> (@Oeq (E -C-> G)) ==> (@Oeq (@Dprod D E -C-> @Dprod F G))
as PROD_map_eq_compat.
intros f g feq s r seq.
simpl. apply fcont_eq_intro.
intros x. case x. clear x. intros d e.
assert (fd:=fcont_eq_elim feq d). assert (fe := fcont_eq_elim seq e).
auto.
Qed.

Lemma PROD_map_simpl :  forall (D1 D2 D3 D4:cpo)(f:D1-c>D3)(g:D2-c>D4)(p:Dprod D1 D2), 
      PROD_map f g p = pair (f (fst p)) (g (snd p)).
trivial.
Save.

Definition PROD_Map1 (D E F G:cpo) : (D -C-> F) -> (E -C-> G) -M-> Dprod D E -C-> Dprod F G.
intros. exists (@PROD_map D E F G X). unfold monotonic.
intros f0 f1 leq. intros x. simpl ; auto.
Defined.

Definition PROD_Map (D E F G:cpo) : (D -C-> F) -M-> (E -C-> G) -M-> Dprod D E -C-> Dprod F G.
intros. exists (@PROD_Map1 D E F G). unfold monotonic.
intros f0 f1 leq. intros x. simpl ; auto.
Defined.

Lemma PROD_Map_continuous2 : forall D E F G, continuous2 (PROD_Map D E F G).
intros. unfold continuous2. intros c1 c2.
intros p. auto.
Save.

Definition PROD_MAP (D E F G:cpo) : (D -C-> F) -C-> (E -C-> G) -C-> Dprod D E -C-> Dprod F G :=
  (continuous2_cont (@PROD_Map_continuous2 D E F G)).

Definition duplicatem (a:cpo) : a -m> Dprod a a.
intros D; exists (fun x => PAIR D D x x).
auto.
Defined.

Definition duplicate (a:cpo) : a-c> Dprod a a.
intros D; exists (duplicatem D).
red.
intro.
auto.
Defined.

Lemma duplicate_simpl: forall (D:cpo) (e:D), duplicate D e = (e,e).
auto.
Qed.

Definition PROD_fun A B C f g:= (@PROD_map C C A B f g) @_ (duplicate C).

Definition SWAP D E := PROD_fun (SND D E) (FST D E).

Lemma PROD_fun_simpl: forall (D E F:cpo) (f:D -C-> E) (g:D -C-> F) d,
           PROD_fun f g d == PAIR _ _ (f d) (g d).
auto.
Qed.

Add Parametric Morphism (D E F:cpo) : (@PROD_fun E F D)
with signature (@Oeq (D -C-> E)) ==> (@Oeq (D -C-> F)) ==> (@Oeq (D -C-> Dprod E F))
as PROD_fun_eq_compat.
intros f0 f1 feq g0 g1 geq.
refine (fcont_eq_intro _).
intro d.
assert (gg:=fcont_eq_elim geq d). assert (ff := fcont_eq_elim feq d).
auto.
Qed.

Lemma PROD_map_fun: forall A B C D f g, @PROD_map A B C D f g == PROD_fun (f @_ FST _ _) (g @_ SND _ _).
intros. unfold PROD_fun. apply fcont_eq_intro. intros p. case p.
intros a b. auto.
Qed.

Definition sum_left (D1 D2:cpo) (c:natO -m> Osum D1 D2) (ex:{d| c O = inl D2 d}) : (natO -m> D1).
intros D1 D2 c ex.
destruct ex as [d ex].
exists (fun x => match c x with | inl y => y | inr _ => d end).
unfold monotonic.
intros n m lm. assert (c n <= c m). auto. generalize H. clear H.
assert ((O <= n) % nat) as no by omega. assert (c O <= c n) as con by auto. rewrite ex in con. clear ex no.
generalize con. clear con.
case (c n). intros d1 _. case (c m). intros d1'. auto.
intros d2' incon. inversion incon. intros d1' incon. inversion incon.
Defined.

Definition sum_right (D1 D2:cpo) (c:natO -m> Osum D1 D2) (ex:{d | c O = inr D1 d}) : (natO -m> D2).
intros D1 D2 c ex.
destruct ex as [d ex].
exists (fun x => match c x with | inr y => y | inl _ => d end).
unfold monotonic.
intros n m lm. assert (c n <= c m). auto. generalize H. clear H.
assert ((O <= n) % nat) as no by omega. assert (c O <= c n) as con by auto. rewrite ex in con. clear ex no.
generalize con. clear con.
case (c n). intros d1 _. case (c m). intros d1'. auto.
intros d2' incon. inversion incon. intros d1' _. case (c m). intros tt incon. inversion incon.
intros d2' l. auto.
Defined.

Definition SuminjProof: forall (D1 D2:cpo) (c:natO -m> Osum D1 D2) (n:nat),
      {d | c n = inl D2 d} + {d | c n = inr D1 d}.
intros D1 D2 c n.
assert ((0 <= n)%nat). omega.
assert (c 0 <= c n). auto. clear H.
case_eq (c 0).
intros d1 c0.
case_eq (c n). intros dn cnd.
left.
exists dn. auto.
intros d2 incon. rewrite incon in H0. rewrite c0 in H0. inversion H0.

intros d2 c0. right.
case_eq (c n). intros dn cnd.
rewrite cnd in H0. rewrite c0 in H0. inversion H0.
intros dn cn. exists dn. auto.
Defined.

Definition sum_lub (D1 D2:cpo) (c: natO -m> Osum D1 D2) : (Osum D1 D2).
intros D1 D2 c. case (SuminjProof c 0).
intros X. refine (inl D2 (lub (sum_left X))).
intro X. refine (inr D1 (lub (sum_right X))).
Defined.

Definition Dsum : cpo -> cpo -> cpo.
intros D1 D2. exists (Osum D1 D2) (@sum_lub D1 D2).
intros c n.
case_eq (c 0).
intros d1 c0.
simpl. case_eq (c n). intros dn cn.
unfold sum_lub. destruct (SuminjProof c 0) as [inj|inj].
assert (sum_left inj n = dn).
destruct inj. simpl. rewrite cn. auto.
rewrite <- H.
apply le_lub.

destruct inj as [dd c0n]. rewrite c0n in c0. inversion c0.
intros d2 cn.
assert ((0 <= n) % nat) as A by omega.
assert (c 0 <= c n) by auto. rewrite c0 in H. rewrite cn in H. inversion H.

intros d2 c0. simpl. case_eq (c n). intros dn cn.
unfold sum_lub. destruct (SuminjProof c 0) as [[d inj] | inj].
rewrite inj in c0. inversion c0.
assert ((0 <= n) % nat) as A by omega. assert (c 0 <= c n) by auto.
rewrite c0 in H. rewrite cn in H. inversion H.
intros dn cn. unfold sum_lub. destruct (SuminjProof c 0) as [[dd inj]|[d inj]].
rewrite c0 in inj. inversion inj.
assert (sum_right (exist (fun dd => c 0 = inr D1 dd) d inj) n = dn).
simpl. rewrite cn. auto.
rewrite <- H.
apply (le_lub).

intros c x. case x ; clear x ; intros x cn. unfold sum_lub.
case (SuminjProof c 0). intros s.
assert (lub (sum_left s) <= x).
apply (lub_le). intros n. specialize (cn n). destruct s as [d c0].
simpl. case_eq (c n). intros dn dnl. rewrite dnl in cn. auto.
intros dn cnr. rewrite cnr in cn. inversion cn. auto.
intros s. destruct s. simpl. specialize (cn 0).
rewrite e in cn. inversion cn.

unfold sum_lub. case (SuminjProof c 0). intros s.
destruct s. simpl. specialize (cn 0). rewrite e in cn. inversion cn.
intros s. assert (lub (sum_right s) <= x).
apply lub_le. intros n. specialize (cn n). destruct s as [d c0].
simpl. case_eq (c n). intros dl cnl. rewrite cnl in cn. inversion cn.
intros dl dll. rewrite dll in cn. auto. auto.
Defined.

Definition Sum_fun : forall (D1 D2 D3:cpo) (f:D1 -m> D3) (g:D2-m> D3),
                      Dsum D1 D2 -m> D3.
intros. exists (fun p => match p with inl y => (f y) | inr y => (g y) end).
unfold monotonic. intros x y.
case_eq x. intros xx _. clear x. case_eq y. intros yy _ ll. auto.
intros yy _ incon. inversion incon.
intros xx _. case_eq y. intros yy _ incon. inversion incon.
intros yy _ ll. auto.
Defined.

Definition SUM_fun (D1 D2 D3:cpo)(f:D1 -c> D3) (g:D2 -c> D3) :
       Dsum D1 D2 -c> D3.
intros. exists (Sum_fun (fcontit f) (fcontit g)).
unfold continuous. intros c.
simpl. case_eq (sum_lub c).
intros ll. intros sl. unfold sum_lub in sl.
destruct (SuminjProof c 0).
inversion sl. clear sl.
destruct s as [dd c0]. assert (continuous (fcontit f)) as fC by auto.
unfold continuous in fC. rewrite fC.
apply lub_le_compat. simpl. intros n.
case_eq (c n). intros t cn. simpl. simpl in cn. rewrite cn. auto.
intros t cn. simpl. simpl in cn. rewrite cn.
assert (c 0 <= c n). assert (0 <= n) % nat by omega. auto.
simpl in c0. simpl in H. rewrite c0 in H. rewrite cn in H. inversion H.
inversion sl.

intros ll. intros sl. unfold sum_lub in sl. destruct (SuminjProof c 0).
inversion sl. inversion sl. clear sl. destruct s as [dd c0].
assert (continuous (fcontit g)) as gC by auto.
unfold continuous in gC. rewrite gC. apply lub_le_compat.
simpl. intros n.
simpl. case_eq (c n). intros t cn. simpl. simpl in cn. rewrite cn.
assert (c 0 <= c n). assert (0 <= n)%nat as A by omega. auto.
simpl in H. rewrite c0 in H. rewrite cn in H. inversion H.
intros t cn. simpl in cn. rewrite cn. auto.
Defined.

Definition INL (D1 D2:cpo) : D1 -c> Dsum D1 D2.
intros. exists (@Inl D1 D2). unfold continuous.
intros c. simpl. apply lub_le_compat.
simpl. auto.
Defined.

Definition INR (D1 D2:cpo) : D2 -c> Dsum D1 D2.
intros. exists (@Inr D1 D2). unfold continuous. intros c. simpl. apply lub_le_compat.
auto.
Defined.

Definition SUM_map D1 D2 D3 D4 (f:D1 -c> D3) (g:D2 -c> D4) := 
 SUM_fun (INL D3 D4 @_ f) (INR _ _ @_ g).

Definition SUM_Map1 (D E F G:cpo) : (D -C-> F) -> (E -C-> G) -M-> Dsum D E -C-> Dsum F G.
intros. exists (@SUM_map D E F G X).
unfold monotonic. intros x y xyl. intros xx ; case xx ; auto.
Defined.

Definition SUM_Map (D E F G:cpo) : (D -C-> F) -M-> (E -C-> G) -M-> Dsum D E -C-> Dsum F G.
intros. exists (@SUM_Map1 D E F G). unfold monotonic.
intros x y xyl. intros xx yy ; case xx ; case yy ; simpl ; auto.
Defined.

Lemma SUM_Map_continuous2 : forall D E F G, continuous2 (SUM_Map D E F G).
intros. unfold continuous2. intros c1 c2.
intros p. case p ; auto.
Save.

Definition SUM_MAP (D E F G:cpo) : (D -C-> F) -C-> (E -C-> G) -C-> Dsum D E -C-> Dsum F G :=
  (continuous2_cont (@SUM_Map_continuous2 D E F G)).

Lemma SUM_map_simpl (D E F G:cpo) f g : SUM_MAP D E F G f g = SUM_map f g.
intros. auto.
Qed.

Lemma SUM_fun_simpl (D E F:cpo) f g s : @SUM_fun D E F f g s = Sum_fun (fcontit f) (fcontit g) s.
intros. auto.
Qed.

Lemma Dsum_lcom: forall D1 D2 D3 f g, @SUM_fun D1 D2 D3 f g @_ INL D1 D2 == f.
intros. apply fcont_eq_intro. intros s. auto.
Qed.

Lemma Dsum_rcom: forall D1 D2 D3 f g, @SUM_fun D1 D2 D3 f g @_ INR _ _ == g.
intros. apply fcont_eq_intro. auto.
Qed.

Lemma Dsum_unique: forall D1 D2 D3 f g h, h @_ INL _ _ == f -> h @_ INR _ _ == g ->
            h == @SUM_fun D1 D2 D3 f g.
intros. apply fcont_eq_intro. intros x.
case x. clear x. intros t.
assert (xx:= fcont_eq_elim H t). auto.
intros t. assert (xx:= fcont_eq_elim H0 t). auto.
Qed.

Definition curry (D1 D2 D3 : cpo) (f:Dprod D1 D2 -c> D3) : D1 -c> (D2-C->D3) :=
fcont_COMP D1 (D2-C->Dprod D1 D2) (D2-C->D3) 
                          (fcont_COMP D2 (Dprod D1 D2) D3 f) (PAIR D1 D2).

Definition Curry : forall (D1 D2 D3 : cpo), (Dprod D1 D2 -c> D3) -m> D1 -c> (D2-C->D3).
       intros; exists (curry (D1:=D1)(D2:=D2)(D3:=D3)); red; intros; auto.
Defined.

Lemma Curry_simpl : forall (D1 D2 D3 : cpo) (f:Dprod D1 D2 -C-> D3) (x:D1) (y:D2), 
       Curry D1 D2 D3 f x y = f (x,y).
trivial.
Save.

Definition CURRY : forall (D1 D2 D3 : cpo), (Dprod D1 D2 -C-> D3) -c> D1 -C-> (D2-C->D3).
       intros; exists (Curry D1 D2 D3); red; intros; auto.
Defined.

Lemma CURRY_simpl : forall (D1 D2 D3 : cpo) (f:Dprod D1 D2 -C-> D3), 
       CURRY D1 D2 D3 f = Curry D1 D2 D3 f.
trivial.
Save.

Definition EV D E := ((AP D E) @2_ (FST _ _)) (SND _ _).

Add Parametric Morphism D E : (@EV D E)
with signature (@Oeq (Dprod (D -C-> E) D)) ==> (@Oeq E)
as EV_eq_compat.
intros p0. case p0. clear p0. intros a0 a1 y. case y. clear y. intros b0 b1 peq.
assert (p0:=Dprod_eq_elim_fst peq). simpl in p0.
assert (p1 := Dprod_eq_elim_snd peq). simpl in p1.
clear peq. assert (a0 a1 == b0 b1).
assert (stable a0) as sa by auto. unfold stable in sa.
specialize (sa _ _ p1). rewrite sa.
apply (fcont_eq_elim p0 b1).
clear p0 p1. auto.
Qed.

Lemma FST_PROD_fun: forall D E F f (g:D -C-> F), FST E F @_ PROD_fun f g == f.
intros. apply fcont_eq_intro. auto.
Qed.

Lemma SND_PROD_fun: forall D E F (f:D -C-> E) g, SND E F @_ PROD_fun f g == g.
intros. apply fcont_eq_intro. auto.
Qed.

Lemma curry_com: forall D E F (f:Dprod D E -C-> F),
        f == EV _ _ @_ (PROD_map (curry f) (ID E)).
intros. simpl. apply fcont_eq_intro. intros d. case d. clear d.
intros d e. auto.
Qed.

Lemma curry_unique: forall D E F (f:D -c> E -C-> F) g,
         g == EV _ _ @_ (PROD_map f (ID E)) -> f == curry g.
intros. apply fcont_eq_intro. intros a. simpl. apply fcont_eq_intro. intros b.
assert (xx:=fcont_eq_elim H (a,b)). auto.
Qed.

Add Parametric Morphism (D E F : cpo) : (@curry D E F)
with signature (@Oeq (Dprod D E -C-> F)) ==> (@Oeq D) ==> (@Oeq (E -C-> F))
as curry_eq_compat.
intros f1 f2 Eqf d1 d2 Eqd.
assert (monotonic (curry (D1:=D) (D2:=E) (D3:=F))) as M. auto.
unfold monotonic in M.
assert (M1 := M f1 f2). assert (M2 := M f2 f1).
destruct Eqf as [eq1 eq2]. specialize (M1 eq1). specialize (M2 eq2). clear eq1 eq2.
apply Ole_antisym.
simpl.
intros x. simpl in M1. specialize (M1 d1 x).
eapply (Ole_trans M1 (z:= fcontit f2 (d2,x))).
assert (monotonic f2) as Mf by auto.
unfold monotonic in Mf. apply Mf.
auto.

simpl. intros x. simpl in M2. specialize (M2 d2 x).
eapply (Ole_trans M2 (z:= fcontit f1 (d1,x))).
assert (monotonic f1). auto.
unfold monotonic in H. apply H. auto.
Qed.

Hint Resolve curry_eq_compat_Morphism.

Lemma curry_simpl: forall D E F (f:(Dprod E F) -C-> D) x y, curry f x y = f (x,y).
intros D E F f x y.
unfold curry.
repeat (try (rewrite fcont_COMP_simpl) ; try (rewrite fcont_comp_simpl) ; try (rewrite PAIR_simpl)).
simpl. auto.
Qed.

(*
Add Parametric  Morphism (C D E:cpo) : (@curry C D E)
with signature (@Oeq (Dprod C D -C-> E)) ==> (@Oeq (C -C-> (D -C-> E)))
as curry_eq_compat2.
intros. simpl. apply fcont_eq_intro. simpl. intros c.  simpl. apply fcont_eq_intro. intros d.
rewrite curry_simpl. rewrite curry_simpl. auto.
Qed.
*)

Definition uncurry D E F :=
      curry (EV _ _ @_ (PROD_fun (EV _ _ @_ PROD_fun (FST (D -C-> E -C-> F) _)
                                                     (FST D E @_ SND _ _))
                                 (SND _ _ @_ SND _ _))).

Add Parametric Morphism (D E F:cpo) : (@uncurry D E F)
with signature (@Oeq (D -c> (E -C-> F))) ==> (@Oeq ((Dprod D E) -c> F))
as uncurry_eq_compat.
intros. apply fcont_eq_intro ; auto.
Qed.
Hint Resolve uncurry_eq_compat_Morphism.

Lemma Uncurry_simpl : forall (D1 D2 D3 : cpo) (f:D1 -c> (D2-C->D3)) (p:Dprod D1 D2), 
       uncurry D1 D2 D3 f p = f (fst p) (snd p).
trivial.
Save.

(** ** Indexed product of cpo's *)

Definition Oprodi (I:Type)(O:I->ord) : ord.
intros; exists (forall i:I, O i) (fun p1 p2 => forall i:I, p1 i <= p2 i); intros; auto.
apply Ole_trans with (y i); trivial.
Defined.

Lemma Oprodi_eq_intro : forall (I:Type)(O:I->ord) (p q : Oprodi O), (forall i, p i == q i) -> p==q.
intros; apply Ole_antisym; intro i; auto.
Save.

Lemma Oprodi_eq_elim : forall (I:Type)(O:I->ord) (p q : Oprodi O), p==q -> forall i, p i == q i.
intros; apply Ole_antisym; case H; auto.
Save.

Definition Proj (I:Type)(O:I->ord) (i:I) : Oprodi O -m> O i.
intros.
exists (fun (x:Oprodi O) => x i); red; intuition.
Defined.

Lemma Proj_simpl : forall  (I:Type)(O:I->ord) (i:I) (x:Oprodi O),
            Proj O i x = x i.
trivial.
Save.

Definition Dprodi (I:Type)(D:I->cpo) : cpo.
intros; exists (Oprodi D) (fun (f : natO -m> Oprodi D) (i:I) => lub (Proj D i @ f));
intros; simpl; intros; auto.
apply le_lub with (f:= Proj (fun x : I => D x) i @ f) (n:=n).
apply lub_le; simpl; intros.
apply (H n i).
Defined.

Lemma Dprodi_lub_simpl : forall (I:Type)(Di:I->cpo)(h:natO-m>Dprodi Di)(i:I),
            lub h i = lub (c:=Di i) (Proj Di i @ h).
trivial.
Save.

Lemma Dprodi_eq_intro : forall I O (p1 p2:@Dprodi I O),
     (forall i, Proj O i p1 == Proj O i p2) -> p1 == p2.
intros I O p1 p2 C.
split ; intro i ; specialize (C i) ; simpl in C; auto.
Qed.

Lemma Dprodi_eq_elim : forall I O i (p1 p2: @Dprodi I O),
     p1 == p2 -> Proj O i p1 == Proj O i p2.
intros. simpl. simpl in H. destruct H as [X Y].
split ; auto.
Qed.

Hint Resolve Dprodi_eq_intro.
Hint Immediate Dprodi_eq_elim.

Lemma Dprodi_le_intro : forall I O (p1 p2:@Dprodi I O),
     (forall i, Proj O i p1 <= Proj O i p2) -> p1 <= p2.
intros I O p1 p2 C.
intro i ; specialize (C i) ; simpl in C; auto.
Qed.

Lemma Dprodi_le_elim : forall I O i (p1 p2: @Dprodi I O),
     p1 <= p2 -> Proj O i p1 <= Proj O i p2.
auto.
Qed.

Hint Resolve Dprodi_le_intro.
Hint Immediate Dprodi_le_elim.

Lemma Proj_cont : forall (I:Type)(Di:I->cpo) (i:I), 
                    continuous (D1:=Dprodi Di) (D2:=Di i) (Proj Di i).
red; intros; simpl; trivial.
Save.

Definition PROJ (I:Type)(Di:I->cpo) (i:I) : Dprodi Di -c> Di i := 
      mk_fconti (Proj_cont (Di:=Di) i).

Lemma PROJ_simpl : forall (I:Type)(Di:I->cpo) (i:I)(d:Dprodi Di),
               PROJ Di i d = d i.
trivial.
Save.

Lemma Dprodi_unique: forall I O D (f g : D -C-> @Dprodi I O),
    (forall i, PROJ O i @_ f == PROJ O i @_ g) -> f == g.
intros. simpl.
apply fcont_eq_intro. intros x. simpl.
refine (Dprodi_eq_intro _).
intros i. specialize (H i). assert (h:=fcont_eq_elim H x).
auto.
Qed.

Definition Prodi_fun I O D (f:forall i:I, D -m> O i) : D -m> Oprodi O.
intros. exists (fun d => fun i => f i d).
unfold monotonic. intros d0 d1 deq. simpl. intros i.
assert (monotonic (f i)) as M by auto.
apply M. auto.
Defined.

Definition PRODI_fun I O D (f:forall i:I, D -c> O i) : D -c> Dprodi O.
intros. exists (@Prodi_fun I _ D (fun i => fcontit (f i))).
unfold continuous. intros c. simpl.
intros i. assert (continuous (fcontit (f i))) as C by auto.
unfold continuous in C. rewrite C.
refine (lub_le_compat _). auto.
Defined.

Lemma PROJ_com_point: forall D I O (f:forall i, D -C-> O i) (i:I) d,
                 PROJ O i (PRODI_fun f d) = f i d.
auto.
Qed.

Lemma PROJ_com: forall D I O (f:forall i, D -C-> O i) (i:I),
                 PROJ O i @_ PRODI_fun f == f i.
auto using fcont_eq_intro.
Qed.

Definition PRODI_map I O1 O2 (f:forall i:I, O1 i -c> O2 i) :=
 PRODI_fun (fun i => f i @_ PROJ O1 i).

Lemma Dprodi_continuous : forall (D:cpo)(I:Type)(Di:I->cpo)
    (f:D -m> Dprodi Di), (forall i, continuous (Proj Di i @ f)) -> 
    continuous f.
red; intros; intro i.
apply Ole_trans with (lub (c:=Di i) ((Proj Di i @ f) @ h)); auto.
exact (H i h).
Save.

Definition Dprodi_lift : forall (I J:Type)(Di:I->cpo)(f:J->I),
             Dprodi Di -m> Dprodi (fun j => Di (f j)).
intros.
exists (fun (p:Dprodi Di) j => p (f j)); red; auto.
Defined.

Lemma Dprodi_lift_simpl : forall (I J:Type)(Di:I->cpo)(f:J->I)(p:Dprodi Di),
             Dprodi_lift Di f p = fun j => p (f j).
trivial.
Save.

Lemma Dprodi_lift_cont : forall (I J:Type)(Di:I->cpo)(f:J->I),
             continuous (Dprodi_lift Di f).
intros; apply Dprodi_continuous; red; simpl; intros; auto.
Save.

Definition DLIFTi (I J:Type)(Di:I->cpo)(f:J->I) : Dprodi Di -c> Dprodi (fun j => Di (f j)) 
             := mk_fconti (Dprodi_lift_cont (Di:=Di) f).

Definition Dmapi : forall (I:Type)(Di Dj:I->cpo)(f:forall i, Di i -m> Dj i), 
            Dprodi Di -m> Dprodi Dj.
intros; exists (fun p i => f i (p i)); red; auto.
Defined.

Lemma Dmapi_simpl : forall (I:Type)(Di Dj:I->cpo)(f:forall i, Di i -m> Dj i) (p:Dprodi Di) (i:I),
Dmapi f p i = f i (p i).
trivial.
Save.


(* Stuff
(** *** Particular cases with one or two elements *)

Section Product2.

Definition I2 := bool.
Variable DI2 : bool -> cpo.

Definition DP1 := DI2 true.
Definition DP2 := DI2 false.

Definition PI1 : Dprodi DI2 -c> DP1 := PROJ DI2 true.
Definition pi1 (d:Dprodi DI2) := PI1 d.

Definition PI2 : Dprodi DI2 -c> DP2 := PROJ DI2 false.
Definition pi2 (d:Dprodi DI2) := PI2 d.

Definition pair2 (d1:DP1) (d2:DP2) : Dprodi DI2 := bool_rect DI2 d1 d2.

Lemma pair2_le_compat : forall (d1 d'1:DP1) (d2 d'2:DP2), d1 <= d'1 -> d2 <= d'2 
            -> pair2 d1 d2 <= pair2 d'1 d'2.
intros; intro b; case b; simpl; auto.
Save.

Definition Pair2 : DP1 -m> DP2 -m> Dprodi DI2 := le_compat2_mon pair2_le_compat.

Definition PAIR2 : DP1 -c> DP2 -C-> Dprodi DI2.
apply continuous2_cont with (f:=Pair2).
red; intros; intro b.
case b; simpl; apply lub_le_compat; auto.
Defined.

Lemma PAIR2_simpl : forall (d1:DP1) (d2:DP2), PAIR2 d1 d2 = Pair2 d1 d2.
trivial.
Save.

Lemma Pair2_simpl : forall (d1:DP1) (d2:DP2), Pair2 d1 d2 = pair2 d1 d2.
trivial.
Save.

Lemma pi1_simpl : forall  (d1: DP1) (d2:DP2), pi1 (pair2 d1 d2) = d1.
trivial.
Save.

Lemma pi2_simpl : forall  (d1: DP1) (d2:DP2), pi2 (pair2 d1 d2) = d2.
trivial.
Save.

Definition DI2_map (f1 : DP1 -c> DP1) (f2:DP2 -c> DP2) 
               : Dprodi DI2 -c> Dprodi DI2 :=
                 DMAPi (bool_rect (fun b:bool => DI2 b -c>DI2 b) f1 f2).

Lemma Dl2_map_eq : forall (f1 : DP1 -c> DP1) (f2:DP2 -c> DP2) (d:Dprodi DI2),
               DI2_map f1 f2 d == pair2 (f1 (pi1 d)) (f2 (pi2 d)).
intros; simpl; apply Oprodi_eq_intro; intro b; case b; trivial.
Save.
End Product2.
Hint Resolve Dl2_map_eq.

Section Product1.
Definition I1 := unit.
Variable D : cpo.

Definition DI1 (_:unit) := D.
Definition PI : Dprodi DI1 -c> D := PROJ DI1 tt.
Definition pi (d:Dprodi DI1) := PI d.

Definition pair1 (d:D) : Dprodi DI1 := unit_rect DI1 d.

Definition pair1_simpl : forall (d:D) (x:unit), pair1 d x = d.
destruct x; trivial.
Defined.

Definition Pair1 : D -m> Dprodi DI1.
exists pair1; red; intros; intro d.
repeat (rewrite pair1_simpl);trivial.
Defined.


Lemma Pair1_simpl : forall (d:D), Pair1 d = pair1 d.
trivial.
Save.

Definition PAIR1 : D -c> Dprodi DI1.
exists Pair1; red; intros; repeat (rewrite Pair1_simpl).
intro d; rewrite pair1_simpl.
rewrite (Dprodi_lub_simpl (Di:=DI1)).
apply lub_le_compat; intros.
intro x; simpl; rewrite pair1_simpl; auto.
Defined.

Lemma pi_simpl : forall  (d:D), pi (pair1 d) = d.
trivial.
Save.

Definition DI1_map (f : D -c> D) 
               : Dprodi DI1 -c> Dprodi DI1 :=
                 DMAPi (fun t:unit => f).

Lemma DI1_map_eq : forall (f : D -c> D) (d:Dprodi DI1),
               DI1_map f d == pair1 (f (pi d)).
intros; simpl; apply Oprodi_eq_intro; intro b; case b; trivial.
Save.
End Product1.

Hint Resolve DI1_map_eq.
End of Stuff *)

(** ** Lifting *)

(* This is an attempt to tweak Christine's flat cpos *)

Section Lift_cpo.
Variable D : cpo.

(** ** Definition *)

CoInductive Dflat : Type := Eps : Dflat -> Dflat | Val : D -> Dflat.

Lemma DL_inv : forall d, d = match d with Eps x => Eps x | Val d => Val d end.
destruct d; auto.
Save.
Hint Resolve DL_inv.

(** ** Removing Eps steps *)

Definition pred d : Dflat := match d with Eps x => x | Val _ => d end.

Fixpoint pred_nth (d:Dflat) (n:nat) {struct n} : Dflat := 
    match n with 0 => d
              |S m => match d with Eps x => pred_nth x m
                                 | Val _ => d
                      end
    end.

(*
Fixpoint eps_nth (d:Dflat) (n:nat) {struct n} : Dflat :=
    match n with 0 => d
             | S m => Eps (eps_nth d m)
    end.
*)

Lemma pred_nth_val : forall x n, pred_nth (Val x) n = Val x.
destruct n; simpl; trivial.
Save.
Hint Resolve pred_nth_val.

Lemma pred_nth_Sn_acc : forall n d, pred_nth d (S n) = pred_nth (pred d) n.
destruct d; simpl; auto.
Save.

Lemma pred_nth_Sn : forall n d, pred_nth d (S n) = pred (pred_nth d n).
induction n; intros; auto.
destruct d.
exact (IHn d).
rewrite pred_nth_val; rewrite pred_nth_val; simpl; trivial.
Save.

(** ** Order *)
CoInductive DLle : Dflat -> Dflat -> Prop := 
                   DLleEps : forall x y,  DLle x y -> DLle (Eps x) (Eps y)
                |  DLleEpsVal : forall x d,  DLle x (Val d) -> DLle (Eps x) (Val d)
                |  DLleVal : forall d d' n y, pred_nth y n = Val d' -> d <= d' -> DLle (Val d) y.

Hint Constructors DLle.

Lemma DLle_rec : forall R : Dflat -> Dflat -> Prop,
   (forall x y, R (Eps x) (Eps y) -> R x y) ->
   (forall x d, R (Eps x) (Val d) -> R x (Val d)) ->
   (forall d y, R (Val d) y -> exists n,exists d', pred_nth y n = Val d' /\ d<=d')
   -> forall x y, R x y -> DLle x y.
intros R REps REpsVal RVal; cofix; destruct x; intros.
destruct y; intros.
apply DLleEps; apply DLle_rec; auto.
apply DLleEpsVal; apply DLle_rec; auto.
case (RVal t y H); intros n H1.
destruct H1.
apply DLleVal with x n; auto.
intuition.
intuition.
Save.

(** *** Properties of the order *)

Lemma DLle_refl : forall x, DLle x x.
intros x. apply DLle_rec with (R:= fun x y => forall n, pred_nth x n = pred_nth y n) ; try (clear x).
intros x y C n. specialize (C (S n)). simpl in C. auto.
intros x d C n. specialize (C (S n)). simpl in C. rewrite C. case n ; auto.
intros d y C. exists 0. exists d. simpl. specialize (C 0). simpl in C. auto.
intros n ; auto.
Save.
Hint Resolve DLle_refl.

Lemma DLvalval : forall d x, DLle (Val d) x -> 
 exists n, exists d', pred_nth x n = Val d' /\ d<=d'.
intros d x H.
inversion H.
exists n; exists d'.
intuition.
Qed.

Lemma pred_nth_epsS n x y: Eps y = pred_nth x n -> y = pred_nth x (S n).
intros n. induction n. intros x y E. simpl in E. rewrite <- E. simpl. auto.
intros x. case x ; clear x. intros x y. intros E. simpl in E. specialize (IHn _ _ E). auto.
intros x y. simpl. intros incon. inversion incon.
Qed.

Lemma pred_nth_valS n d y: Val d = pred_nth y n -> Val d = pred_nth y (S n).
intros n ; induction n. intros d y E. simpl in E. rewrite <- E. auto.
intros d y. case y ; clear y. intros y E. simpl in E. specialize (IHn _ _ E). rewrite IHn. auto.
intros dd E. simpl in E. inversion E. rewrite <- H0 in *. clear dd H0 E. simpl. auto.
Qed.

Lemma pred_nth_comp y : forall m0 n1 m1 : nat,
   m0 = n1 + m1 -> pred_nth y m0 = pred_nth (pred_nth y n1) m1.
intros y m0. induction m0. intros n1 m1. simpl. intros E. assert (n1 = 0) by omega. rewrite H in *.
simpl. assert (m1 = 0) as A by omega. rewrite A in *. auto.
intros n1 m1. case m1 ; clear m1. case n1 ; clear n1. simpl. intros incon. inversion incon.
intros m1 E. assert (m0 = m1) as EE by omega. rewrite EE in *. auto.
intros m1. intros E. assert (m0 = n1 + m1) as A by omega. specialize (IHm0 _ _ A).
clear E A. generalize IHm0. clear IHm0. case_eq (pred_nth y m0).
intros d E. rewrite <- (pred_nth_epsS (sym_eq E)). intros EE. rewrite (pred_nth_epsS EE). auto.
intros d E EE.
rewrite <- (pred_nth_valS (sym_eq E)). rewrite (pred_nth_valS EE). auto.
Qed.

Lemma DLle_pred_nth x y n : DLle x y -> DLle (pred_nth x n) (pred_nth y n).
intros x y n. induction n. simpl. auto. intros C.
specialize (IHn C). inversion IHn.
rewrite (pred_nth_epsS H) in *. clear H x0. rewrite (pred_nth_epsS H0) in *. clear y0 H0. auto.
rewrite (pred_nth_epsS H) in *. clear H x0. rewrite H0 in *.
rewrite <- (pred_nth_valS H0). rewrite <- H0 in *. auto.
rewrite <- H2 in *. rewrite <- H in *.
rewrite <- (pred_nth_valS H).
apply DLleVal with (n:=n0) (d':=d').
rewrite H2 in H0. rewrite <- (@pred_nth_comp y (n+n0)) in H0 ; auto.
rewrite <- (@pred_nth_comp y (S n+n0)) ; auto. rewrite (pred_nth_valS (sym_eq H0)). auto. auto.
Qed.

Lemma DLleEps_right : forall x y,  DLle x y -> DLle x (Eps y).
intros x y C. apply DLle_rec with (R:= fun x y => forall n, DLle (pred_nth x n) (pred_nth y (S n))).
clear x y C. intros x y C n. specialize (C (S n)). auto.
clear x y C. intros x y C n. specialize (C (S n)). auto.
clear x y C. intros d y C. specialize (C 0).
assert (DLle (Val d) (pred_nth y 1)) as CC by auto. clear C.
destruct (DLvalval CC) as [n [dd [X Y]]]. exists (S n). exists dd. split ; auto.
generalize X. clear X. destruct y. simpl. auto. simpl. case n ; auto.
intros n. simpl.
apply DLle_pred_nth ; auto.
Qed.
Hint Resolve DLleEps_right.

Lemma DLleEps_left : forall x y,  DLle x y -> DLle (Eps x) y.
intros x y C. apply DLle_rec with (R:= fun x y => forall n, DLle (pred_nth x (S n)) (pred_nth y n)).
clear x y C. intros x y C n. specialize (C (S n)). auto.
clear x y C. intros x y C n. specialize (C (S n)). destruct n ; auto.
clear x y C. intros d y C. specialize (C 0).
simpl in C.
destruct (DLvalval C) as [n [dd [X Y]]]. exists n. exists dd. split ; auto.
intros n. simpl. apply DLle_pred_nth ; auto.
Qed.
Hint Resolve DLleEps_left.

Lemma DLle_pred_left : forall x y, DLle x y -> DLle (pred x) y.
destruct x; destruct y; simpl; intros; trivial.
inversion H; auto.
inversion H; trivial.
Save.

Lemma DLle_pred_right : forall x y, DLle x y -> DLle x (pred y).
destruct x; destruct y; simpl; intros; trivial.
inversion H; auto.
inversion H; trivial.
destruct n; simpl in H1.
discriminate H1.
apply DLleVal with d' n; auto.
Save.

Hint Resolve DLle_pred_left DLle_pred_right.

Lemma DLle_pred : forall x y, DLle x y -> DLle (pred x) (pred y).
auto.
Save.

Hint Resolve DLle_pred.

Lemma DLle_pred_nth_left : forall n x y, DLle x y -> DLle (pred_nth x n) y.
induction n; intros.
simpl; auto.
rewrite pred_nth_Sn; auto.
Save.

Lemma DLle_pred_nth_right : forall n x y,
      DLle x y -> DLle x (pred_nth y n).
induction n; intros.
simpl; auto.
rewrite pred_nth_Sn; auto.
Save.

Hint Resolve DLle_pred_nth_left DLle_pred_nth_right.


(* should be called leq *)
Lemma DLleVal_eq : forall x y, DLle (Val x) (Val y) -> x<=y.
intros x y H; inversion H.
destruct n; simpl in H1; injection H1;intro; subst y; auto.
Save.
Hint Immediate DLleVal_eq.

(* New *)
Lemma DLle_leVal : forall x y, x<=y -> DLle (Val x) (Val y).
intros x y H.
apply (DLle_rec (R := fun x' y' => x'=Val x /\ y'=Val y)).
intros.
destruct H0; discriminate.
intros.
destruct H0; discriminate.
intros.
destruct H0.
subst y0.
exists 0.
exists y.
simpl.
split; trivial.
injection H0.
intro; subst d; auto.
auto.
Qed.


Lemma DLnvalval : forall n y d z, pred_nth y n = Val d -> DLle y z ->
 exists m, exists d', pred_nth z m = Val d' /\ d <= d'.
induction n.
intros.
dcase y.
intros.
subst y.
simpl in H.
discriminate.
intros.
subst y.
simpl in H.
rewrite H in H0. 
apply (DLvalval H0).

intros.
apply (IHn (pred y) d z).
rewrite <- (pred_nth_Sn_acc).
assumption.
apply DLle_pred_left.
assumption.
Qed.


(* This one's no longer true, of course.
   Should probably replace with antisym?
Lemma DLleVal_sym : forall x y, DLle (Val x) y -> DLle y (Val x).
intros x y H; inversion H.
rewrite <- H1; apply DLle_pred_nth_right; auto.
Save.
*)

Lemma DLleEpsn n x z xx : pred_nth x n = Val xx -> DLle (Val xx) z -> DLle x z.
intros n. induction n.
intros x z xx. simpl. intros C. rewrite C. auto.
intros x. case x ; clear x. intros d z xx. simpl. intros C DD.
specialize (IHn _ z _ C DD).
generalize IHn. case z. intros zz CC. apply DLleEps. apply (DLle_pred_right CC).
intros zz CC. apply DLleEpsVal. apply CC.
intros x z xx C. simpl in C. inversion C. auto.
Qed.

Lemma DLle_trans : forall x y z, DLle x y -> DLle y z -> DLle x z.
intros x y z D0 D1.
apply DLle_rec with (R:=fun x z => (forall xx n, pred_nth x n = Val xx -> DLle x z)).
clear x y z D0 D1. intros x y C xx n X.
specialize (C xx (S n) X).
apply (DLle_pred C).

clear x y z D0 D1. intros x y C xx n X.
specialize (C _ (S n) X).
apply (DLle_pred_left C).

intros d yy C.
specialize (C d 0).
destruct (fun Z => DLvalval (C Z)) as [n [dd [ZZ Z]]] ; auto.
exists n. exists dd. split ; auto.

intros xx n C.
destruct (DLnvalval C D0) as [m [yy [P Q]]].
destruct (DLnvalval P D1) as [k [zz [Pz Qz]]].
apply (DLleEpsn C). apply (DLleVal Pz). apply (Ole_trans Q Qz).
Qed.

(*
Lemma DLtranscase : forall x d d', DLle x (Val d) -> d<=d' -> DLle x (Val d').
intros x d d' C P.
apply (DLle_trans C).
assert (pred_nth (Val d') 0 = Val d') as X by auto.
apply (DLleVal X P).
Qed.
*)

(** *** Declaration of the ordered set *)
Definition DL_ord : ord.
exists Dflat DLle; intros; auto.
apply DLle_trans with y; trivial.
Defined.

Canonical Structure DL_ord.


(** ** Definition of the cpo structure *)

Lemma eq_Eps : forall x, x == Eps x.
intros; apply Ole_antisym; repeat red; auto.
Save.
Hint Resolve eq_Eps.

(** *** Bottom is given by an infinite chain of Eps *)
CoFixpoint DL_bot : DL_ord := Eps DL_bot.

Lemma DL_bot_eq : DL_bot = Eps DL_bot.
transitivity match DL_bot with Eps x => Eps x | Val d => Val d end; auto.
Save.

Lemma DLless_cond : forall (x y : DL_ord), (forall xx, x == Val xx -> x <= y) -> DLle x y.
intros x y P. apply DLle_rec with (R:=fun x y => forall xx, x == Val xx -> x <= y).
intros d0 d1 IH. intros d00 dv.
rewrite (eq_Eps d0) in dv. rewrite (eq_Eps d0).
specialize (IH _ dv).
rewrite (eq_Eps d1). auto.

intros d0 d1 IH dd dv. rewrite (eq_Eps d0) in dv.
specialize (IH _ dv). rewrite (eq_Eps d0). auto.

intros d0 d1 IH.
specialize (IH _ (Oeq_refl _)).
auto using (DLvalval IH).
apply P.
Qed.

Lemma DL_bot_least : forall x, DL_bot <= x.
intros x. apply DLless_cond.
intros xx [_ C].
destruct (DLvalval C) as [n [dd [P Q]]].
assert False as F. induction n. simpl in P. rewrite DL_bot_eq in P. inversion P.
rewrite DL_bot_eq in P. simpl in P. apply (IHn P). inversion F.
Qed.

(** *** More properties of elements in the flat domain *)

(* This is essentially the same, when fixed up, as DLvalval that I proved above
   I've repeated it here, though
*)
Lemma DLle_Val_exists_pred : 
      forall x d, Val d <= x -> exists k, exists d', pred_nth x k = Val d'
            /\ d <= d'.
intros x d H; inversion H; eauto.
Save.

Lemma Val_exists_pred_le : 
      forall x d, (exists k, pred_nth x k = Val d) -> Val d <= x.
destruct 1; intros.
apply DLleVal with d x0; trivial.
Save.
Hint Immediate DLle_Val_exists_pred Val_exists_pred_le.

Lemma Val_exists_pred_eq : 
      forall x d, (exists k, pred_nth x k = Val d) -> (Val d:DL_ord) == x.
intros.
split.
auto.
destruct H.
rewrite <- H.
apply DLle_pred_nth_right.
auto.
Save.

(** *** Construction of least upper bounds *)

Definition isEps x := match x with Eps _ => True | _ => False end.

Lemma isEps_Eps : forall x, isEps (Eps x).
repeat red; auto.
Save.

Lemma not_isEpsVal : forall d, ~ (isEps (Val d)).
repeat red; auto.
Save.
Hint Resolve isEps_Eps not_isEpsVal.

Lemma isEps_dec : forall x, {d:D|x=Val d}+{isEps x}.
destruct x; auto.
left.
exists t; auto.
Defined.

(* Slightly more radical rewrite, trying to use equality *)

Lemma allvalsfromhere : forall (c:natO -m> DL_ord) n d i,
  c n == Val d -> exists d', c (n+i) == Val d' /\ d <= d'.
intros c n d i (Hnv1,Hnv2).
assert (((Val d):DL_ord) <= c (n+i)).
apply Ole_trans with (c n); auto.
apply fmonotonic.
intuition.
assert (HH:= @DLle_Val_exists_pred (c (n+i)) d H).
destruct HH as (k,(d',Hkd')).
exists d'.
split.
apply Oeq_sym; apply Val_exists_pred_eq; firstorder.
tauto.
Qed.


Definition hasVal x := exists j, exists d, pred_nth x j = Val d.
Definition valuable := {x | hasVal x}.

Definition projj : valuable -> DL_ord.
intro.
destruct X.
exact x.
Defined.


(* Redoing extract using constructive epsilon *)


Definition findindexandval (x : valuable) : {kd | match kd with (k,d) => (pred_nth (projj x) k) = Val d end}.
intro x; destruct x.
unfold hasVal in h; simpl.
cut {n | exists d, pred_nth x n = Val d}.
intro; destruct H.
dcase (pred_nth x x0).
intros; rewrite H in e; absurd (exists d0, Eps d = Val d0).
intro.
destruct H0.
discriminate.
assumption.
intros.
exists (x0,t).
assumption.

apply constructive_indefinite_description_nat.
intro.
case (isEps_dec (pred_nth x x0)).
intro; left.
destruct s.
exists x1.
assumption.
intro H; right.
intro; destruct H0.
rewrite H0 in H; simpl in H.
contradiction.
exact h.
Defined.


Definition extract (x : valuable) : D.
intro x; generalize (findindexandval x); intro.
destruct X.
destruct x0.
exact t.
Defined.


Lemma extractworks : forall x, projj x == Val (extract x).
intro x; unfold extract.
dcase (findindexandval x).
intros.
destruct x0.
apply Oeq_sym.
apply Val_exists_pred_eq.
exists n.
assumption.
Qed.


Lemma vinj : forall d d',Val d == Val d' -> d == d'.
intros d d' (H1,H2).
split; auto.
Qed.

Lemma vleinj : forall d d', Val d <= Val d' -> d<=d'.
auto.
Qed.

Lemma extractmono : forall (x y : valuable), (projj x) <= (projj y) -> extract x <= extract y.
intros x y H.
apply vleinj.
rewrite <- (extractworks x).
rewrite <- (extractworks y).
assumption.
Qed.

(* This is the simpler one that just takes an equality *)

Lemma eqValpred : forall x d, x == Val d -> exists k, exists d', pred_nth x k = Val d' /\ d==d'.
intros.
destruct H as (H1,H2).
assert (HH := DLle_Val_exists_pred H2).
destruct HH as (k,(d',HH)).
exists k; exists d';split.
tauto.
split.
tauto.
apply vleinj.
destruct HH.
rewrite <- H.
auto.
Qed.

Definition makechain (c:natO -m> DL_ord) k d (Hck:c k == Val d) : nat -> D.
intros c k d Hck i.
assert (HH:= @allvalsfromhere c k d i Hck).
eapply extract.
exists (c (k+i)).
destruct HH as (d',(Hd',HHH)).
unfold hasVal.
generalize (eqValpred Hd').
firstorder.
Defined.

Definition makechainm (c:natO -m> DL_ord) k d (Hck:c k == Val d) : natO -m> D.
intros.
exists (makechain Hck).
unfold monotonic.
intros.
unfold makechain.
apply extractmono.
simpl.
apply (fmonotonic c).
intuition.
Defined.


Definition cpred (c:natO -m> DL_ord) : natO-m>DL_ord.
intro c; exists (fun n => pred (c n)); red.
intros x y H; apply DLle_pred.
apply (fmonotonic c); auto.
Defined.

Lemma fValIdx : forall (c:natO -m> DL_ord) (n:nat), 
        {dk: (D*nat)%type | match dk with (d,k) => k<n /\ c k == Val d end} + {forall k, k<n -> isEps (c k)}.
induction n.
right; intros. destruct (lt_n_O k H).
case IHn; intros.
destruct s as [(d,k) [H ck]].
left; exists (d,k).
auto with arith.

case (isEps_dec (c n)); intros.
destruct s as [d H]. left.
exists (d,n); intuition.

right; intros.
assert (L:=lt_n_Sm_le _ _ H). destruct L ; auto with arith.
Defined.

Definition createchain (c:natO -m> DL_ord) (n:nat) 
  (X:{dk: (D*nat)%type | match dk with (d,k) => k<n /\ c k == Val d end}) :
   (natO -m> D) :=
match X with | exist (d,k) (conj Hk Hck) => @makechainm c k d Hck
end.

CoFixpoint DL_lubn (c:natO-m> DL_ord) (n:nat) : DL_ord := 
    match fValIdx c n with inleft X => Val (lub (createchain X))
                                |  inright _  => Eps (DL_lubn (cpred c) (S n))
    end.

Lemma DL_lubn_inv : forall (c:natO-m> DL_ord) (n:nat), DL_lubn c n = 
     match fValIdx c n with inleft X => Val (lub (createchain X))
                                |  inright _  => Eps (DL_lubn (cpred c) (S n))
    end.
intros; rewrite (DL_inv (DL_lubn c n)).
simpl; case (fValIdx c n); trivial. 
Qed.

Lemma makechainworks : forall (c:natO -m> DL_ord) k dk (H2:c k == Val dk) i (d:D), c (k+i) == (Val d) -> makechain H2 i == d.
intros.
apply vinj.
rewrite <- H.
apply Oeq_sym.
unfold makechain.
rewrite <- extractworks.
simpl.
auto.
Qed.

Lemma chainluble : forall (c:natO -m> DL_ord) k dk (Hkdk : c k == Val dk) k' dk' (Hkdk': c k' == Val dk'), 
   lub (makechainm Hkdk) <= lub (makechainm Hkdk').
intros;apply lub_le.
intro n.
assert (Hkn := allvalsfromhere n Hkdk).
destruct Hkn as (dkn,(Hckn,Hdkdkn)).
setoid_replace (makechainm Hkdk n) with (dkn).
assert (Hkk'n := allvalsfromhere (n+k) Hkdk').
destruct Hkk'n as (dkk'n,(Hckk'n,Hddd)).
apply Ole_trans with dkk'n.
apply vleinj; rewrite <- Hckn; rewrite <- Hckk'n.
apply fmonotonic.
intuition.
setoid_replace dkk'n with (makechainm Hkdk' (n+k)).
apply le_lub.
unfold makechainm.
apply Oeq_sym; apply makechainworks.
assumption.
unfold makechainm; apply makechainworks.
assumption.
Qed.

Lemma chainlubinv : forall (c:natO -m> DL_ord) k dk (Hkdk : c k == Val dk) k' dk' (Hkdk': c k' == Val dk'), 
   lub (makechainm Hkdk) == lub (makechainm Hkdk').
intros;split;apply chainluble.
Qed.

Lemma pred_lubn_Val : forall (d:D)(n k p:nat) (c:natO-m> DL_ord),
             (n <k+p)%nat -> pred_nth (c n) k = Val d
                                   -> exists d', DL_lubn c p == Val d'.

induction k; intros.
(* Base case *)
rewrite (DL_lubn_inv c p); simpl.
simpl in H0.
case (fValIdx c p); intros.
case s; intros (d',k) (H1,H2); auto.
simpl.
exists (lub (makechainm H2)).
auto.

absurd (c n = Val d).
intro Hsilly.
assert (isEps (c n)).
apply i.
omega.
rewrite Hsilly in H1.
simpl in H1.
tauto.
assumption.

(* Induction *)
rewrite (DL_lubn_inv c p).
case (fValIdx c p); intros.
unfold createchain.
dependent inversion s.
simpl.
destruct x.
dependent inversion y.
exists (lub (makechainm o)).
auto.

assert (n<k + S p).
omega.
assert (pred_nth (cpred c n) k = Val d).
unfold cpred.
simpl.
rewrite <- pred_nth_Sn_acc.
assumption.
specialize (IHk (S p) (cpred c) H1 H2 ).
destruct IHk as (d',IH); exists d'.
rewrite <-IH; auto.
Qed.

Lemma eqDLeq : forall d d', d == d' -> Val d == (Val d').
  intros.
  destruct H; split; apply DLle_leVal.
  assumption.
  assumption.
Qed.

Lemma predeq : forall x, x == pred x.
  intro; split; auto.
Qed.

Lemma cpredsamelub : forall (c:natO -m> DL_ord) k dk (H1:c k == Val dk) (H2: (cpred c) k == Val dk),
                      lub (makechainm H1) == lub (makechainm H2).
intros; split.
apply lub_le; intro.
setoid_replace (makechainm H1 n) with (makechainm H2 n).
apply le_lub.
assert (Hkn := allvalsfromhere n H1).
destruct Hkn as (dkn,(Hckn,Hdkdkn)).
unfold makechainm; simpl.
setoid_rewrite (makechainworks H1 Hckn).
assert (HH : ((cpred c) (k+n) == Val dkn)).
unfold cpred; simpl.
rewrite <- Hckn.
apply Oeq_sym; apply predeq.
setoid_rewrite (makechainworks H2 HH).
auto.

apply lub_le; intro.
setoid_replace (makechainm H2 n) with (makechainm H1 n).
apply le_lub.
assert (Hkn := allvalsfromhere n H2).
destruct Hkn as (dkn,(Hckn,Hdkdkn)).
unfold makechainm; simpl.
setoid_rewrite (makechainworks H2 Hckn).
assert (HH : ((c) (k+n) == Val dkn)).
unfold cpred in Hckn; simpl in Hckn.
rewrite <- Hckn; apply predeq.
setoid_rewrite (makechainworks H1 HH); auto.
Qed.


Lemma attempt2 : forall kl (c:natO -m> DL_ord) p d, pred_nth (DL_lubn c p) kl = Val d 
                    -> exists k, exists dk, exists Hkdk:c k == Val dk, d == lub (makechainm Hkdk).
induction kl; intros.
rewrite DL_lubn_inv in H; simpl in H.
dcase (fValIdx c p).
intros.
rewrite H0 in H.
destruct s; destruct x; inversion y.

unfold createchain in H.
exists n; exists t.
exists H2.
clear H0.
generalize H.
clear H.
dependent inversion y.
intro.
apply vinj.
rewrite <- H.

apply eqDLeq.
apply chainlubinv.

(* then we get an impossible case *)

Focus 2.
(* this is the tricky one *)
assert (HH := IHkl (cpred c) (S p) d).
dcase (fValIdx c p).
intros.

(* making it up as I go along *)
rewrite DL_lubn_inv in H.
rewrite H0 in H; simpl in H.
clear H0.
destruct s.
destruct x.
generalize H; dependent inversion y; intros.
unfold createchain in H0.
exists n; exists t; exists o.
apply vinj.
rewrite <- H0.
auto.

clear IHkl.
intros.
rewrite DL_lubn_inv in H.
rewrite H0 in H; simpl in H.
specialize (HH H).
(* AHA now we're getting somewhere at last, albeit rather clunkily! *)
destruct HH as (k,(dk,(Hkdk,HH))).
exists k; exists dk.

unfold cpred in Hkdk; simpl in Hkdk.
assert (HHH : c k == Val dk).
rewrite <- Hkdk.
apply predeq.
exists HHH.
rewrite HH.

apply Oeq_sym.
apply cpredsamelub.

(* now back to trivial one *)
intros.
rewrite H0 in H.
discriminate.
Qed.

Lemma pred_nth_sum : forall x m n, pred_nth x (m+n) = pred_nth (pred_nth x m) n.
induction m.
simpl.
intuition.
intro n; replace (S m + n) with (m + S n); intuition.
rewrite IHm.
rewrite (@pred_nth_Sn_acc n (pred_nth x m)).
rewrite pred_nth_Sn.
reflexivity.
Qed.

(* first just wrap pred_lubn_Val to use equality *)
Lemma chainVallubnVal : forall (c:natO -m> DL_ord) n d p, c n == Val d -> exists d', DL_lubn c p == Val d'.
intros.
assert (HH := eqValpred H).
destruct HH as (k,(d'',(Hcnk,Hdd''))).
apply (pred_lubn_Val (n:=n) (k:=k+n+1) (p:=p) (c:=c) (d:=d'')).
intuition.
replace (k+n+1) with (k+(n+1)).
rewrite pred_nth_sum.
rewrite Hcnk.
auto.
omega.
Qed.

Lemma chainVallubnlub : forall (c:natO -m> DL_ord) n d (H : c n == Val d) p, exists d',  DL_lubn c p == Val d' /\ d' == lub (makechainm H).
intros.
assert (HH := chainVallubnVal p H ).
destruct HH as (d',HH).
exists d'.
split. assumption.
assert (HHH := eqValpred HH).
destruct HHH as (k,(d'',(Hk,Hd'd''))).
assert (HHH := attempt2 Hk).
destruct HHH as (p',(dp',(Hp',Heq))).
rewrite Hd'd''.
rewrite Heq.
apply chainlubinv.
Qed.


Definition DL_lub (c:natO-m> DL_ord) := DL_lubn c 1.


Lemma DL_lub_upper : forall c:natO-m> DL_ord, forall n, c n <= DL_lub c.
intros c n. apply DLless_cond.
intros d C.
destruct (chainVallubnlub C 1) as [d' [Ln L]].
unfold DL_lub. rewrite Ln.
assert (X:=makechainworks C). specialize (X 0 d). assert (n+0 = n) as A by omega. rewrite A in X. clear A.
specialize (X C). apply Ole_trans with (y:=Val d). auto.
assert (pred_nth (Val d') 0 = Val d') as Y by auto.
apply (DLleVal Y). rewrite L.
apply Ole_trans with (y:=(makechainm (c:=c) (k:=n) (d:=d) C) 0). simpl. auto.
apply le_lub.
Qed.

(* Just realized I should have had Val as a morphism ages ago *)
Add Parametric Morphism : Val
with signature (@Oeq D) ==> (@Oeq DL_ord)
as Val_eq_compat.
intros.
apply eqDLeq.
assumption.
Qed.


Lemma DLle_Val_exists_eq : forall y d, Val d <= y -> exists d', y == Val d' /\ d <= d'.
intros y d H; inversion H.
exists d'.
split.
symmetry.
apply Val_exists_pred_eq.
exists n; trivial.
assumption.
Qed.


Lemma DL_lub_least : forall (c:natO-m> DL_ord) a, 
                      (forall n, c n <= a) -> DL_lub c <= a.
intros c a C.
apply DLless_cond. intros d X.
destruct (eqValpred X) as [n [dd [P Q]]].
destruct (attempt2 P) as [m [d' [cm Y]]].
apply Ole_trans with (y:= (Val dd)).
refine (proj2 (Val_exists_pred_eq _)). exists n. auto.
assert (Z:= C m). rewrite cm in Z.
destruct (DLvalval Z) as [mm [e [R S]]].
apply (DLleVal R). rewrite Y.
apply lub_le.
intros nn.
assert (A:=makechainworks cm).
specialize (A nn).
clear S Y Z.
destruct (allvalsfromhere nn cm) as [nnv [S T]].
specialize (A _ S). rewrite A. clear A.
assert (a == Val e) as E.
apply Oeq_sym.
apply (Val_exists_pred_eq). exists mm. apply R.
assert (Z := C (m+nn)).
rewrite E in Z. rewrite S in Z. apply (vleinj Z).
Qed.


(** *** Declaration of the lifted cpo *)

Definition DL : cpo.
exists DL_ord  DL_lub; intros.
(* apply DL_bot_least. *)
apply DL_lub_upper.
apply DL_lub_least; trivial.
Defined.

Lemma lubval: forall (c:natO -m> DL) d, lub c == Val d ->
                exists k, exists d', c k == Val d' /\ d' <= d.
intros c d l.
destruct l as [lubval1 lubval2].
destruct (DLle_Val_exists_pred lubval2) as [k xx].
destruct xx as [f' [lubval lf]].
assert (lub c == Val f') as lubval'.
apply Ole_antisym.
eapply Ole_trans. apply lubval1. auto using DLle_leVal.
eapply (DLleVal). apply lubval. apply Ole_refl.
assert (f' <= d).
destruct lubval' as [v1 v2].
assert (Val f' <= Val d).
apply (Ole_trans v2 lubval1).
apply (DLleVal_eq H).
assert (d == f') as feq. auto.
clear H lubval1 lubval2 lf.

assert (xx:=attempt2 lubval).
destruct xx as [n [f0 [chnval lf']]].
exists n. exists f0. split. apply chnval.
assert (w:=makechainworks chnval).
specialize (w 0 f0). assert (n+0 = n) by omega. rewrite H in w.
specialize (w chnval).
destruct w as [_ L2].
refine (Ole_trans L2 (Ole_trans (le_lub (makechainm chnval) 0) _)).
rewrite <- lf'. auto.
Qed.

Lemma DLvalgetchain: forall (c:natO -m> DL) k d, c k == Val d -> exists c':natO -m> D, forall n, c(k+n) == Val (c' n).
intros c k d chk.
exists (makechainm chk).
intros n.
destruct (allvalsfromhere n chk) as [d' [chd ld]].
refine (Oeq_trans chd _).
apply eqDLeq.
apply (Oeq_sym (makechainworks chk chd)).
Qed.

Hint Immediate DLle_leVal.

Definition eta_m : D -m> DL.
exists (fun (x:D) => Val x).
unfold monotonic.
intros;auto.
Defined.

Definition eta : D -c> DL.
exists (eta_m).
red.
intro c.
set (LHS := eta_m (lub c)).
unfold DL, DL_lub.
unfold lub.
rewrite DL_lubn_inv.
simpl.
subst LHS.
simpl.
apply DLle_leVal.
apply lub_le; intro n.
setoid_replace (c n) with (makechainm  (c:= eta_m @ c) (k:=0) (d := c 0) (Oeq_refl_eq (refl_equal (Val (c 0)))) n).
apply le_lub.
unfold makechainm.
simpl.
symmetry. apply makechainworks.
simpl.
trivial.
Defined.

Add Parametric Morphism : (@eta)
with signature (@Oeq D) ==> (@Oeq (DL))
as eta_eq_compat.
intros d0 d1 Eq.
assert (monotonic (eta)). auto.
unfold monotonic in H.
specialize (H d0 d1). auto using Ole_antisym.
Qed.

Hint Resolve eta_eq_compat.

End Lift_cpo.

CoFixpoint kleisli (D E : cpo) (M : DL_ord D) (N : D -c> DL E) : DL E :=
 match M with Eps M' => Eps (kleisli  M' N)
            | Val d => N d
 end.

Lemma kleisli_inv : forall D E (M:DL D) (N: D -c> DL E), kleisli  M N =
 match M with Eps M' => Eps (kleisli  M' N)
            | Val d => N d
 end.
intros.
transitivity  (match kleisli M N with Eps x => Eps x | Val d => Val d end).
apply (DL_inv (D:=E)).
case M.
auto.
intro d.
simpl.
symmetry.
apply DL_inv.
Qed.

(* Monad stuff *)
Lemma left_unit : forall (D E : cpo) (d:D) (N : D -c> DL E), kleisli (Val d) N == N d.
intros; rewrite kleisli_inv.
trivial.
Qed.

Lemma kleisli_Eps: forall D E (x : DL D) (N:D -c> DL E), kleisli (Eps x) N = Eps (kleisli x N).
intros.
rewrite kleisli_inv. auto.
Qed.

Lemma kleisli_Val: forall (D:cpo) E (d:D) (N:D -c> DL E), kleisli (Val d) N = N d.
Proof.
  intros.
  rewrite kleisli_inv.
  auto.
Qed.

(* Revising kleislival to work with nats first cos that'll make the induction work better... *)
Lemma kleislipredValVal : forall (D E : cpo) k (e: tcpo E) (M: DL_ord D) (N : D -c> DL E), pred_nth (kleisli M N) k = Val e -> exists d, M == Val d /\ N d == Val e.
induction k.

intros.
 simpl in H.
 rewrite kleisli_inv in H.
 dcase M; intros.
 rewrite H0 in H.
  discriminate.
 rewrite H0 in H.
  exists t.
  split.
  trivial.
  rewrite H; auto.

intros.
 rewrite kleisli_inv in H.
 dcase M; intros.
 rewrite H0 in H.
 simpl pred_nth in H.
 assert (Hfoo := IHk _ _ _ H).
 destruct Hfoo as (d0,(H1,H2)).
 exists d0.
 split.
 rewrite <- H1.
 apply Oeq_sym.
 apply eq_Eps.
 assumption.

 rewrite H0 in H.
 exists t.
 split.
 auto.
 rewrite <- H.
 (* this is bloody obvious, why has auto foresaken me? *)
 Lemma predn_DLeq : forall D n (x:DL_ord D), x == pred_nth x n.
 induction n; intros.
  simpl; auto.
  rewrite pred_nth_Sn.
  set (GRR := pred_nth x n).
  rewrite (IHn x).
  subst GRR.
  apply predeq.
 Qed.

 apply (@predn_DLeq E (S k) (N t)).
Qed.

Lemma kleisli_Valeq: forall (D E:cpo) v (d:D) M, v == Val d ->
   kleisli v M == (M d:DL_ord (E)).
Proof.
intros D E v d M veq.
destruct veq as [v1 v2].
destruct (DLle_Val_exists_pred v2) as [n [dd [pd ld]]].
assert (dd == d) as DD. apply Ole_antisym.
assert (xx:=DLle_pred_nth_left). specialize (xx _ n _ _ v1). rewrite pd in xx.
apply vleinj. apply xx. apply ld. clear ld.

assert (M d == M dd). assert (stable M) by auto. unfold stable in H.
apply H. auto. rewrite H. clear d DD H v1 v2.

  generalize dependent v.
  induction n.
  intros v P. simpl in P. rewrite P in *.
  rewrite kleisli_Val. auto.

  intros v P. destruct v.
    rewrite kleisli_Eps.
    simpl in P.
    apply Oeq_trans with (kleisli v M). apply (Oeq_sym (@eq_Eps E (kleisli v M))).
    rewrite (IHn _ P) ; auto.

  inversion P as [Eq]. rewrite Eq in *. simpl in P.
  assert (pred_nth (Val dd) n = Val dd). destruct n ; auto.
  apply (IHn _ H).
Qed.

Hint Resolve DLle_leVal.

Lemma kleisliValVal : forall (D E : cpo) (M: DL D) (N : D -c> DL E) e,
                kleisli M N == Val e -> exists d, M == Val d /\ N d == Val e.
intros D E M N e [kv1 kv2].
destruct (DLvalval kv2) as [n [d [pd ld]]].
destruct (kleislipredValVal pd) as [md [vm ndm]].
exists md. split. apply vm. refine (Oeq_trans ndm _).
rewrite (kleisli_Valeq N vm) in kv1. rewrite ndm in kv1.
split. apply kv1.
auto.
Qed.

Definition kleisli_m (D E:cpo) : (DL_ord D) -> (D-c> DL E) -m> DL E.
intros.
exists (kleisli X).
unfold monotonic.
intros f g L. simpl. apply DLless_cond.
intros e C. destruct (kleisliValVal C) as [dd [P Q]].
rewrite (kleisli_Valeq f P). rewrite (kleisli_Valeq g P).
apply (fcont_le_elim L).
Defined.

Definition kleisli_c (D E: cpo) : (DL_ord D) -> (D -C-> DL E) -c> DL E.
intros D E f.
exists (kleisli_m E f).
unfold continuous.
intros h.
simpl.
apply DLless_cond.
intros xx C.
destruct (kleisliValVal C) as [f' [P Q]].
rewrite (kleisli_Valeq _ P). rewrite (fcont_lub_simpl h).
apply Ole_trans with (y:=lub (kleisli_m E f @ h)) ; auto.
apply (@lub_le_compat _ (h <__> f')). apply fmon_le_intro.
intros n. rewrite fmon_comp_simpl. rewrite (kleisli_Valeq (h n) P).
auto.
Defined.

Definition kleisli_mc (D E : cpo) : (DL_ord D) -m> (D -C-> DL E) -C-> DL E.
intros. exists (@kleisli_c D E).
unfold monotonic.
intros x y xyL h.
simpl. apply DLless_cond.
intros e C. destruct (kleisliValVal C) as [d [P Q]].
rewrite (kleisli_Valeq _ P). rewrite P in xyL.
destruct (DLvalval xyL) as [n [dd [X Y]]].
assert (y == Val dd) as vy. apply Oeq_sym. apply Val_exists_pred_eq.
exists n. apply X. rewrite (kleisli_Valeq _ vy).
assert (monotonic h) as M by auto. apply M. auto.
Defined.

Lemma eta_val: forall D f, eta D f = Val f.
auto.
Qed.

Definition kleisli_C (D E: cpo) : DL D -c> (D -C-> DL E) -C-> DL E.
intros D E. exists (kleisli_mc D E).
unfold continuous.
intros c.
simpl.
intro h. apply DLless_cond.
intros e C.
destruct (kleisliValVal C) as [d [V hd]].
rewrite (kleisli_Valeq _ V).
destruct (lubval V) as [k [dk [ck ld]]].
destruct (DLvalgetchain ck) as [cc ceq].
apply Ole_trans with (y:=lub (fcontit h @ cc)).
assert (continuous (fcontit h)) as ch by auto.
rewrite <- (lub_comp_eq _ ch).
assert (monotonic h) as mh by auto.
apply mh.
apply vleinj.
rewrite <- V.
rewrite <- eta_val. assert (continuous (fcontit (eta D))) as ce by auto.
rewrite (lub_comp_eq cc ce).
apply Ole_trans with (y:=lub c) ; auto.
rewrite (lub_lift_left c k).
refine (lub_le_compat _).
apply fmon_le_intro.
intros m. apply (proj1 (ceq m)).

apply Ole_trans with
    (y:= lub ((Fcontit (D -C-> DL E) (DL E) @ (kleisli_mc D E @ c)) <_> h)).
apply Oeq_le.
refine (Oeq_trans _ (Oeq_sym (lub_lift_left _ k))).
refine (@lub_eq_lift _ _ _ k _).
intros x kx. simpl. specialize (ceq x).
rewrite (kleisli_Valeq h ceq). auto.
auto.
Defined.

Definition kleisli_cc : forall D E, (D -C-> DL E) -c> DL D -C-> DL E :=
fun D E => curry (fcont_comp (uncurry _ _ _ (@kleisli_C D E)) (SWAP _ _)).

Add Parametric Morphism D E : (@kleisli_cc D E)
with signature (@Oeq (D -C-> DL E)) ==> (@Oeq (DL D -C-> DL E))
as kleisli_cc_eq_compat.
intros.
apply Ole_antisym ; auto.
Qed.

Lemma kleisli_C_simpl: forall D E f a, kleisli_C D E a f = kleisli a f.
intros. auto.
Qed.

Lemma kleisli_c_simpl: forall D E (f:D -C-> DL E) a, kleisli_c E a f = kleisli a f.
intros. auto.
Qed.

Lemma kleisli_cc_simpl: forall (D E:cpo) f (a:DL D), kleisli_cc D E f a = kleisli a f.
intros ; auto.
Qed.

Lemma kleisli_comp: forall D E F (f:DL E -c> DL F) (g:D -c> DL E) d,
   (forall x xx, f x == Val xx -> exists y, x == Val y) ->
         kleisli d (f @_ g) == f (kleisli d g).
intros D E F f g d S. split. simpl.
apply DLless_cond.

intros ff kv.
destruct (kleisliValVal kv) as [dd [dv fdd]].
rewrite (kleisli_Valeq (f @_ g) dv). rewrite fdd.
assert (f (g dd) == f (kleisli d g)).
assert (stable f) as sf. auto.
unfold stable in sf. apply sf.
rewrite (kleisli_Valeq g dv). auto.
rewrite <- H. rewrite fcont_comp_simpl in fdd. auto.

simpl. apply DLless_cond.
intros xx fx.
specialize (S _ _ fx).
destruct S as [e ke].
destruct (kleisliValVal ke) as [dd [de ge]].
rewrite (kleisli_Valeq (f @_ g) de).
rewrite fcont_comp_simpl. 
assert (monotonic f) as M by auto.
apply M. rewrite <- ge in ke.
auto.
Qed.

Lemma kleisli_cc_comp: forall D E F (f:D -c> DL E) (g:E -c> DL F),
       kleisli_cc _ _ g @_ kleisli_cc _ _ f == kleisli_cc _ _ (kleisli_cc _ _ g @_ f).
intros.
refine (fcont_eq_intro _).
intros d. rewrite fcont_comp_simpl.
repeat (rewrite kleisli_cc_simpl).
rewrite kleisli_comp.
rewrite kleisli_cc_simpl. auto.
intros e ff. rewrite kleisli_cc_simpl.
intros kl. destruct (kleisliValVal kl) as [dd [edd gd]].
exists dd. auto.
Qed.

Lemma kleisli_leq: forall (D E F:cpo) a b (f:D -C-> DL E) (g: F -C-> DL E),
      (forall aa, a == Val aa -> exists bb, b == Val bb /\ f aa <= g bb) ->
   @kleisli D E a f <= @kleisli F E b g.
Proof.
intros D E F a b f g V. simpl. apply DLless_cond.
intros xx [_ kl].
destruct (DLle_Val_exists_pred kl) as [n [dd [pd ld]]].
destruct (kleislipredValVal pd) as [aa [va fa]].
rewrite (kleisli_Valeq f va). specialize (V _ va).
destruct V as [bb [vb fg]]. rewrite (kleisli_Valeq g vb). apply fg.
Qed.

Lemma kleisli_eq: forall (D E F:cpo) a b (f:D -C-> DL E) (g: F -C-> DL E),
      (forall aa, a == Val aa -> exists bb, b == Val bb /\ f aa == g bb) ->
      (forall bb, b == Val bb -> exists aa, a == Val aa /\ f aa == g bb) ->
   @kleisli D E a f == @kleisli F E b g.
Proof.
intros D E F a b f g V1 V2.
apply Ole_antisym.
apply kleisli_leq.
intros aa va. destruct (V1 _ va) as [bb [vb fa]].
exists bb. auto.
apply kleisli_leq. intros bb vb.
destruct (V2 _ vb) as [aa [f1 f2]].
exists aa. auto.
Qed.

Lemma kleisli_eta_com: forall D E f, kleisli_cc D E f @_ eta _ == f.
intros D E f. refine (fcont_eq_intro _).
intros d. rewrite fcont_comp_simpl. rewrite kleisli_cc_simpl.
rewrite eta_val. rewrite kleisli_Val. auto.
Qed.

Lemma kleisli_unit: forall D d, kleisli d (eta D) == d.
intros D d.
split. simpl. apply DLless_cond.

intros dd kl. destruct (kleisliValVal kl) as [d1 [P1 P2]].
rewrite eta_val in P2. rewrite P2 in P1.
rewrite (kleisli_Valeq (eta D) P1). auto.

simpl. apply DLless_cond.
intros dd kl.
rewrite (kleisli_Valeq (eta D) kl). auto.
Qed.

Lemma kleisli_cc_unit: forall D, kleisli_cc D D (eta _) == ID _.
intros D. refine (fcont_eq_intro _).
intros d. rewrite kleisli_cc_simpl. rewrite ID_simpl. simpl.
assert (xx := kleisli_unit). simpl. simpl in xx. apply xx.
Qed.

Lemma kleisli_cc_comp2: forall D E F (f:D -C-> E) (g:E -C-> DL F),
         kleisli_cc _ _ (g @_ f) == kleisli_cc _ _ g @_ kleisli_cc _ _ (eta _ @_ f).
intros D E F f g.
refine (fcont_eq_intro _).
intros d. repeat (rewrite kleisli_cc_simpl).
rewrite fcont_comp_simpl. repeat (rewrite kleisli_cc_simpl).
refine (kleisli_eq _ _).
intros dd dv. exists (f dd). rewrite fcont_comp_simpl. split ; auto.
rewrite (kleisli_Valeq (eta _ @_ f) dv). rewrite fcont_comp_simpl. auto.
intros ee kl.
destruct (kleisliValVal kl) as [dd P]. exists dd.
split. destruct P ; auto.
rewrite fcont_comp_simpl in *. destruct P as [_ P].
rewrite eta_val in P.
assert (stable g) as sg by auto. apply sg.
apply (vinj P).
Qed.

Definition mu D := kleisli_cc _ _ (ID (DL D)).

Lemma mu_natural D E (f:D -c> E) :
   kleisli_cc _ _ (eta _ @_ f) @_ mu _ == mu _ @_ kleisli_cc _ _ (eta _ @_ (kleisli_cc _ _ (eta _ @_ f))).
intros D E f.
unfold mu. rewrite <- kleisli_cc_comp2. rewrite kleisli_cc_comp.
rewrite fcont_comp_unitL. rewrite fcont_comp_unitR.
auto.
Qed.

Lemma mumu D : mu _ @_ kleisli_cc _ _ (eta _ @_ mu _) == mu D @_ mu _.
intros D. unfold mu. rewrite <- kleisli_cc_comp2. rewrite kleisli_cc_comp.
repeat (rewrite fcont_comp_unitL). rewrite fcont_comp_unitR. auto.
Qed.

Lemma mukl D : mu D @_ kleisli_cc _ _ (eta _ @_ eta _) == ID _.
intros D. unfold mu.
rewrite <- kleisli_cc_comp2. rewrite fcont_comp_unitL. rewrite kleisli_cc_unit. auto.
Qed.

Add Parametric Morphism (D E:cpo) : (@kleisli D E)
with signature (@Oeq (DL D)) ==> (@Oeq (D -C-> DL E)) ==> (@Oeq (DL E))
as kleisli_eq_compat.
intros d0 d1 deq f0 f1 feq.
repeat (rewrite <- kleisli_cc_simpl).
assert (monotonic (kleisli_cc D E f0)) as M1 by auto.
assert (monotonic (kleisli_cc D E f1)) as M0 by auto.
assert (monotonic (kleisli_cc D E)) as M2 by auto.
unfold monotonic in *.
destruct deq as [eq1 eq2].
destruct feq as [feq1 feq2].
apply Ole_antisym.
eapply Ole_trans.
apply (M1 _ _ eq1).
apply (M2 _ _ feq1).
eapply Ole_trans.
apply (M0 _ _ eq2).
apply (M2 _ _ feq2).
Qed.

Hint Resolve kleisli_eq_compat_Morphism.

Add Parametric Morphism (D E:cpo) : (@kleisli_C D E)
with signature (@Oeq (DL D)) ==> (@Oeq ((D -C-> DL E) -C-> (DL E)))
as kleisli_C_eq_compat.
intros d0 d1 dEq.
assert (monotonic (kleisli_C D E)) as M. auto.
unfold monotonic in M.
apply Ole_antisym.
specialize (M d0 d1) ; auto.
specialize (M d1 d0) ; auto.
Qed.

Add Parametric Morphism (D E:cpo) : (@kleisli_c D E)
with signature (@Oeq (DL D)) ==> (@Oeq (D -C-> DL E)) ==> (@Oeq (DL E))
as kleisli_c_eq_compat.
intros d0 d1 dEq f0 f1 fEq.
assert (monotonic (@kleisli_mc D E)) by auto.
simpl in H. unfold monotonic in H.
apply Ole_antisym.
specialize (H d0 d1).
assert (d0 <= d1). auto.
specialize (H H0).
assert (M := fcont_le_elim H).
specialize (M f0).
eapply Ole_trans. apply M.
assert (monotonic (kleisli_c E d0)) as M1. auto.
unfold monotonic in M1. specialize (M1 f0 f1). auto.

specialize (H d1 d0).
assert (d1 <= d0). auto.
specialize (H H0).
assert (M := fcont_le_elim H).
specialize (M f1).
eapply Ole_trans. apply M.
assert (monotonic (kleisli_c E d0)) as M1. auto.
unfold monotonic in M1. specialize (M1 f1 f0). auto.
Qed.

(* Constant maps are continuous *)
Definition kab (O1 O2:ord) (b : O2): O1-m>O2.
intros; exists (fun a => b).
red. auto.
Defined.

Definition K (D1 D2 : cpo) (d:D2) : D1 -c> D2.
intros; exists (kab D1 d).
red.
intros.
simpl.
setoid_replace d with ((kab D1 (O2:=D2) d @ h) 0).
apply le_lub.
simpl.
trivial.
Defined.


(* Now that I've got bottomless cpos, I need ones with bottom for fixpoints.
   Could work with things that are lifted, but that's a bit literal.
   Could work with algebras for the lift monad, but that's probably going to be heavy in Coq
   So I'll just add a new class of things with bottom, like Christine's original cpos but with 
   a new layer of stuff. Hope I can do enough implicit stuff not to make this too disgusting.
*)
(*
Record pcpo : Type := mk_pcpo 
  {tpcpo:>cpo; D0 : tpcpo; Dbot : forall x:tpcpo, D0 <= x}.


Definition Lift (D:cpo) : pcpo.
intro D; exists (DL D) (DL_bot D).
intro.
red. simpl.
apply (DL_bot_least).
Defined.
*)

Lemma PROD_fun_compl (D E F G:cpo) (f : D -C-> E) (g:D -C-> F) (h:G -C-> D) : 
        PROD_fun f g @_ h == PROD_fun (f @_ h) (g @_ h).
intros D E F G f g h.
simpl. apply (@Dprod_unique E F G).
rewrite FST_PROD_fun. rewrite <-fcont_comp_assoc. rewrite FST_PROD_fun. auto.
rewrite SND_PROD_fun. rewrite <- fcont_comp_assoc. rewrite SND_PROD_fun. auto.
Qed.

Lemma PROD_fun_compr (D E F G H : cpo) (f:H -c> D) (g : D -c> E) (h:H -c> F) (k:F -c> G) :
       PROD_fun (g @_ f) (k @_ h) == PROD_map g k @_ PROD_fun f h.
intros D E F G H f g h k. apply (@Dprod_unique E G H).
rewrite FST_PROD_fun. rewrite <-fcont_comp_assoc. rewrite PROD_map_fun.
rewrite FST_PROD_fun. rewrite fcont_comp_assoc. rewrite FST_PROD_fun. auto.
rewrite SND_PROD_fun. rewrite <-fcont_comp_assoc. rewrite PROD_map_fun.
rewrite SND_PROD_fun. rewrite fcont_comp_assoc. rewrite SND_PROD_fun. auto.
Qed.


Record alg : Type := mk_alg
  { acpo :> cpo; strength : DL acpo -c> acpo;
    st_eta : strength @_ eta _ == ID _;
    st_kl : strength @_ kleisli_cc _ _ (eta _ @_ strength) == strength @_ kleisli_cc _ _ (ID _) }.

Definition Aprod : alg -> alg -> alg.
intros A1 A2. exists (Dprod (acpo A1) (acpo A2))
     (PROD_fun (strength A1 @_ (kleisli_cc _ _ (eta _ @_ FST _ _)))
               (strength A2 @_ (kleisli_cc _ _ (eta _ @_ SND _ _)))).
rewrite PROD_fun_compl.
apply (@Dprod_unique A1 A2 (Dprod A1 A2)). rewrite FST_PROD_fun.
rewrite fcont_comp_unitR. rewrite fcont_comp_assoc.
rewrite kleisli_eta_com. rewrite <- fcont_comp_assoc. rewrite st_eta. rewrite fcont_comp_unitL. auto.
rewrite SND_PROD_fun.
rewrite fcont_comp_unitR. rewrite fcont_comp_assoc.
rewrite kleisli_eta_com. rewrite <- fcont_comp_assoc. rewrite st_eta. rewrite fcont_comp_unitL. auto.

refine (@Dprod_unique (acpo A1) (acpo A2) _ _ _ _ _). rewrite <- fcont_comp_assoc.
rewrite FST_PROD_fun. rewrite <- fcont_comp_assoc. rewrite FST_PROD_fun.
rewrite fcont_comp_assoc. rewrite fcont_comp_assoc.
rewrite kleisli_cc_comp. rewrite <- fcont_comp_assoc. rewrite kleisli_eta_com.
rewrite fcont_comp_assoc. rewrite FST_PROD_fun. rewrite kleisli_cc_comp.
rewrite fcont_comp_unitR.
rewrite <- fcont_comp_assoc. rewrite kleisli_cc_comp2. rewrite <- fcont_comp_assoc.
rewrite st_kl. rewrite fcont_comp_assoc. rewrite <- kleisli_cc_comp2. rewrite fcont_comp_unitL.
auto.

rewrite <- fcont_comp_assoc.
rewrite SND_PROD_fun. rewrite <- fcont_comp_assoc. rewrite SND_PROD_fun.
rewrite fcont_comp_assoc. rewrite fcont_comp_assoc.
rewrite kleisli_cc_comp. rewrite <- fcont_comp_assoc. rewrite kleisli_eta_com.
rewrite fcont_comp_assoc. rewrite SND_PROD_fun. rewrite kleisli_cc_comp.
rewrite fcont_comp_unitR.
rewrite <- fcont_comp_assoc. rewrite kleisli_cc_comp2. rewrite <- fcont_comp_assoc.
rewrite st_kl. rewrite fcont_comp_assoc. rewrite <- kleisli_cc_comp2. rewrite fcont_comp_unitL.
auto.
Defined.

Definition AF : cpo -> alg.
intros D. exists (DL D) (kleisli_cc _ _ (ID _)).
rewrite kleisli_eta_com. auto.
rewrite <- kleisli_cc_comp2. rewrite fcont_comp_unitL.
rewrite kleisli_cc_comp. rewrite fcont_comp_unitR. auto.
Defined.

Definition Application := fun a b =>
    curry ((kleisli_C (a -C-> DL b) b @2_ (FST _ _)) (kleisli_C a b @_ SND _ _)).

Definition Operator2 := fun A B C (F:Dprod A B -C-> DL C) =>
  (Application _ _) @_ (kleisli_cc _ _ (eta _ @_ (curry F))).

Definition Smash := fun A B => @Operator2 _ _ (Dprod A B) (eta _).

Definition LStrength := fun A B => uncurry _ _ _ (Smash _ _) @_ PROD_map (eta A) (ID (DL B)).

Definition RStrength := fun A B => uncurry _ _ _ (Smash _ _) @_ PROD_map (ID (DL A)) (eta B).

Definition KLEISLIR := fun A B C f => kleisli_cc (Dprod A B) C f @_ LStrength A B.
Definition KLEISLIL := fun A B C f => kleisli_cc (Dprod A B) C f @_ RStrength A B.

Lemma uncurry_simpl: forall D E F f p1 p2, uncurry D E F f (p1,p2) = f p1 p2.
intros ; auto.
Qed.

Lemma Operator2Val: forall (D E F:cpo) (h:Dprod D E -C-> DL F) d e f,
     Operator2 h d e == Val f -> exists d', exists e', d == Val d' /\ e == Val e' /\ h (d',e') == Val f.
intros D E F h d e f.
unfold Operator2. unfold Application. unfold uncurry.
repeat (try (rewrite fcont_comp_simpl) ; try (rewrite curry_simpl) ; try (rewrite kliesli_cc_simpl) ;
        try (rewrite PROD_map_simpl) ; simpl ; try (rewrite FST_simpl) ; try (rewrite SND_simpl) ;
        try (rewrite kleisli_C_simpl) ; try (rewrite fcont_comp2_simpl) ; try (rewrite kleisli_c_simpl) ;
        try (rewrite ID_simpl)).
intros c.
destruct (kleisliValVal c) as [f' [cc1 cc2]].
rewrite kleisli_C_simpl in *. rewrite kleisli_cc_simpl in *.
destruct (kleisliValVal cc2) as [e' [eeq feeq]]. clear cc2. clear c.
destruct (kleisliValVal cc1) as [d' [deq deeq]]. clear cc1.
rewrite fcont_comp_simpl in *.
exists d'. exists e'. split. auto. split. auto.
rewrite eta_val in *. assert (xx:=fcont_eq_elim (vinj deeq) e'). clear deeq deq eeq. rewrite <- feeq.
rewrite curry_simpl in xx. auto.
Qed.

Lemma Operator2_simpl: forall (E F D:cpo) (f:Dprod E F -C-> DL D) v1 v2,
          Operator2 f (Val v1) (Val v2) == f (v1,v2).
intros E F D f.
intros v1 v2.
unfold Operator2. unfold Application. unfold uncurry.
repeat (try (rewrite fcont_comp_simpl) ; try (rewrite curry_simpl) ;
        try (rewrite fcont_comp2_simpl) ;
        try (rewrite PROD_map_simpl) ; try (rewrite FST_simpl) ;
        try (rewrite SND_simpl) ; try (rewrite ID_simpl) ;
        try (rewrite kleisli_cc_simpl) ; simpl).
rewrite kleisli_C_simpl.
refine (Oeq_trans (kleisli_eq_compat _ (Oeq_refl _)) _).
rewrite kleisli_Val.
rewrite fcont_comp_simpl.
rewrite eta_val.
apply Oeq_refl. rewrite kleisli_Val. rewrite kleisli_C_simpl. rewrite kleisli_Val.
rewrite curry_simpl. auto.
Qed.

Lemma KLEISLIL_eq (A A' B B' C:cpo) : forall (f:Dprod A B -C-> DL C) (f':Dprod A' B' -C-> DL C)
            a b a' b',
    (forall aa aa', a == Val aa -> a' == Val aa' -> f (aa,b) == f' (aa',b')) ->
    (forall aa, a == Val aa -> exists aa', a' == Val aa') ->
    (forall aa', a' == Val aa' -> exists aa, a == Val aa) ->
  @KLEISLIL A B C f (a,b) == @KLEISLIL A' B' C f' (a',b').
Proof.
intros A A' B B' C f f' a b a' b' V1 V2 V3.
unfold KLEISLIL. unfold kleisli_cc. intros.
unfold RStrength. repeat (rewrite fcont_comp_simpl).
repeat (rewrite PROD_map_simpl). simpl. repeat (rewrite ID_simpl). simpl.
repeat (rewrite curry_simpl). repeat (rewrite fcont_comp_simpl).
unfold SWAP.
refine (Oeq_trans (app_eq_compat (Oeq_refl _) (PROD_fun_simpl _ _ _)) _).
refine (Oeq_trans _ (app_eq_compat (Oeq_refl _) (Oeq_sym (PROD_fun_simpl _ _ _)))).
repeat (rewrite FST_simpl ; rewrite SND_simpl ; rewrite PAIR_simpl). simpl.
repeat (rewrite uncurry_simpl). repeat (rewrite kleisli_C_simpl). unfold Smash.
refine (kleisli_eq _ _).
intros aa sa.
destruct (Operator2Val sa) as [a1 [b1 [Pa1 [Pb1 pv]]]].
destruct (V2 _ Pa1) as [aa' Paa].
exists (aa', b'). specialize (V1 a1 aa' Pa1 Paa). rewrite <- V1.
rewrite eta_val in *.
split.
refine (Oeq_trans (app_eq_compat (app_eq_compat (Oeq_refl (Operator2 (eta _))) Paa) (Oeq_refl _)) _).
rewrite Operator2_simpl. auto.
auto.
refine (app_eq_compat (Oeq_refl _) _).
refine (Oeq_trans (Oeq_sym (vinj pv)) _).
refine (PAIR_eq_compat (Oeq_refl _) _).
apply (Oeq_sym (vinj Pb1)).

intros bb ob. destruct (Operator2Val ob) as [aa' [bb' [av P]]].
specialize (V3 _ av). destruct V3 as [aa va].
specialize (V1 _ _ va av).
exists (aa,b).
split. rewrite eta_val.
refine (Oeq_trans (app_eq_compat (app_eq_compat (Oeq_refl (Operator2 (eta (Dprod A B)))) va) (Oeq_refl (Val b))) _).
rewrite Operator2_simpl. rewrite eta_val. auto.
refine (Oeq_trans V1 _).
refine (app_eq_compat (Oeq_refl _) _).
destruct P as [P1 P2]. rewrite eta_val in P2. refine (Oeq_trans _ (vinj P2)).
rewrite eta_val in P1.
refine (PAIR_eq_compat (Oeq_refl _) (vinj P1)).
Qed.

Lemma KLEISLIR_eq (A A' B B' C:cpo) : forall (f:Dprod A B -C-> DL C) (f':Dprod A' B' -C-> DL C)
            a b a' b',
    (forall aa aa', b == Val aa -> b' == Val aa' -> f (a,aa) == f' (a',aa')) ->
    (forall aa, b == Val aa -> exists aa', b' == Val aa') ->
    (forall aa', b' == Val aa' -> exists aa, b == Val aa) ->
  @KLEISLIR A B C f (a,b) == @KLEISLIR A' B' C f' (a',b').
Proof.
intros A A' B B' C f f' a b a' b' V1 V2 V3.
unfold KLEISLIR. unfold kleisli_cc. intros.
unfold LStrength. repeat (rewrite fcont_comp_simpl).
repeat (rewrite PROD_map_simpl). repeat (rewrite ID_simpl).
repeat (rewrite curry_simpl). repeat (rewrite fcont_comp_simpl).
unfold SWAP.
refine (Oeq_trans (app_eq_compat (Oeq_refl _) (PROD_fun_simpl _ _ _)) _).
refine (Oeq_trans _ (app_eq_compat (Oeq_refl _) (Oeq_sym (PROD_fun_simpl _ _ _)))).
repeat (rewrite FST_simpl ; rewrite SND_simpl ; rewrite PAIR_simpl). simpl.

repeat (rewrite uncurry_simpl). repeat (rewrite kleisli_C_simpl). unfold Smash.
refine (kleisli_eq _ _).
intros aa sa.
destruct (Operator2Val sa) as [a1 [b1 [Pa1 [Pb1 pv]]]].
destruct (V2 _ Pb1) as [aa' Paa].
exists (a',aa'). specialize (V1 _ aa' Pb1 Paa). rewrite <- V1.
rewrite eta_val in *.
split.
refine (Oeq_trans (app_eq_compat (Oeq_refl (Operator2 (eta _) (Val a'))) (Paa )) _).
rewrite Operator2_simpl. auto.
refine (app_eq_compat (Oeq_refl _) _).
refine (Oeq_trans (Oeq_sym (vinj pv)) _).
refine (PAIR_eq_compat _ (Oeq_refl _)).
apply (Oeq_sym (vinj Pa1)).

intros bb ob. destruct (Operator2Val ob) as [aa' [bb' [av [bv P]]]].
specialize (V3 _ bv). destruct V3 as [aa va].
specialize (V1 _ _ va bv).
exists (a,aa).
split. rewrite eta_val.
refine (Oeq_trans (app_eq_compat (Oeq_refl (Operator2 (eta (Dprod A B)) (Val a))) va) _).
rewrite Operator2_simpl. rewrite eta_val. auto.
refine (Oeq_trans V1 _).
refine (app_eq_compat (Oeq_refl _) _).
rewrite eta_val in P. refine (Oeq_trans _ (vinj P)).
rewrite eta_val in av.
refine (PAIR_eq_compat (vinj av) (Oeq_refl _)).
Qed.


Add Parametric Morphism (D E F:cpo) : (@KLEISLIL D E F)
with signature (@Oeq (Dprod D E -C-> DL F)) ==> (@Oeq (Dprod (DL D) E -c> DL F))
as KLEISLIL_eq_compat.
intros f0 f1 feq.
apply fcont_eq_intro. intros de. simpl.
unfold KLEISLIL. repeat (rewrite fcont_comp_simpl).
repeat (rewrite kleisli_cc_simpl). refine (kleisli_eq_compat (Oeq_refl _) _). auto.
Qed.

Add Parametric Morphism (D E F:cpo) : (@KLEISLIR D E F)
with signature (@Oeq (Dprod D E -C-> DL F)) ==> (@Oeq (Dprod D (DL E) -c> DL F))
as KLEISLIR_eq_compat.
intros f0 f1 feq.
apply fcont_eq_intro. intros de. simpl.
unfold KLEISLIR. repeat (rewrite fcont_comp_simpl).
repeat (rewrite kleisli_cc_simpl). refine (kleisli_eq_compat (Oeq_refl _) _). auto.
Qed.

Lemma KLEISLIL_comp: forall D E F G (f:Dprod D E -C-> DL F) g h,
        KLEISLIL f @_ PROD_fun g h == KLEISLIL (f @_ PROD_map (ID _) h) @_ PROD_fun g (ID G).
intros.
apply fcont_eq_intro.
intros gg. rewrite fcont_comp_simpl. unfold PROD_fun. rewrite fcont_comp_simpl. rewrite duplicate_simpl.
rewrite PROD_map_simpl. simpl. repeat (rewrite fcont_comp_simpl). rewrite duplicate_simpl. rewrite PROD_map_simpl.
simpl. refine (KLEISLIL_eq _ _ _).
intros d0 d1 vd0 vd1. rewrite fcont_comp_simpl. rewrite PROD_map_simpl. simpl. repeat (rewrite ID_simpl). simpl.
rewrite vd0 in vd1. assert (vv:=vinj vd1). assert (stable f) as sf by auto. unfold stable in sf.
apply (sf (d0,h gg) (d1,h gg)). apply PAIR_eq_compat. auto. auto.
intros d0 vd0. exists d0. auto. intros aa vv. eexists ; apply vv.
Qed.

Lemma KLEISLIR_comp: forall D E F G (f:Dprod D E -C-> DL F) g h,
        KLEISLIR f @_ PROD_fun h g == KLEISLIR (f @_ PROD_map h (ID _)) @_ PROD_fun (ID G) g.
intros.
apply fcont_eq_intro.
intros gg. rewrite fcont_comp_simpl. unfold PROD_fun. rewrite fcont_comp_simpl. rewrite duplicate_simpl.
rewrite PROD_map_simpl. simpl. repeat (rewrite fcont_comp_simpl). rewrite duplicate_simpl. rewrite PROD_map_simpl.
simpl. refine (KLEISLIR_eq _ _ _).
intros d0 d1 vd0 vd1. rewrite fcont_comp_simpl. rewrite PROD_map_simpl. simpl. repeat (rewrite ID_simpl). simpl.
rewrite vd0 in vd1. assert (vv:=vinj vd1). assert (stable f) as sf by auto. unfold stable in sf.
apply (sf (h gg,d0) (h gg,d1)). apply PAIR_eq_compat. auto. auto.
intros d0 vd0. exists d0. auto. intros aa vv. eexists ; apply vv.
Qed.

Lemma KLEISLIL_unit: forall D E F G (f:Dprod D E -C-> DL F) (g:G -C-> D) h,
    KLEISLIL f @_ PROD_fun (eta _ @_ g) h == f @_ PROD_fun g h.
intros.
unfold KLEISLIL. unfold RStrength.
repeat (rewrite fcont_comp_assoc).
refine (fcont_eq_intro _). intros gg. repeat (rewrite fcont_comp_simpl). unfold PROD_fun. rewrite PROD_map_simpl. simpl.
rewrite uncurry_simpl. unfold Smash. rewrite ID_simpl. rewrite eta_val. simpl.
rewrite kleisli_cc_simpl. refine (Oeq_trans (kleisli_eq_compat (Operator2_simpl _ _ _) (Oeq_refl _)) _).
rewrite eta_val. rewrite kleisli_Val. auto.
Qed.

Lemma KLEISLIR_unit: forall D E F G (f:Dprod E D -C-> DL F) (g:G -C-> D) h,
    KLEISLIR f @_ PROD_fun h (eta _ @_ g) == f @_ PROD_fun h g.
intros.
unfold KLEISLIR. unfold LStrength.
repeat (rewrite fcont_comp_assoc).
refine (fcont_eq_intro _). intros gg. repeat (rewrite fcont_comp_simpl). unfold PROD_fun. rewrite PROD_map_simpl. simpl.
rewrite uncurry_simpl. unfold Smash. rewrite ID_simpl. rewrite eta_val. simpl.
rewrite kleisli_cc_simpl. refine (Oeq_trans (kleisli_eq_compat (Operator2_simpl _ _ _) (Oeq_refl _)) _).
rewrite eta_val. rewrite kleisli_Val. auto.
Qed.

Lemma curry_compr A B C D (f:Dprod A B -c> C) (g:D -c> A) :
    curry f @_ g == curry (f @_ PROD_map g (ID _)).
intros A B C D f g.
apply curry_unique.
assert (PROD_map (curry f @_ g) (ID B) == PROD_map (curry f) (ID B) @_ PROD_map g (ID B)) as E.
apply (@Dprod_unique (B -C-> C) B (Dprod D B)).
repeat (rewrite PROD_map_fun). rewrite FST_PROD_fun. rewrite <- fcont_comp_assoc.
rewrite FST_PROD_fun. repeat (rewrite fcont_comp_assoc). rewrite FST_PROD_fun. auto.
repeat (rewrite PROD_map_fun). rewrite SND_PROD_fun. rewrite <- fcont_comp_assoc.
rewrite SND_PROD_fun. repeat (rewrite fcont_comp_assoc). rewrite SND_PROD_fun.
repeat (rewrite fcont_comp_unitL). auto.
rewrite E. clear E. rewrite <- fcont_comp_assoc. rewrite <- curry_com.
auto.
Qed.

(*
Definition fcont_alg : cpo -> alg -> alg.
intros D A. exists (D -C-> A) (curry (strength A @_ (KLEISLIL (eta _ @_ (EV D A))))).
apply fcont_eq_intro. intros f.
rewrite fcont_comp_simpl. rewrite ID_simpl. simpl.
apply fcont_eq_intro. intros d. rewrite curry_simpl. rewrite fcont_comp_simpl.
assert (kl:=fcont_eq_elim (@KLEISLIL_unit _ _ _ _ (eta A @_ EV D A) (K _ f) (ID _)) d).
repeat (rewrite fcont_comp_simpl in kl).
assert ((PROD_fun (A:=DL (D -C-> A)) (B:=D) (C:=D)
            (eta (D -C-> A) @_ K D (D2:=D -C-> A) f) 
            (ID D) d) == (eta (D -C-> A) f, d)) as F by auto.
assert (X:=app_eq_compat (Oeq_refl (KLEISLIL (eta A @_ EV D A))) F).
rewrite X in kl. clear F X.
rewrite (app_eq_compat (Oeq_refl (strength A)) kl). clear kl.
assert (X:=fcont_eq_elim (st_eta A) (EV D A
           (PROD_fun (A:=D -C-> A) (B:=D) (C:=D) (K D (D2:=D -C-> A) f)
              (ID D) d))). auto.

rewrite curry_compr.
refine (Oeq_sym _).
apply curry_unique.
rewrite fcont_comp_assoc.




apply fcont_eq_intro.
intros f''. repeat (rewrite fcont_comp_simpl).
refine (fcont_eq_intro _). intros d.
repeat (rewrite curry_simpl).
repeat (rewrite fcont_comp_simpl).
assert (X:=st_kl A).

*)

(** Discrete cpo on a Set. This should really be an arbitrary Setoid, I guess. *)
Definition Discrete_ord (X:Type) : ord.
intro X.
exists X (fun x y => x=y); intuition.
subst.
trivial.
Defined.

Definition Discrete (X:Type) : cpo.
intro X; exists (Discrete_ord X) (fun (c:natO -m> Discrete_ord X) => c 0).
 intros; assert (f 0 <= f n).
 apply fmonotonic.
 intuition.
 intuition.

 intros.
 apply H.
Defined.

(* Messing with classes *)
Class Pointed (D:cpo) := {
 PBot : D;
 Pleast : forall d:D, PBot <= d }.

Hint Resolve @Pleast.

Instance lift_pointed D : Pointed (DL D).
intro D.
exists (DL_bot D).
intro. red. simpl.
apply DL_bot_least.
Defined.

Definition strict D E {pd:Pointed D} {pe:Pointed E} (f:D -C-> E) := f PBot == PBot.

Record fstricti (D1 D2 : cpo) (PD1:Pointed D1) (PD2:Pointed D2) : Type
  := mk_fstricti {fstrictit : D1 -c> D2; fstrict : strict fstrictit}.

Definition fstricti_fun (D1 D2 : cpo) (PD1:Pointed D1) (PD2:Pointed D2) (f:fstricti PD1 PD2) : D1 -> D2 := fun x =>
fstrictit f x.
Coercion fstricti_fun : fstricti >-> Funclass.

Definition fstrict_ord (D1 D2:cpo) (PD1:Pointed D1) (PD2:Pointed D2) : ord.
intros D1 D2 PD1 PD2.
exists (fstricti PD1 PD2) (fun (f g:fstricti PD1 PD2) => fstrictit f <= fstrictit g).
intros; auto.
intros; auto.
apply Ole_trans with (fstrictit y); auto.
Defined.

Infix "-s>" := fstrict_ord (at level 30, right associativity) : O_scope.

(* Now Christine has a whole bunch of auxiliary lemmas for continuous
   maps, which I copy here, assuming they'll turn out to be useful
*)

Lemma fstrict_le_intro : forall (D1 D2:cpo) (PD1:Pointed D1) (PD2:Pointed D2) (f g : PD1 -s> PD2), (forall x, f x <=
 g x) -> f <= g.
trivial.
Save.

Lemma fstrict_le_elim : forall (D1 D2:cpo) (PD1:Pointed D1) (PD2:Pointed D2) (f g : PD1 -s> PD2), f <= g -> forall x
, f x <= g x.
trivial.
Save.

Lemma fstrict_eq_intro : forall (D1 D2:cpo) (PD1:Pointed D1) (PD2:Pointed D2) (f g : PD1 -s> PD2), (forall x, f x ==
 g x) -> f == g.
intros; apply Ole_antisym; apply fstrict_le_intro; auto.
Save.

Lemma fstrict_eq_elim : forall (D1 D2:cpo) (PD1:Pointed D1) (PD2:Pointed D2) (f g : PD1 -s> PD2), f == g -> forall x
, f x == g x.
intros; apply Ole_antisym; apply fstrict_le_elim; auto.
Save.

Lemma fstrict_monotonic : forall (D1 D2:cpo) (PD1:Pointed D1) (PD2:Pointed D2) (f : PD1 -s> PD2) (x y : D1),
            x <= y -> f x <= f y.
intros; apply (fmonotonic (fcontit (fstrictit f)) H).
Save.
Hint Resolve fstrict_monotonic.

Lemma fstrict_stable : forall (D1 D2:cpo) (PD1:Pointed D1) (PD2:Pointed D2) (f : PD1 -s> PD2) (x y : D1),
            x == y -> f x == f y.
intros; apply (fmon_stable (fcontit (fstrictit f)) H).
Save.
Hint Resolve fstrict_monotonic.

Lemma strict_lub  : forall (D1 D2 : cpo) (PD1:Pointed D1) (PD2:Pointed D2)  (f:natO -m> (D1 -c> D2)),
                                        (forall n, strict (f n)) ->
                                         strict  (lub (c:=D1-C->D2) f).
intros. unfold strict.
rewrite fcont_lub_simpl.
split ; auto.
apply lub_le.
intro n.
rewrite fcont_app_simpl.
unfold strict in H.
rewrite (H n).
auto.
Defined.

Definition Fstrictit (D1 D2 : cpo) (PD1:Pointed D1) (PD2:Pointed D2) : (PD1 -s> PD2)-m> (D1-c> D2).
intros D1 D2 PD1 PD2; exists (fstrictit (D1:=D1) (D2:=D2) (PD1:=PD1) (PD2:=PD2)).
auto.
Defined.

Definition fstrict_lub (D1 D2 : cpo) (PD1:Pointed D1) (PD2:Pointed D2) : (natO -m> PD1 -s> PD2) -> (PD1 -s> PD2).
intros D1 D2 PD1 PD2 f.
exists (lub (c:= D1 -C->D2) (Fstrictit PD1 PD2 @ f)).
apply strict_lub.
intros; simpl.
apply (fstrict (f n)).
Defined.

Definition fstrict_cpo (D1 D2:cpo) (PD1:Pointed D1) (PD2:Pointed D2) : cpo.
intros D1 D2 PD1 PD2; exists (fstrict_ord PD1 PD2)
                     (fstrict_lub (D1:=D1) (D2:=D2) (PD1:=PD1) (PD2:=PD2));
simpl;intros; auto.
apply le_lub with (f := ((Fcontit D1 D2 @ (Fstrictit PD1 PD2 @ f)) <_> x))
                  (n := n).
Defined.

Infix "-S->" := fstrict_cpo (at level 30, right associativity) : O_scope.

Definition fstrict_comp : forall (D1 D2 D3:cpo) {PD1:Pointed D1} {PD2:Pointed D2} {PD3:Pointed D3}, 
    (PD2 -s> PD3) -> (PD1-s> PD2) -> PD1 -s> PD3.
intros D1 D2 D3 PD1 DP2 PD3 f g.
exists ((fstrictit f) @_ (fstrictit g)).
red.
rewrite fcont_comp_simpl.
apply Oeq_trans with (fstrictit f (PBot)).
apply fcont_stable.
apply (fstrict g).
apply (fstrict f).
Defined.




Lemma strictVal: forall D E (f:DL D -c> DL E), (forall d e, f d == Val e -> exists dd, d == Val dd) -> strict f.
intros. unfold strict.
split ; auto.

simpl. apply DLless_cond.
intros xx Fx. specialize (H _ _ Fx). destruct H as [d incon].
simpl in incon.
assert (yy := eqValpred incon).
destruct yy as [n [dd [pp _]]].
assert False as F. induction n.
simpl in pp. rewrite DL_bot_eq in pp. inversion pp.
rewrite DL_bot_eq in pp. simpl in pp. auto. inversion F.
Qed.

Lemma kleisli_bot: forall D E (f:D -c> DL E), kleisli (@PBot (DL D) _) f == PBot.
intros.
refine (strictVal (f:=kleisli_cc _ _ f) _).
intros d e. rewrite kleisli_cc_simpl.
intros kl. destruct (kleisliValVal kl) as [dd [P _]]. exists dd. auto.
Qed.


(* Empty cpo *)
Inductive DEmp : Set := .

Definition DZero : cpo := Discrete DEmp.

Definition DOne : cpo := Discrete unit.

Definition Sierp : cpo := DL DOne.

Lemma DOne_final: forall D (f g:D -C-> DOne), f == g.
intros. refine (fcont_eq_intro _). intros d. case (f d) ; case (g d) ; auto.
Qed.

Instance DOne_pointed : Pointed DOne.
exists tt.
intros d. case d. auto.
Defined.

Instance prod_pointed A B { pa : Pointed A} {pb : Pointed B} : Pointed (Dprod A B).
intros.
destruct pa.
destruct pb.
exists (PAIR A B PBot0 PBot1).
intro.
auto.
Defined.

Lemma K_com: forall (D E A:cpo) (f:D -C-> E) (e:A), @K D A e == @K E A e @_ f.
Proof.
intros D E A f e. apply fcont_eq_intro. intros d. rewrite fcont_comp_simpl. auto.
Qed.

Lemma K_simpl: forall (D E:cpo) (e:E) (d:D), K E d e = d.
auto.
Qed.

(*
Lemma KK_simpl: forall (D E:cpo) e d, @K E D d e == d.
auto.
Qed.*)

Instance fun_pointed A B {pb : Pointed B} : Pointed (A -C-> B).
intros; destruct pb.
exists (K A PBot0).
intro; simpl.
intro; auto.
Defined.

Lemma PBot_app D E {pd:Pointed E} : forall d, @PBot _ (fun_pointed D E) d == PBot.
intros D E pd d. split. case pd. auto.
auto.
Qed.

Lemma PBot_comp D F {pd:Pointed F} : forall E (h:E -C-> D),
            @PBot _ (fun_pointed D F) @_ h == @PBot _ (fun_pointed E F).
intros D F pd E h. apply fcont_eq_intro.
intros e. rewrite fcont_comp_simpl. rewrite PBot_app. rewrite PBot_app.
auto.
Qed.



(** ** Fixpoints *)

Section Fixpoints.
 Variable D : cpo.
 Context {dp : Pointed D}.

(* Variable D: cpo. *)
Variable f : D -m>D.

Hypothesis fcont : continuous f. 

Fixpoint iter_ n (* D [Pointed D] (f : D -m> D) *) : D := match n with O => PBot | S m => f (iter_ m (* f *)) end.

Lemma iter_incr (* D [Pointed D] (f : D -m> D)*) : forall n, iter_ n (* f *) <= f (iter_ n (*f *)).
induction n; simpl; auto.
(* apply Pleast. *)
Save.

Hint Resolve iter_incr.

Definition iter : natO -m> D.
exists iter_; red; intros.
induction H; simpl; auto.
apply Ole_trans with (iter_ m); auto.
Defined.

Definition fixp : D := lub iter.

Lemma fixp_le : fixp <= f fixp.
unfold fixp.
apply Ole_trans with (lub (fmon_comp f iter)); auto.
Save.
Hint Resolve fixp_le.

Lemma fixp_eq : fixp == f fixp.
apply Ole_antisym; auto.
unfold fixp.
apply Ole_trans with (1:= fcont iter).
rewrite (lub_lift_left iter (S O)); auto.
Save.

Lemma fixp_inv : forall g, f g <= g -> fixp <= g.
unfold fixp; intros.
apply lub_le.
induction n; intros; auto.
simpl; apply Ole_trans with (f g); auto.
Save.

End Fixpoints.
Hint Resolve fixp_le fixp_eq fixp_inv.

Definition fixp_cte D {pd:Pointed D} : forall (d:D), fixp (fmon_cte D d) == d.
intros; apply fixp_eq with (f:=fmon_cte D d); red; intros; auto.
unfold fmon_cte at 1; simpl.
apply le_lub with (f:=fmon_comp (fmon_cte D (O2:=D) d) h) (n:=O); simpl; auto.
Save.
Hint Resolve fixp_cte.


Lemma fixp_le_compat D {pd:Pointed D} : forall (*(D:cpo)*) (f g : D-m>D), f<=g -> fixp f <= fixp g.
intros; unfold fixp.
apply lub_le_compat.
intro n; induction n; simpl; auto.
apply Ole_trans with (g (iter_ (D:=D) f n)); auto.
Save.
Hint Resolve fixp_le_compat.


Add Parametric Morphism D {pd : Pointed D} : (@fixp D pd)
with signature (@Oeq (D -m>D)) ==> (@Oeq D)
as fixp_eq_compat.
intros; apply Ole_antisym; auto.
Save.
Hint Resolve fixp_eq_compat.

Definition Fixp D {pd: Pointed D} : (D-m>D) -m> D.
intros; exists (@fixp D pd (* (D:=D) *)).
auto.
Defined. 

Lemma Fixp_simpl D {pd:Pointed D} : forall (* (D:cpo)*) (f:D-m>D), Fixp f = fixp f.
trivial.
Save.

Definition Iter D {pd:Pointed D} : (D-M->D) -m> (natO -M->D).
intros; exists (fun (f:D-M->D) => iter (D:=D) f); red; intros.
intro n; induction n; simpl; intros; auto.
apply Ole_trans with (y (iter_ (D:=D) x n)); auto.
Defined.

Lemma IterS_simpl D {pd:Pointed D} : forall (* (D:cpo)*) f n, Iter f (S n) = f (Iter f n).
trivial.
Save.

Lemma iterS_simpl D {pd:Pointed D} : forall (* (D:cpo)*) f n, iter f (S n) = f (iter  f n).
trivial.
Save.

Lemma iter_continuous D {pd:Pointed D} :(* forall (D:cpo), *)
    forall h : natO -m> (D-M->D), (forall n, continuous (h n)) -> 
                  iter (lub h) <= lub (Iter (* D*) @ h).
red; intros; intro k.
induction k; auto.
(* botbot above just so auto works here *)

rewrite iterS_simpl.
apply Ole_trans with (lub h  (lub (c:=natO -M-> D) (Iter (*D*) @ h) k)); auto.
repeat (rewrite fmon_lub_simpl).
apply Ole_trans with
       (lub (c:=D) (lub (c:=natO-M->D) (double_app h ((Iter (* D*) @ h) <_> k)))).
apply lub_le_compat; simpl; intros.
apply Ole_trans with (lub ((h x)@((Iter (*D*) @ h) <_> k))); auto.
apply Ole_trans 
  with (lub (fmon_diag  (double_app h ((Iter (*D*) @ h) <_> k)))); auto.
case (double_lub_diag (double_app h ((Iter (*D*) @ h) <_> k))); intros L _.
apply Ole_trans with (1:=L); auto.
Save.

Hint Resolve iter_continuous.

Lemma iter_continuous_eq D {pd:Pointed D} :
    forall h : natO -m> (D-M->D), (forall n, continuous (h n)) -> 
                  iter (lub h) == lub (Iter (* D*) @ h).
intros; apply Ole_antisym; auto.
exact (lub_comp_le (Iter (*D*)) h).
Save.


Lemma fixp_continuous D {pd:Pointed D} : forall (* (D:cpo)*) (h : natO -m> (D-M->D)), 
       (forall n, continuous (h n)) -> fixp (lub h) <= lub (Fixp (*D*) @ h).
intros; unfold fixp.
rewrite (iter_continuous_eq H).
simpl; rewrite <- lub_exch_eq; auto.
Save.
Hint Resolve fixp_continuous.

Lemma fixp_continuous_eq D {pd:Pointed D} : forall (h : natO -m> (D -M-> D)), 
       (forall n, continuous (h n)) -> fixp (lub h) == lub (Fixp (*D*) @ h).
intros; apply Ole_antisym; auto.
exact (lub_comp_le (D1:=D -M-> D) (Fixp (*D*)) h).
Save.

Definition FIXP (D:cpo) {p:Pointed D} : (D-C->D) -c> D.
intros D p.
exists (Fixp @ (Fcontit D D)).
unfold continuous.
intros h.
rewrite fmon_comp_simpl.
rewrite Fixp_simpl.
apply Ole_trans with (fixp (D:=D) (lub (c:=D-M->D) (Fcontit D D@h))); auto.
apply Ole_trans with  (lub (Fixp (*D*) @ (Fcontit D D@h))); auto.
apply fixp_continuous; intros n.
change (continuous (D1:=D) (D2:=D) (fcontit (h n))); auto.
Defined.

Lemma FIXP_simpl D {pd:Pointed D} : forall (*(D:cpo)*) (f:D-c>D), FIXP (*D*) f = Fixp (*D*) (fcontit f).
trivial.
Save.

Lemma FIXP_le_compat D {pd:Pointed D} : forall (*(D:cpo)*) (f g : D-C->D),
            f <= g -> FIXP (*D*) f <= FIXP (*D*) g.
intros; repeat (rewrite FIXP_simpl); repeat (rewrite Fixp_simpl).
apply fixp_le_compat; auto.
Save.
Hint Resolve FIXP_le_compat.

Add Parametric Morphism (D:cpo) (p:Pointed D) : (@FIXP D p)
with signature (@Oeq (D -C-> D)) ==> (@Oeq D)
as FIXP_eq_compat.
intros; apply Ole_antisym; auto.
Qed.

Hint Resolve FIXP_eq_compat.
Hint Resolve FIXP_eq_compat_Morphism.

Lemma FIXP_eq D {pd:Pointed D} : forall (f:D-c>D), FIXP (*D*) f == f (FIXP (*D*) f).
intros; rewrite FIXP_simpl; rewrite Fixp_simpl.
apply (fixp_eq (f:=fcontit f)).
auto.
Save.
Hint Resolve FIXP_eq.

Lemma FIXP_com: forall E D (f:E -C-> (D -C-> D)) {pd:Pointed D}, FIXP @_ f == EV _ _ @_ PROD_fun f (FIXP @_ f).
intros E D f P. apply fcont_eq_intro. intros e. repeat (rewrite fcont_comp_simpl).
rewrite FIXP_eq. auto.
Qed.

Lemma FIXP_inv D {pd:Pointed D} : forall (f:D-c>D)(g : D), f g <= g -> FIXP (*D*) f <= g.
intros; rewrite FIXP_simpl; rewrite Fixp_simpl; apply fixp_inv; auto.
Save.

(** *** Iteration of functional *)
Lemma FIXP_comp_com D {pd:Pointed D} : forall (f g:D-c>D),
       g @_ f <= f @_ g-> FIXP (*D*) g <= f (FIXP (*D*) g).
intros; apply FIXP_inv.
apply Ole_trans with (f (g (FIXP (*D*) g))).
assert (X:=fcont_le_elim H (FIXP (*D*) g)). repeat (rewrite fcont_comp_simpl in X). apply X.
apply fcont_monotonic.
case (FIXP_eq g); trivial.
Save.

Lemma FIXP_comp D {pd:Pointed D} : forall (f g:D-c>D),
       g @_ f <= f @_ g -> f (FIXP (*D*) g) <= FIXP (*D*) g -> FIXP (*D*) (f @_ g) == FIXP (*D*) g.
intros; apply Ole_antisym.
(* fix f @_ g <= fix g *)
apply FIXP_inv.
rewrite fcont_comp_simpl.
apply Ole_trans with (f (FIXP (*D*) g)); auto.
(* fix g <= fix f @_ g *)
apply FIXP_inv.
assert (g (f (FIXP (*D*) (f @_ g))) <= f (g (FIXP (*D*) (f @_ g)))).
specialize (H (FIXP (f @_ g))). repeat (rewrite fcont_comp_simpl in H). apply H.
case (FIXP_eq (f@_g)); intros.
apply Ole_trans with (2:=H3).
apply Ole_trans with (2:=H1).
apply fcont_monotonic.
apply FIXP_inv.
rewrite fcont_comp_simpl.
apply fcont_monotonic.
apply Ole_trans with (1:=H1); auto.
Save.

Fixpoint fcont_compn D {pd:Pointed D} (f:D-c>D) (n:nat) {struct n} : D-c>D := 
             match n with O => f | S p => fcont_compn f p @_ f end.

Lemma fcont_compn_com D {pd:Pointed D} : forall (f:D-c>D) (n:nat), 
            f @_ (fcont_compn f n) <= fcont_compn f n @_ f.
induction n; auto.
simpl fcont_compn.
apply Ole_trans with ((f @_ fcont_compn (D:=D) f n) @_ f); auto.
Save.

Lemma FIXP_compn D {pd:Pointed D} : 
     forall  (f:D-c>D) (n:nat), FIXP (fcont_compn f n) == FIXP f.
induction n; auto.
simpl fcont_compn.
apply FIXP_comp.
apply fcont_compn_com.
apply Ole_trans with (fcont_compn (D:=D) f n (FIXP (fcont_compn (D:=D) f n))); auto.
apply Ole_trans with (FIXP (fcont_compn (D:=D) f n)); auto.
Save.

Lemma fixp_double D {pd:Pointed D} : forall (f:D-c>D), FIXP (f @_ f) == FIXP f.
intros; exact (FIXP_compn f (S O)).
Save.


(** *** Induction principle *)
Definition admissible (D:cpo) (P:D->Prop) := 
          forall f : natO -m> D, (forall n, P (f n)) -> P (lub f).

Definition fixp_ind D {pd:Pointed D} : forall  (F:D -m> D)(P:D->Prop),
       admissible P -> P PBot -> (forall x, P x -> P (F x)) -> P (fixp F).
intros D pD F P Adm Pb Pc; unfold fixp.
apply Adm. intros n.
induction n; simpl; auto.
Defined.

Definition SubOrd (D:ord) (P:D -> Prop) : ord.
intros D I. exists ({x : tord D | I x})
   (fun (x' y:{x : D | I x}) => match x' with exist x _ => match y with exist y
_ => @Ole D x y end end).
intros x1. case x1. auto.
intros x1. case x1. clear x1. intros x ix y1. case y1. clear y1.
intros y iy z1. case z1. clear z1. intros z iz l1 l2.
apply (Ole_trans l1 l2).
Defined.

Definition SubOrde (D:cpo) (P:D -> Prop) (d:D)  : P d -> SubOrd P.
intros. exists d. auto.
Defined.

Definition InheritFunm (D E:cpo) (Q:E->Prop) (f:D -m> E) (p:forall d, Q (f d)) :
 D -m> SubOrd Q.
intros D E Q f p.
exists (fun d => @SubOrde E Q (f d) (p d)).
unfold monotonic. intros x y lxy. auto.
Defined.

Definition Forgetm D P : (@SubOrd D P) -m> D.
intros D P. exists (fun (O:SubOrd P) => match O with exist x _ => x end).
unfold monotonic. intros x. case x. clear x. intros x px y. case y. clear y.
intros ; auto.
Defined.

Definition Subchainlub (D:cpo) (P:D -> Prop) (I:admissible P) (c:natO -m> (SubOrd
 P)) : SubOrd P.
intros D P I c.
exists (@lub D (Forgetm P @ c)).
unfold admissible in I. specialize (I (Forgetm P @ c)).
apply I. intros i. simpl. case (c i). auto.
Defined.

Definition SubCPO (D:cpo) (P:D -> Prop) (I: admissible P) : cpo.
intros D P I. exists (SubOrd P) (Subchainlub I).
intros c n. unfold Subchainlub.  simpl. case_eq (c n). intros x Px cn.
refine (Ole_trans _ (@le_lub _ (Forgetm P @ c) n)).
simpl. rewrite cn. auto.
intros c d C. unfold Subchainlub. simpl.
case_eq d. intros dd pd de.
refine (lub_le _). intros n. simpl. specialize (C n).
case_eq (c n). intros d1 pd1 cn. rewrite cn in C.
simpl in C. rewrite de in C. auto.
Defined.

Definition InheritFun (D E:cpo) (Q:E -> Prop) (IQ : admissible Q) (f:D -c> E)
                  (p:forall d, Q (f d)) : D -c> SubCPO IQ.
intros. exists (InheritFunm p). unfold continuous.
intros c. simpl.
rewrite lub_comp_eq.
apply lub_le. intros n.
refine (Ole_trans _ (le_lub (Forgetm Q @ (InheritFunm p @ c)) n)).
auto. auto.
Defined.

Lemma InheritFun_simpl: forall (D E:cpo) (Q:E -> Prop) (IQ : admissible Q) (f:D -c> E)
                  (p:forall d, Q (f d)) d, InheritFun IQ p d = InheritFunm p d.
intros. auto.
Qed.

Definition Forget (D:cpo) (P:D -> Prop) (I: admissible P) : (SubCPO I) -c> D.
intros D P I. exists (@Forgetm D P).
unfold continuous. auto.
Defined.

Lemma ForgetP D P fs d : P (@Forget D P fs d).
intros. case d. auto.
Qed.

Add Parametric Morphism (D:cpo) (P:D -> Prop) (I:admissible P) : (@Forget D P I)
with signature (@Ole _) ++> (@Ole _)
as Forget_le_compat.
intros d0 d1 dle. auto.
Qed.

Add Parametric Morphism (D:cpo) (P:D -> Prop) (I:admissible P) : (@Forget D P I)
with signature (@Oeq _) ==> (@Oeq _)
as Forget_eq_compat.
intros d0 d1 dle. split ; auto.
Qed.

Lemma Forget_leinj: forall (D:cpo) (P:D -> Prop) (I:admissible P) (d:SubCPO I) e, Forget I d <= Forget I e -> d <= e.
intros D P I d e. case e. clear e. intros e Pe. case d. clear d. intros d Pd.
auto.
Qed.

Hint Resolve Forget_leinj.

Lemma Forget_inj : forall (D:cpo) (P:D -> Prop) (I:admissible P) (d:SubCPO I) e, Forget I d == Forget I e -> d == e.
intros. split ; auto.
Qed.

Hint Resolve Forget_inj.

Lemma InheritFun_eq_compat D E Q Qcc f g X XX : Oeq f g ->
    (@InheritFun D E Q Qcc f X) == (@InheritFun D E Q Qcc g XX).
intros. refine (fcont_eq_intro _). intros d. auto.
Qed.

Lemma ForgetInherit (D E:cpo) (P:D -> Prop) PS f B : Forget PS @_ @InheritFun E D P PS f B == f.
intros D E P PS f B.
refine (fcont_eq_intro _). intros d.
rewrite fcont_comp_simpl. rewrite InheritFun_simpl.
simpl. auto.
Qed.

Lemma InheritFun_comp: forall D E F P I (f:E -C-> F) X (g:D -C-> E) XX,
      @InheritFun _ F P I f X @_ g == @InheritFun _ F P I (f @_ g) XX.
intros. refine (fcont_eq_intro _). intros d. repeat (rewrite fcont_comp_simpl).
split ; auto.
Qed.
